#!/usr/bin/env python
# -*- coding: utf-8 -*-

##########################################
## NOTE: This module was autogenerated. ##
## Contains no user-servicable parts!!! ##
##########################################

import re

VERSION = 20210419.190716

def rei(regex: str) -> re.Pattern:
    """Return compiled `re.Pattern` with `regex` as pattern, and the IGNORECASE flag.

    Args:
        regex (str): Regular expression pattern.

    Returns:
        re.Pattern: Compiled version of `regex`, with IGNORECASE flag.
    """
    return re.compile(regex, flags=re.I)

modern_plural_of = {
    "Achinese": "Achinese",
    "Alfurese": "Alfurese",
    "Amboinese": "Amboinese",
    "Americanese": "Americanese",
    "Amoyese": "Amoyese",
    "Andamanese": "Andamanese",
    "Angolese": "Angolese",
    "Annamese": "Annamese",
    "Antananarivo": "Antananarivos",
    "Aragonese": "Aragonese",
    "Arakanese": "Arakanese",
    "Asherah": "Asherahs",
    "Assamese": "Assamese",
    "Atlas": "Atlantes",
    "Auslese": "Auslesen",
    "Avignonese": "Avignonese",
    "Azerbaijanese": "Azerbaijanese",
    "Bahaman": "Bahamans",
    "Balinese": "Balinese",
    "Bamako": "Bamakos",
    "Barquisimeto": "Barquisimetos",
    "Bengalese": "Bengalese",
    "Bernese": "Bernese",
    "Bhutanese": "Bhutanese",
    "Biro": "Biros",
    "Bolognese": "Bolognese",
    "Bolzano": "Bolzanos",
    "Borghese": "Borghese",
    "Bostonese": "Bostonese",
    "Boto": "Botos",
    "Buginese": "Buginese",
    "Burman": "Burmans",
    "Burmese": "Burmese",
    "Cairo": "Cairos",
    "Calabrese": "Calabrese",
    "Camaldolese": "Camaldolese",
    "Cantonese": "Cantonese",
    "Carlylese": "Carlylese",
    "Cassinese": "Cassinese",
    "Celanese": "Celanese",
    "Ceylonese": "Ceylonese",
    "Chicago": "Chicagos",
    "Chimango": "Chimangos",
    "Chinese": "Chinese",
    "Colombo": "Colombos",
    "Colorado": "Colorados",
    "Congoese": "Congoese",
    "Congolese": "Congolese",
    "Czech": "Czechs",
    "Dongolese": "Dongolese",
    "Draco": "Dracos",
    "Esperanto": "Esperantos",
    "Faro": "Faros",
    "Faroese": "Faroese",
    "Ferrarese": "Ferrarese",
    "Filipino": "Filipinos",
    "Foochowese": "Foochowese",
    "Gabunese": "Gabunese",
    "Galla": "Gallas",
    "Genevese": "Genevese",
    "Genoese": "Genoese",
    "German": "Germans",
    "Gestapo": "Gestapos",
    "Gilbertese": "Gilbertese",
    "Goanese": "Goanese",
    "Greensboro": "Greensboros",
    "Guaiabero": "Guaiaberos",
    "Guianese": "Guianese",
    "Hainanese": "Hainanese",
    "Hararese": "Hararese",
    "Harlemese": "Harlemese",
    "Havanese": "Havanese",
    "Heavenese": "Heavenese",
    "Hiroshiman": "Hiroshimans",
    "Hoosierese": "Hoosierese",
    "Hottentotese": "Hottentotese",
    "Hunanese": "Hunanese",
    "I": "we",
    "Idaho": "Idahos",
    "Indochinese": "Indochinese",
    "Iquico": "Iquicos",
    "Japanese": "Japanese",
    "Javanese": "Javanese",
    "Jerry": "Jerrys",
    "Johnsonese": "Johnsonese",
    "Kakapo": "Kakapos",
    "Kanarese": "Kanarese",
    "Kinkimavo": "Kinkimavos",
    "Kiplingese": "Kiplingese",
    "Kokako": "Kokakos",
    "Kongoese": "Kongoese",
    "Kongolese": "Kongolese",
    "Kosovo": "Kosovos",
    "Lapith": "Lapiths",
    "Lapponese": "Lapponese",
    "Lebanese": "Lebanese",
    "Leonese": "Leonese",
    "Lesotho": "Lesothos",
    "Lilo": "Lilos",
    "Liman": "Limans",
    "Londonese": "Londonese",
    "Lorrainese": "Lorrainese",
    "Lucchese": "Lucchese",
    "Lyonese": "Lyonese",
    "Macanese": "Macanese",
    "Macassarese": "Macassarese",
    "Madurese": "Madurese",
    "Majuro": "Majuros",
    "Malabarese": "Malabarese",
    "Malabo": "Malabos",
    "Maltese": "Maltese",
    "Maputo": "Maputos",
    "Maracaibo": "Maracaibos",
    "Mary": "Marys",
    "Messinese": "Messinese",
    "Mexico": "Mexicos",
    "Milanese": "Milanese",
    "Milano": "Milanos",
    "Modenese": "Modenese",
    "Monaco": "Monacos",
    "Montenegro": "Montenegros",
    "Morocco": "Moroccos",
    "Muqdisho": "Muqdishos",
    "Muranese": "Muranese",
    "Nakayaman": "Nakayamans",
    "Nankingese": "Nankingese",
    "Navarrese": "Navarrese",
    "Nenets": "Nenets",
    "Nepalese": "Nepalese",
    "Niasese": "Niasese",
    "Nicobarese": "Nicobarese",
    "Ningbo": "Ningbos",
    "Nipponese": "Nipponese",
    "Norman": "Normans",
    "Oklahoman": "Oklahomans",
    "Orinoco": "Orinocos",
    "Orlando": "Orlandos",
    "Oslo": "Oslos",
    "Otomi": "Otomies",
    "Panaman": "Panamans",
    "Pangasinan": "Pangasinans",
    "Paramaribo": "Paramaribos",
    "Pardusco": "Parduscos",
    "Parmese": "Parmese",
    "Pekingese": "Pekingese",
    "Piedmontese": "Piedmontese",
    "Pistoiese": "Pistoiese",
    "Pluto": "Plutos",
    "Polonese": "Polonese",
    "Porto": "Portos",
    "Porto Novo": "Porto Novos",
    "Porto-Novo": "Porto-Novos",
    "Portuguese": "Portuguese",
    "Purum": "Purums",
    "Quito": "Quitos",
    "Rom": "Roma",
    "Romagnese": "Romagnese",
    "Roman": "Romans",
    "Romanese": "Romanese",
    "Romany": "Romanies",
    "Sacramento": "Sacramentos",
    "Sangirese": "Sangirese",
    "Santiago": "Santiagos",
    "Sapporo": "Sapporos",
    "Sarajevo": "Sarajevos",
    "Sarawakese": "Sarawakese",
    "Selman": "Selmans",
    "Senegalese": "Senegalese",
    "Shavese": "Shavese",
    "Shawanese": "Shawanese",
    "Siamese": "Siamese",
    "Sienese": "Sienese",
    "Sikkimese": "Sikkimese",
    "Silurus": "Siluri",
    "Singhalese": "Singhalese",
    "Sinhalese": "Sinhalese",
    "Sogdianese": "Sogdianese",
    "Sonaman": "Sonamans",
    "Spaetlese": "Spaetlesen",
    "Sudanese": "Sudanese",
    "Sundanese": "Sundanese",
    "Swahilese": "Swahilese",
    "Tacoman": "Tacomans",
    "Taiko": "Taikos",
    "Tenggerese": "Tenggerese",
    "Timorese": "Timorese",
    "Tirolese": "Tirolese",
    "Tocharese": "Tocharese",
    "Togo": "Togos",
    "Tokyo": "Tokyos",
    "Tonkinese": "Tonkinese",
    "Torinese": "Torinese",
    "Torino": "Torinos",
    "Tornese": "Tornesi",
    "Toronto": "Torontos",
    "Tyrolese": "Tyrolese",
    "Vermontese": "Vermontese",
    "Veronese": "Veronese",
    "Viennese": "Viennese",
    "Vietnamese": "Vietnamese",
    "Virgo": "Virgos",
    "Wenchowese": "Wenchowese",
    "Whitmanese": "Whitmanese",
    "Yakiman": "Yakimans",
    "Yamoussoukro": "Yamoussoukros",
    "Yengeese": "Yengeese",
    "Yokohaman": "Yokohamans",
    "Yuman": "Yumans",
    "Yunnanese": "Yunnanese",
    "Zibo": "Zibos",
    "abacus": "abacuses",
    "aboiteau": "aboiteaus",
    "abortorium": "abortoria",
    "abortus": "abortuses",
    "abscissa": "abscissas",
    "abstractum": "abstracta",
    "acarus": "acari",
    "achinese": "achinese",
    "acicula": "aciculae",
    "acinus": "acini",
    "acroama": "acroamata",
    "acropolis": "acropolises",
    "actinia": "actinias",
    "aculeus": "aculei",
    "addendum": "addendums",
    "adieu": "adieus",
    "adjutant general": "adjutant generals",
    "adminiculum": "adminicula",
    "ado": "ados",
    "adviso": "advisoes",
    "adytum": "adyta",
    "aecidium": "aecidia",
    "aegis": "aegises",
    "aelf": "aelfe",
    "aficionado": "aficionados",
    "afreet": "afreets",
    "afrit": "afrits",
    "agape": "agapes",
    "agend": "agends",
    "agendum": "agenda",
    "agent provocateur": "agents provocateurs",
    "aggro": "aggros",
    "agon": "agones",
    "agora": "agorae",
    "agrogorod": "agrogoroda",
    "aide de camp": "aides de camp",
    "aide-de-camp": "aides-de-camp",
    "ala": "alae",
    "alan": "alans",
    "albergo": "alberghi",
    "albino": "albinos",
    "ale wife": "ale wives",
    "ale-wife": "ale-wives",
    "alfurese": "alfurese",
    "alga": "algae",
    "alguacil": "alguacils",
    "alias": "aliases",
    "allargando": "allargandi",
    "allegro": "allegros",
    "alluvium": "alluviums",
    "alto": "altos",
    "alto relievo": "alto relievos",
    "alto-relievo": "alto-relievos",
    "altorelievo": "altorelievos",
    "alula": "alulae",
    "alumna": "alumnae",
    "alumnus": "alumni",
    "alveolus": "alveoli",
    "ambo": "ambos",
    "amboinese": "amboinese",
    "ambulacrum": "ambulacra",
    "ambuscado": "ambuscados",
    "amentum": "amenta",
    "americanese": "americanese",
    "amicus curiae": "amici curiae",
    "ammo": "ammos",
    "amoeba": "amoebas",
    "amoretto": "amoretti",
    "amorino": "amorini",
    "amoyese": "amoyese",
    "amphipneust": "amphipneusts",
    "amphora": "amphorae",
    "anacoluthon": "anacoluthons",
    "anathema": "anathemas",
    "ancona": "ancone",
    "andamanese": "andamanese",
    "angioma": "angiomata",
    "angolese": "angolese",
    "angustia": "angustiae",
    "anlage": "anlages",
    "annamese": "annamese",
    "anta": "antae",
    "antananarivo": "antananarivos",
    "antefix": "antifixes",
    "antenna": "antennas",
    "anthemion": "anthemia",
    "anthodium": "anthodia",
    "anti serum": "anti serums",
    "anti-serum": "anti-serums",
    "antiserum": "antisera",
    "antrum": "antra",
    "apex": "apexes",
    "aphelion": "aphelia",
    "aphis": "aphides",
    "apothecium": "apothecia",
    "apparatchik": "apparatchiks",
    "apparatus": "apparatuses",
    "appendix": "appendixes",
    "appressorium": "appressoria",
    "apsis": "apsides",
    "apterium": "apteria",
    "aptychus": "aptychi",
    "aquarium": "aquariums",
    "aragonese": "aragonese",
    "arakanese": "arakanese",
    "arango": "arangoes",
    "arboretum": "arboreta",
    "arcanum": "arcana",
    "archegonium": "archegonia",
    "archipelago": "archipelagos",
    "arcosolium": "arcosolia",
    "areola": "areolae",
    "argumentum": "argumenta",
    "arista": "aristae",
    "armadillo": "armadillos",
    "arrha": "arrhae",
    "asbestos": "asbestoses",
    "ascidium": "ascidia",
    "asherah": "asherahs",
    "assamese": "assamese",
    "asterias": "asteriae",
    "astrocytoma": "astrocytomas",
    "asylum": "asylums",
    "asyndeton": "asyndeta",
    "ataman": "atamans",
    "atlas": "atlases",
    "atman": "atmas",
    "aureus": "aurei",
    "aurora": "auroras",
    "auslese": "ausleses",
    "auspex": "auspices",
    "auto": "autos",
    "auto de fe": "auto de fes",
    "auto-de-fe": "auto-de-fes",
    "autobahn": "autobahns",
    "autochthon": "autochthons",
    "autostrada": "autostradas",
    "avicularium": "avicularia",
    "avignonese": "avignonese",
    "aviso": "avisos",
    "avocado": "avocados",
    "ax": "axes",
    "axe": "axes",
    "axilla": "axillae",
    "axis": "axes",
    "axon": "axons",
    "axopodium": "axopodia",
    "azerbaijanese": "azerbaijanese",
    "baal": "baals",
    "bacillus": "bacilli",
    "bacterium": "bacteria",
    "bahaman": "bahamans",
    "bajocco": "bajocchi",
    "balinese": "balinese",
    "ballabile": "ballabili",
    "ballata": "ballate",
    "ballerina": "ballerinas",
    "ballista": "ballistas",
    "bamako": "bamakos",
    "bandeau": "bandeaus",
    "bandit": "bandits",
    "barquisimeto": "barquisimetos",
    "barrico": "barricoes",
    "base": "bases",
    "basilica": "basilicas",
    "basis": "bases",
    "bass": "basses",
    "basso": "bassos",
    "bateau": "bateaus",
    "bath": "baths",
    "bathos": "bathoses",
    "batz": "batzes",
    "beau": "beaus",
    "beau geste": "beau gestes",
    "bechuana": "bechuanas",
    "beef": "beefs",
    "beerenauslese": "beerenauslesen",
    "bel esprit": "beaux esprits",
    "bel homme": "beaux hommes",
    "belle epoque": "belle epoques",
    "bema": "bemas",
    "bengalese": "bengalese",
    "bernese": "bernese",
    "bersagliere": "bersaglieri",
    "bhikshu": "bhikshus",
    "bhutanese": "bhutanese",
    "bias": "biases",
    "bierstube": "bierstubes",
    "bifolium": "bifolia",
    "bijou": "bijoux",
    "bilbo": "bilboes",
    "bimbo": "bimbos",
    "bingo": "bingos",
    "biro": "biros",
    "biscotto": "biscotti",
    "bison": "bisons",
    "blastema": "blastemata",
    "blemya": "blemyae",
    "blin": "blinis",
    "blitz": "blitzes",
    "blouse": "blouses",
    "blue": "blues",
    "blues": "blues",
    "bodegon": "bodegones",
    "bolero": "boleros",
    "bolognese": "bolognese",
    "bolus": "boluses",
    "bolzano": "bolzanos",
    "bon mot": "bon mots",
    "bon vivant": "bon vivants",
    "bona fides": "bonis fidebus",
    "bongo": "bongos",
    "bordereau": "bordereaux",
    "borghese": "borghese",
    "bostonese": "bostonese",
    "boto": "botos",
    "bozzetto": "bozzetti",
    "braccio": "braccia",
    "brachiopod": "brachiopods",
    "branchiopod": "branchiopods",
    "bratwurst": "bratwursts",
    "bravo": "bravos",
    "break away": "break aways",
    "break-away": "break-aways",
    "breakaway": "breakaways",
    "bream": "bream",
    "breech": "breeches",
    "breeches": "breeches",
    "bregma": "bregmata",
    "brigadier general": "brigadier generals",
    "britches": "britches",
    "bronchitis": "bronchitises",
    "bronchus": "bronchi",
    "brother": "brothers",
    "brother german": "brothers german",
    "brother-german": "brothers-german",
    "bruschetta": "bruschette",
    "bubo": "buboes",
    "buffalo": "buffalos",
    "buginese": "buginese",
    "bulimus": "bulimi",
    "bulla": "bullae",
    "bureau": "bureaus",
    "burman": "burmans",
    "burmese": "burmese",
    "burro": "burros",
    "bursa": "bursae",
    "bursch": "burschen",
    "bursitis": "bursitises",
    "busto": "bustos",
    "buzz": "buzzes",
    "bylina": "bylinas",
    "cabaletta": "cabalettas",
    "cactus": "cactuses",
    "caddis": "caddises",
    "caduceus": "caducei",
    "caecum": "caeca",
    "caffe macchiato": "caffe macchiatos",
    "caffe ristretto": "caffe ristrettos",
    "caiman": "caimans",
    "cairo": "cairos",
    "calabrese": "calabrese",
    "calamistrum": "calamistra",
    "calathus": "calathi",
    "calix": "calices",
    "callus": "calluses",
    "calypso": "calypsos",
    "calzone": "calzones",
    "camaldolese": "camaldolese",
    "camera obscura": "camera obscuras",
    "canaliculus": "canaliculi",
    "candelabrum": "candelabra",
    "cannabis": "cannabises",
    "cantharus": "canthari",
    "canto": "cantos",
    "cantonese": "cantonese",
    "cantus": "cantuses",
    "canvas": "canvases",
    "canzonetta": "canzonettas",
    "capataz": "capataces",
    "capo": "capos",
    "cappuccino": "cappuccinos",
    "carabiniere": "carabinieri",
    "carcinoma": "carcinomas",
    "cardo": "cardines",
    "cardophagus": "cardophagi",
    "carex": "carices",
    "cargador": "cargadores",
    "caribou": "caribous",
    "carlylese": "carlylese",
    "carmen": "carmens",
    "carp": "carp",
    "carte de visite": "cartes de visite",
    "carte-de-visite": "cartes-de-visite",
    "caryatid": "caryatids",
    "caryopsis": "caryopsides",
    "casino": "casinos",
    "cassinese": "cassinese",
    "cassone": "cassones",
    "castrato": "castrati",
    "castrum": "castra",
    "casus belli": "casus belli",
    "cathisma": "cathismata",
    "caudex": "caudices",
    "caulis": "caules",
    "cavea": "caveae",
    "cayman": "caymans",
    "cecidium": "cecidia",
    "celanese": "celanese",
    "celia": "celiae",
    "cello": "cellos",
    "cementoma": "cementomata",
    "centesimo": "centesimi",
    "centumvir": "centumviri",
    "cercus": "cerci",
    "ceriman": "cerimans",
    "ceylonese": "ceylonese",
    "chaeta": "chaetae",
    "chaos": "chaoses",
    "chapeau": "chapeaus",
    "chapplis": "chapplies",
    "charisma": "charismas",
    "chassis": "chassis",
    "chateau": "chateaus",
    "cherub": "cherubs",
    "chervonetz": "chervontzi",
    "cheval": "chevaux",
    "cheval de bataille": "chevaux de bataille",
    "cheval-de-bataille": "chevaux-de-bataille",
    "chicago": "chicagos",
    "child": "children",
    "chimango": "chimangos",
    "chinese": "chinese",
    "chintz": "chintzes",
    "chitarrone": "chitarroni",
    "chlamydia": "chlamydias",
    "choana": "choanae",
    "chorda": "chordae",
    "chorus": "choruses",
    "chou": "choux",
    "chromidium": "chromidia",
    "chromonema": "chromonemata",
    "chrysalis": "chrysalises",
    "cicatrix": "cicatrices",
    "cilantro": "cilantros",
    "cimex": "cimices",
    "cinclis": "cinclides",
    "cistophorus": "cistophorori",
    "cladus": "cladi",
    "claustrum": "claustra",
    "clausula": "clausulae",
    "clavula": "clavulae",
    "cleithrum": "cleithra",
    "clepsydra": "clepsydras",
    "clipper": "clippers",
    "clippers": "clippers",
    "clitoris": "clitorises",
    "clostridium": "clostridia",
    "co": "coes",
    "coagulum": "coagula",
    "coccidium": "coccidia",
    "cochito": "cochitos",
    "coco": "cocos",
    "cod": "cod",
    "codex": "codices",
    "coe": "coes",
    "coelia": "coeliae",
    "cognoscente": "cognoscenti",
    "coitus": "coituses",
    "collegium": "collegia",
    "collegium musicum": "collegia musica",
    "colombo": "colombos",
    "colorado": "colorados",
    "columbarium": "columbaria",
    "columna": "columnae",
    "coma": "comae",
    "comatula": "comatulae",
    "commando": "commandos",
    "compendium": "compendiums",
    "competent": "competentes",
    "concertino": "concertinos",
    "concerto grosso": "concerti grossi",
    "concetto": "concetti",
    "condo": "condos",
    "condottiere": "condottieri",
    "conductus": "conducti",
    "coney": "coneys",
    "conferva": "confervae",
    "confluvium": "confluvia",
    "congius": "congii",
    "congoese": "congoese",
    "congolese": "congolese",
    "conidium": "conidia",
    "conoscente": "conoscenti",
    "consortium": "consortiums",
    "contadina": "contadine",
    "contadino": "contadini",
    "contagium": "contagia",
    "contango": "contangos",
    "contrafactum": "contrafacta",
    "contralto": "contraltos",
    "contretemps": "contretemps",
    "conversazione": "conversaziones",
    "convolvulus": "convolvuluses",
    "cony": "conys",
    "corgi": "corgis",
    "corgy": "corgys",
    "cornetto": "cornetti",
    "corno": "corni",
    "cornu": "cornua",
    "corona": "coronas",
    "corps": "corps",
    "corpus": "corpuses",
    "corpus vile": "corpora vilia",
    "corpusculum": "corpuscula",
    "corrigendum": "corrigenda",
    "cortex": "cortexes",
    "cosmos": "cosmoses",
    "costa": "costae",
    "coup d'etat": "coups d'etat",
    "court martial": "court martials",
    "court-martial": "court-martials",
    "cow": "cows",
    "coxendix": "coxendices",
    "coz": "cozes",
    "craft": "craft",
    "crafts child": "crafts childer",
    "crafts-child": "crafts-childer",
    "cranium": "craniums",
    "craspedon": "craspeda",
    "craspedum": "craspeda",
    "crazia": "crazie",
    "credo": "credos",
    "cremaster": "cremasters",
    "crescendo": "crescendos",
    "cri de coeur": "cris de coeur",
    "crista": "cristae",
    "criterion": "criteria",
    "crowfoot": "crowfoots",
    "crus": "crura",
    "crusta": "crustae",
    "cubiculum": "cubicula",
    "cul de four": "culs de four",
    "cul de lampe": "culs de lampe",
    "cul de sac": "culs de sac",
    "cul-de-four": "culs-de-four",
    "cul-de-lampe": "culs-de-lampe",
    "cul-de-sac": "culs-de-sac",
    "cultellus": "cultelli",
    "cumulus": "cumuli",
    "cuniculus": "cuniculi",
    "curioso": "curiosos",
    "curriculum": "curriculums",
    "curriculum vitae": "curricula vitae",
    "cuspis": "cuspides",
    "cyano": "cyanos",
    "cyathus": "cyathi",
    "cyclops": "cyclopses",
    "cynocephalus": "cynocephali",
    "cyphella": "cyphellae",
    "cysticercus": "cysticerci",
    "cystidium": "cystidia",
    "cystoma": "cystomata",
    "czech": "czechs",
    "dace": "daces",
    "dais": "daises",
    "datum": "data",
    "debris": "debris",
    "decennium": "decennia",
    "deer": "deer",
    "delirium": "deliriums",
    "demo": "demos",
    "denarius": "denarii",
    "denkmal": "denkmaler",
    "denotatum": "denotata",
    "dentalium": "dentalia",
    "depositum": "depositums",
    "descriptum": "descripta",
    "desideratum": "desiderata",
    "designatum": "designata",
    "desma": "desmas",
    "desman": "desmans",
    "devoto": "devotos",
    "diabetes": "diabetes",
    "diastema": "diastemata",
    "dichasium": "dichasia",
    "dictum": "dictums",
    "didymis": "didymes",
    "die": "dice",
    "differentia": "differentiae",
    "digitalis": "digitalises",
    "dilettante": "dilettantes",
    "diplococcus": "diplococci",
    "diploma": "diplomas",
    "diporpa": "diporpae",
    "dipsas": "dipsades",
    "directrix": "directrices",
    "ditto": "dittos",
    "diverticulum": "diverticula",
    "divertimento": "divertimentos",
    "djinn": "djinns",
    "do": "does",
    "doe": "does",
    "dogma": "dogmas",
    "dolman": "dolmans",
    "dolphinarium": "dolphinariums",
    "dongolese": "dongolese",
    "draco": "dracos",
    "dragoman": "dragomen",
    "drama": "dramas",
    "dreikanter": "dreikanters",
    "dumka": "dumkas",
    "duumvir": "duumvirs",
    "dynamo": "dynamos",
    "edema": "edemas",
    "editio princeps": "editiones principes",
    "efreet": "efreets",
    "eland": "elands",
    "elf": "elves",
    "elk": "elks",
    "ellipse": "ellipses",
    "ellipsis": "ellipses",
    "emblema": "emblememata",
    "embryo": "embryos",
    "embryoma": "embryomata",
    "emporium": "emporiums",
    "enconium": "enconiums",
    "endothelioma": "endotheliomata",
    "enema": "enemas",
    "enigma": "enigmas",
    "ens rationis": "entia rationis",
    "ens reale": "entia realia",
    "entente cordiale": "ententes cordiales",
    "enteron": "entera",
    "entrochus": "entrochi",
    "ephelis": "ephelides",
    "ephemera": "ephemeras",
    "ephemeris": "ephemerises",
    "epicardium": "epicardia",
    "epidermis": "epidermises",
    "epididymis": "epididymises",
    "epilimnion": "epilimnia",
    "epimeron": "epimera",
    "epiphenomenon": "epiphenomena",
    "epithelioma": "epitheliomata",
    "epyllion": "epyllia",
    "equinoctium": "equinoctiums",
    "equisetum": "equisetums",
    "erf": "erfs",
    "ergastulum": "ergastula",
    "eros": "eroses",
    "erratum": "errata",
    "espanol": "espanoles",
    "esperanto": "esperantos",
    "espresso": "espressos",
    "ethos": "ethoses",
    "eucalyptus": "eucalyptuses",
    "eunuch": "eunuchs",
    "euporiston": "euporista",
    "euro": "euros",
    "evirato": "evirati",
    "ex": "exes",
    "exanthema": "exanthemata",
    "exemplum": "exempla",
    "exencephalus": "exencephali",
    "explanandum": "explananda",
    "explicandum": "explicanda",
    "extispex": "extispices",
    "extremum": "extremums",
    "eyas": "eyases",
    "fabella": "fabellae",
    "fabliau": "fabliaux",
    "faciendum": "facienda",
    "faecula": "faeculae",
    "fait accompli": "faits accomplis",
    "falsetto": "falsettos",
    "falx": "falces",
    "familia": "familiae",
    "famulus": "famuli",
    "fanfoot": "fanfoots",
    "farman": "farmans",
    "faro": "faros",
    "faroese": "faroese",
    "fascia": "fascias",
    "fasciculus": "fasciculi",
    "fauna": "faunas",
    "favella": "favellae",
    "faveolus": "faveoli",
    "favissa": "favissae",
    "fecula": "feculae",
    "fedai": "fedais",
    "feis": "fess",
    "felo": "felones",
    "felo de se": "felos de se",
    "felo-de-se": "felos-de-se",
    "felsenmeer": "felsenmeere",
    "femur": "femurs",
    "fenestra": "fenestrae",
    "fermata": "fermatas",
    "ferrarese": "ferrarese",
    "fiasco": "fiascos",
    "fibra": "fibrae",
    "fibrilla": "fibrillae",
    "fibroma": "fibromata",
    "fibula": "fibulas",
    "fidei defensor": "fidei defensores",
    "figurante": "figurantes",
    "filaria": "filarias",
    "filipino": "filipinos",
    "finger end": "finger ends",
    "finger-end": "finger-ends",
    "fioritura": "fioriture",
    "fish": "fish",
    "fizz": "fizzes",
    "flabellum": "flabella",
    "flag staff": "flag staffs",
    "flag-staff": "flag-staffs",
    "flagellum": "flagella",
    "flamenco": "flamencos",
    "flatfoot": "flatfoots",
    "flatus": "flatuses",
    "flocculus": "flocculi",
    "floccus": "flocci",
    "flora": "floras",
    "flounder": "flounder",
    "fly": "flies",
    "focus": "focuses",
    "follis": "folles",
    "fomes": "fomites",
    "foochowese": "foochowese",
    "foot": "feet",
    "foramen": "foramens",
    "formula": "formulas",
    "forum": "forums",
    "fossa": "fossae",
    "fowl": "fowls",
    "fraenum": "fraena",
    "franc archer": "franc archers",
    "franc-archer": "franc-archers",
    "frate": "frati",
    "frenum": "frena",
    "fricandeau": "fricandeaux",
    "frizz": "frizzes",
    "fro": "froes",
    "froe": "froes",
    "frottola": "frottole",
    "frustulum": "frustula",
    "frustum": "frustums",
    "fucus": "fuci",
    "fulcrum": "fulcra",
    "funambulus": "funambuli",
    "fundamentum divisionis": "fundamenta divisionis",
    "fungo porcino": "funghi porchini",
    "fungus": "funguses",
    "furfur": "furfures",
    "furioso": "furiosos",
    "fuzz": "fuzzes",
    "gabunese": "gabunese",
    "galla": "gallas",
    "gallows": "gallows",
    "gametangium": "gametangia",
    "ganglion": "ganglions",
    "gaon": "gaons",
    "gas": "gases",
    "gastrocnemius": "gastrocnemii",
    "geisha": "geishas",
    "geisteswissenschaft": "geisteswissenschaften",
    "gemma": "gemmae",
    "gena": "genae",
    "generalissimo": "generalissimos",
    "genevese": "genevese",
    "genie": "genies",
    "genius": "geniuses",
    "genoese": "genoese",
    "gens": "gentes",
    "genus": "genera",
    "german": "germans",
    "germarium": "germaria",
    "gestapo": "gestapos",
    "ghetto": "ghettos",
    "gigolo": "gigolos",
    "gilbertese": "gilbertese",
    "giuncus": "giunchi",
    "gizmo": "gizmos",
    "glacis": "glacises",
    "gladiolus": "gladioluses",
    "glass": "glasses",
    "glasses": "glasses",
    "glioma": "gliomata",
    "gliosarcoma": "gliosarcomata",
    "glissando": "glissandos",
    "globigerina": "globigerinae",
    "glochidium": "glochidia",
    "glomus": "glomera",
    "glottis": "glottises",
    "glutaeus": "glutei",
    "gluteus": "glutei",
    "gmina": "gminy",
    "go": "goes",
    "goanese": "goanese",
    "gonangium": "gonangia",
    "gonidium": "gonidia",
    "gonimium": "gonimia",
    "goose": "geese",
    "gorgonia": "gorgonias",
    "gousblom": "gousblomme",
    "goy": "goys",
    "graben": "grabens",
    "graffiti": "graffiti",
    "graffito": "graffiti",
    "granita": "granite",
    "grano": "grani",
    "granula": "granulae",
    "gravamen": "gravamina",
    "greensboro": "greensboros",
    "gringo": "gringos",
    "grissino": "grissini",
    "grotto": "grottos",
    "grouse": "grouses",
    "guaiabero": "guaiaberos",
    "guano": "guanos",
    "guardia civil": "guardias civiles",
    "gubernaculum": "gubernacula",
    "guglio": "guglios",
    "guianese": "guianese",
    "gumbo": "gumbos",
    "gumma": "gummas",
    "guttus": "gutti",
    "guy": "guys",
    "gymnasium": "gymnasiums",
    "gymnotus": "gymnoti",
    "gyro": "gyros",
    "gyrus": "gyri",
    "haddock": "haddocks",
    "haemangioma": "haemangiomas",
    "haggis": "haggis",
    "hainanese": "hainanese",
    "hairdo": "hairdos",
    "hake": "hakes",
    "halibut": "halibuts",
    "hallux": "halluces",
    "halteridium": "halteridia",
    "hamadryad": "hamadryads",
    "hamulus": "hamuli",
    "hapax legomenon": "hapax legomenons",
    "hapteron": "haptera",
    "hararese": "hararese",
    "harlemese": "harlemese",
    "harman": "harmans",
    "haruspex": "haruspices",
    "haustellum": "haustella",
    "haustorium": "haustoria",
    "haustrum": "haustra",
    "havanese": "havanese",
    "he": "they",
    "headquarters": "headquarters",
    "heavenese": "heavenese",
    "hedron": "hedrons",
    "heft": "hefte",
    "helion": "helions",
    "helix": "helices",
    "hemangeioma": "hemangeiomas",
    "hemangioma": "hemangiomas",
    "hemelytrum": "hemelytra",
    "hemera": "hemerae",
    "hemi elytrum": "hemi elytra",
    "hemi-elytrum": "hemi-elytra",
    "her": "them",
    "hernia": "hernias",
    "hero": "heroes",
    "heroon": "heroa",
    "heroum": "heroa",
    "herpes": "herpes",
    "herr": "herren",
    "herring": "herrings",
    "hers": "theirs",
    "herself": "themselves",
    "hertz": "hertz",
    "hetaera": "hetaerae",
    "hetaira": "hetaerae",
    "hetman": "hetmans",
    "hiatus": "hiatuses",
    "hijinks": "hijinks",
    "him": "them",
    "himself": "themselves",
    "hinaki": "hinaki",
    "hippo": "hippos",
    "hippocampus": "hippocampi",
    "hippopotamus": "hippopotamuses",
    "hiroshiman": "hiroshimans",
    "his": "theirs",
    "ho": "hoes",
    "hoe": "hoes",
    "holothuria": "holothurias",
    "honorarium": "honorariums",
    "hoof": "hoofs",
    "hoosierese": "hoosierese",
    "hordeolum": "hordeola",
    "hormogonium": "hormogonia",
    "hoti": "hoties",
    "hottentotese": "hottentotese",
    "hsien": "heen",
    "hubris": "hubrises",
    "human": "humans",
    "humerus": "humeri",
    "hunanese": "hunanese",
    "hydra": "hydras",
    "hydraulicon": "hydraulica",
    "hydria": "hydriae",
    "hymenium": "hymenia",
    "hymnarium": "hymnaria",
    "hyperbaton": "hyperbata",
    "hyperbola": "hyperbolas",
    "hypha": "hyphae",
    "hypnum": "hypnums",
    "hypoderma": "hypodermata",
    "hypolimnion": "hypolimnia",
    "i": "we",
    "ibis": "ibises",
    "ichthyosaurus": "ichthyosauri",
    "idaho": "idahos",
    "idee fixe": "idees fixes",
    "idolon": "idola",
    "idolum": "idola",
    "ignoramus": "ignoramuses",
    "ilium": "ilia",
    "imago": "imagos",
    "imbrex": "imbrices",
    "impetigo": "impetigos",
    "impetus": "impetuses",
    "improperium": "improperia",
    "improvisatrice": "improvisatrici",
    "incubus": "incubuses",
    "index": "indexes",
    "indicium": "indicia",
    "indigena": "indigenae",
    "individuum": "individuums",
    "indochinese": "indochinese",
    "indumentum": "indumenta",
    "indusium": "indusia",
    "inferno": "infernos",
    "infima species": "infimae species",
    "info": "infos",
    "inhalatorium": "inhalatoria",
    "innings": "innings",
    "inoculum": "inocula",
    "insula": "insulae",
    "intaglio": "intaglios",
    "interambulacrum": "interambulacra",
    "interaxis": "interaxes",
    "intercalarium": "intercalaria",
    "interleaf": "leaves",
    "intermaxilla": "intermaxillae",
    "intermedium": "intermedia",
    "intermezzo": "intermezzos",
    "internodium": "internodia",
    "interphenomenon": "interphenomena",
    "interradius": "interradii",
    "interregnum": "interregnums",
    "interrex": "interreges",
    "interstitium": "interstitiums",
    "intertrigo": "intertrigos",
    "intervallum": "intervallums",
    "intichiuma": "intichiuma",
    "inuk": "inuit",
    "inukshuk": "inukshuit",
    "involucrum": "involucra",
    "ipse dixit": "ipse dixits",
    "iquico": "iquicos",
    "iris": "irises",
    "ischium": "ischia",
    "isidium": "isidia",
    "ispravnik": "ispravniks",
    "isthmus": "isthmuses",
    "it": "they",
    "iter": "iters",
    "its": "theirs",
    "itself": "themselves",
    "ixodid": "ixodides",
    "jackanapes": "jackanapes",
    "jacobus": "jacobuses",
    "jalapeno": "jalapenos",
    "jambeau": "jambeaus",
    "japanese": "japanese",
    "javanese": "javanese",
    "jazz": "jazzes",
    "jerry": "jerries",
    "jeu": "jeux",
    "jo": "joes",
    "joe": "joes",
    "johnsonese": "johnsonese",
    "judicatum": "judicata",
    "juelet": "iueles",
    "jugum": "juga",
    "jumbo": "jumbos",
    "junto": "juntos",
    "kakapo": "kakapos",
    "kanarese": "kanarese",
    "kantharos": "kantharoi",
    "kar": "kars",
    "karakia": "karakias",
    "karrenfeld": "karrenfelds",
    "kernos": "kernoi",
    "kibbutz": "kibbutzes",
    "kilo": "kilos",
    "kimono": "kimonos",
    "kinkimavo": "kinkimavos",
    "kioko": "kiokos",
    "kiplingese": "kiplingese",
    "kletterschuh": "kletterschuhe",
    "knife": "knives",
    "knopper": "knoppers",
    "kobo": "kobo",
    "kokako": "kokakos",
    "konditorei": "konditorei",
    "kongoese": "kongoese",
    "kongolese": "kongolese",
    "kosovo": "kosovos",
    "kouros": "kouroi",
    "kovsh": "kovshi",
    "kuvasz": "kuvaszok",
    "kylix": "kylixes",
    "labrum": "labra",
    "lacinia": "laciniae",
    "lactobacillus": "lactobacilli",
    "lacuna": "lacunas",
    "lacunar": "lacunars",
    "laika": "laiki",
    "lamella": "lamellae",
    "lamina": "laminae",
    "lapith": "lapiths",
    "lapponese": "lapponese",
    "lar": "lars",
    "larnax": "larnakes",
    "larynx": "larynges",
    "latex": "latexes",
    "leaf": "leaves",
    "lebanese": "lebanese",
    "lebensform": "lebensformen",
    "legionella": "legionellae",
    "legomenon": "legomena",
    "legumen": "legumens",
    "leishmania": "leishmanias",
    "leman": "lemans",
    "lemma": "lemmas",
    "lemniscus": "lemnisci",
    "lemur": "lemurs",
    "lens": "lenses",
    "lentigo": "lentigines",
    "lentiscus": "lentisci",
    "leonese": "leonese",
    "leproma": "lepromas",
    "leprosarium": "leprosaria",
    "leptomonas": "leptomonas",
    "lepton": "lepta",
    "leptospira": "leptospirae",
    "lesotho": "lesothos",
    "leu": "lei",
    "lex humana": "leges humanae",
    "libero": "liberos",
    "libido": "libidos",
    "libretto": "librettos",
    "lido": "lidos",
    "lied": "lieder",
    "lieu": "lieus",
    "lieutenant general": "lieutenant generals",
    "life": "lives",
    "ligamentum": "ligamenta",
    "likuta": "makuta",
    "lilo": "lilos",
    "liman": "limans",
    "limax": "limaces",
    "limbo": "limbos",
    "limes": "limites",
    "limo": "limos",
    "limulus": "limuli",
    "linctus": "linctuses",
    "line out": "line outs",
    "line-out": "line-outs",
    "linea": "lineae",
    "lineno": "linenos",
    "lingo": "lingos",
    "lingula": "lingulae",
    "lino": "linos",
    "lipoma": "lipomata",
    "lis": "lisses",
    "listerella": "listerellas",
    "listeria": "listerias",
    "lithodomus": "lithodomi",
    "lithophyton": "lithophyta",
    "livedo": "livedos",
    "lixivium": "lixivia",
    "loa": "loas",
    "loaf": "loaves",
    "lobulus": "lobuli",
    "lochus": "lochi",
    "loco": "locos",
    "loculus": "loculi",
    "locus": "loci",
    "loggia": "loggias",
    "logion": "logia",
    "logo": "logos",
    "logodaedalus": "logodaedali",
    "loma": "lomata",
    "lomentum": "lomenta",
    "londonese": "londonese",
    "lord lieutenant": "lord lieutenants",
    "lord-lieutenant": "lord-lieutenants",
    "lorrainese": "lorrainese",
    "lorum": "lora",
    "louis": "louis",
    "louse": "lice",
    "lowlife": "lowlifes",
    "lox": "loxes",
    "lucchese": "lucchese",
    "lumbago": "lumbagos",
    "lumbricus": "lumbrici",
    "lumen": "lumens",
    "lustrum": "lustrums",
    "lymphangioma": "lymphangiomata",
    "lymphoma": "lymphomas",
    "lyonese": "lyonese",
    "lyssa": "lyssae",
    "maar": "maars",
    "macacus": "macaci",
    "macanese": "macanese",
    "macaroni": "macaronies",
    "macassarese": "macassarese",
    "macho": "machos",
    "mackerel": "mackerels",
    "macro": "macros",
    "macula": "maculas",
    "madurese": "madurese",
    "maecenas": "maecenases",
    "maestro": "maestros",
    "mafiosa": "mafiose",
    "mafioso": "mafiosos",
    "magilus": "magili",
    "magma": "magmas",
    "magneto": "magnetos",
    "magnifico": "magnificos",
    "magnum opus": "magnum opuses",
    "magus": "magi",
    "major general": "major generals",
    "majuro": "majuros",
    "malabarese": "malabarese",
    "malabo": "malabos",
    "malagma": "malagmata",
    "malleolus": "malleoli",
    "malleus": "mallei",
    "maltese": "maltese",
    "malum in se": "mala in se",
    "man": "men",
    "man child": "men children",
    "man midwife": "men midwives",
    "man milliner": "man milliners",
    "man servant": "men servants",
    "man-child": "men-children",
    "man-midwife": "men-midwives",
    "man-milliner": "man-milliners",
    "man-servant": "men-servants",
    "manchild": "menchildren",
    "mango": "mangos",
    "manifesto": "manifestos",
    "manservant": "menservants",
    "manteau": "manteaus",
    "mantelletta": "mantellette",
    "mantis": "mantises",
    "manubrium": "manubriums",
    "maputo": "maputos",
    "maquereau": "maquereaux",
    "maracaibo": "maracaibos",
    "mare": "mares",
    "maremma": "maremme",
    "marquis": "marquises",
    "mary": "maries",
    "mater lectionis": "matres lectionis",
    "math": "maths",
    "matin": "matins",
    "matinee musicale": "matinees musicales",
    "matzah": "matzahs",
    "mausoleum": "mausoleums",
    "maxilla": "maxillae",
    "maximum": "maximums",
    "me": "us",
    "measles": "measles",
    "meatus": "meatuses",
    "mediastinum": "mediastina",
    "mediatrix": "mediatrices",
    "medico": "medicos",
    "medium": "mediums",
    "medulloblastoma": "medulloblastomas",
    "medusa": "medusas",
    "megalosaurus": "megalosauri",
    "megatherium": "megatheria",
    "melanoma": "melanomas",
    "melastoma": "melastomas",
    "meliceris": "melicerides",
    "melisma": "melismas",
    "memento": "mementos",
    "memo": "memos",
    "memorandum": "memorandums",
    "menage a quatre": "menages a quatre",
    "menage a trois": "menages a trois",
    "meningioma": "meningiomas",
    "meniscus": "menisci",
    "menu": "menus",
    "meretrix": "meretrices",
    "meris": "merides",
    "meshumad": "meshumadim",
    "meshummad": "meshummadim",
    "mesothelioma": "mesotheliomas",
    "mesothelium": "mesothelia",
    "messa di voce": "messe di voce",
    "messinese": "messinese",
    "meta": "metae",
    "metalimnion": "metalimnia",
    "metapleuron": "metapleura",
    "metatarsus": "metatarsi",
    "metro": "metros",
    "metropolis": "metropolises",
    "mew": "mews",
    "mews": "mews",
    "mexico": "mexicos",
    "mezuza": "mezuzoth",
    "mezuzah": "mezuzoth",
    "mezzo rilievo": "mezzo rilievos",
    "mezzo-rilievo": "mezzo-rilievos",
    "miasma": "miasmas",
    "micro": "micros",
    "microbion": "microbia",
    "microcephalus": "microcephali",
    "micrococcus": "micrococci",
    "microfiche": "microfiches",
    "micronucleus": "micronuclei",
    "milanese": "milanese",
    "milano": "milanos",
    "miles gloriosus": "milites gloriosi",
    "milieu": "milieus",
    "milieu interieur": "milieux interieurs",
    "miliola": "miliolae",
    "millennium": "millenniums",
    "mimosa": "mimosas",
    "mine": "ours",
    "minimum": "minimums",
    "minnelied": "minnelieder",
    "minutia": "minutias",
    "minyan": "minyanim",
    "miracidium": "miracidia",
    "mithraeum": "mithraea",
    "mitochondrion": "mitochondria",
    "mixolimnion": "mixolimnia",
    "mo": "moes",
    "moaks": "moakses",
    "modelletto": "modelletti",
    "modello": "modellos",
    "modenese": "modenese",
    "modernus": "moderni",
    "modius": "modii",
    "modulus": "moduluses",
    "modus": "moduses",
    "moe": "moes",
    "molimen": "molimina",
    "momentum": "momentums",
    "monaco": "monacos",
    "monas": "monades",
    "mondo": "mondi",
    "money": "monies",
    "mongoose": "mongooses",
    "monimolimnion": "monimolimnia",
    "mono": "monos",
    "monochasium": "monochasia",
    "monorchis": "monorchides",
    "monosyllabon": "monosyllaba",
    "monota": "monotae",
    "monoxylon": "monoxyla",
    "mons": "montes",
    "monseigneur": "monsiegneurs",
    "monsignor": "monsignors",
    "monsignore": "monsignori",
    "monstre sacre": "monstres sacres",
    "montenegro": "montenegros",
    "moose": "moose",
    "moratorium": "moratoriums",
    "morceau": "morceaux",
    "morocco": "moroccos",
    "moroso": "morosi",
    "mortadella": "mortadellas",
    "mortarium": "mortaria",
    "moscardino": "moscardini",
    "motoscafo": "motoscafi",
    "motto": "mottos",
    "mouse": "mice",
    "mousetrap": "mousetraps",
    "mouth": "mouths",
    "mucro": "mucros",
    "mukim": "mukims",
    "multigravida": "multigravidas",
    "mumps": "mumps",
    "municipium": "municipia",
    "muqdisho": "muqdishos",
    "muranese": "muranese",
    "murex": "murexes",
    "musa": "musas",
    "musee imaginaire": "musees imaginaires",
    "musique concrete": "musiques concretes",
    "mustachio": "mustachios",
    "muzz": "muzzes",
    "mya": "myas",
    "mycoplasma": "mycoplasmas",
    "mycorhiza": "mycorhizas",
    "mycorrhiza": "mycorrhizas",
    "myeloma": "myelomas",
    "myo": "myos",
    "myself": "ourselves",
    "mythos": "mythoi",
    "myxoma": "myxomata",
    "nachtlokal": "nachtlokale",
    "nagaika": "nagaikas",
    "nakayaman": "nakayamans",
    "nankingese": "nankingese",
    "nasutus": "nasuti",
    "naumachia": "naumachias",
    "nauplius": "nauplii",
    "nautilus": "nautiluses",
    "navarrese": "navarrese",
    "naya paisa": "naye paise",
    "nebenkern": "nebenkerne",
    "nebula": "nebulas",
    "nectocalyx": "nectocalyces",
    "nemathecium": "nemathecia",
    "nenets": "nentsi",
    "nepalese": "nepalese",
    "nephridium": "nephridia",
    "nerita": "neritas",
    "neuma": "neumas",
    "neurinoma": "neurinomas",
    "neuroma": "neuromata",
    "neurula": "neurulae",
    "neutrino": "neutrinos",
    "news": "news",
    "nexus": "nexuses",
    "ngwee": "ngwee",
    "niasese": "niasese",
    "nicobarese": "nicobarese",
    "nidus": "niduses",
    "niello": "niellos",
    "nimbus": "nimbuses",
    "ningbo": "ningbos",
    "nipponese": "nipponese",
    "nockerl": "nockerln",
    "noctiluca": "noctilucae",
    "nodum": "noda",
    "nodus": "nodi",
    "noema": "noemata",
    "nomisma": "nomismata",
    "non-ens": "nonentia",
    "norman": "normans",
    "noumenon": "noumena",
    "nouveau pauvre": "nouvelle pauvre",
    "nouveau riche": "nouvelle riche",
    "nouvel arrive": "nouveaux arrives",
    "nova": "novas",
    "novus homo": "novi homines",
    "noxa": "noxae",
    "nubecula": "nubeculae",
    "nucleolonema": "nucleolonemas",
    "nucleolus": "nucleoluses",
    "nucleus": "nucleuses",
    "numen": "numina",
    "nuntius": "nuntii",
    "nympha": "nymphae",
    "nymphaeum": "nymphea",
    "nympheum": "nymphea",
    "oath": "oaths",
    "obelus": "obeli",
    "obolus": "oboli",
    "occiput": "occiputs",
    "oceanarium": "oceanaria",
    "ocellus": "ocelli",
    "octavo": "octavos",
    "octohedron": "octohedrons",
    "octopus": "octopuses",
    "oculus": "oculi",
    "odeum": "odea",
    "oedema": "oedemas",
    "offspring": "offspring",
    "ogle": "oglys",
    "oklahoman": "oklahomans",
    "olfactorium": "olfactoria",
    "omentum": "omenta",
    "ommatidium": "ommatidia",
    "onager": "onagers",
    "one": "some",
    "operculum": "opercula",
    "ophiophagus": "ophiophagi",
    "optimum": "optimums",
    "opus": "opuses",
    "opus magnum": "opera magna",
    "opusculum": "opuscula",
    "oratrix": "oratrices",
    "oregano": "oreganos",
    "organon": "organa",
    "orinoco": "orinocos",
    "orlando": "orlandos",
    "orthoceras": "orthocerata",
    "osculum": "oscula",
    "oslo": "oslos",
    "osmaterium": "osmateria",
    "osmeterium": "osmeteria",
    "ossiculum": "ossicula",
    "osteoclastoma": "osteoclastomas",
    "osteoma": "osteomata",
    "ostium": "ostia",
    "oto": "otos",
    "otomi": "otomies",
    "ottoman": "ottomans",
    "ovarium": "ovaria",
    "ovolo": "ovoli",
    "ovulum": "ovula",
    "ovum": "ova",
    "ox": "oxen",
    "oxybaphon": "oxybapha",
    "oxymoron": "oxymorons",
    "oxyuris": "oxyurides",
    "paha": "paha",
    "paisa": "paisas",
    "pajamas": "pajamas",
    "pala": "palae",
    "palea": "paleae",
    "pallium": "pallia",
    "palpebra": "palpebrae",
    "palpus": "palpi",
    "palus": "pali",
    "panaman": "panamans",
    "pangasinan": "pangasinans",
    "panino": "panini",
    "panorpa": "panorpae",
    "pant": "pants",
    "panto": "pantos",
    "pants": "pants",
    "paparrazo": "paparrazi",
    "papilla": "papillae",
    "papilloma": "papillomata",
    "papula": "papulae",
    "papyrus": "papyruses",
    "parabola": "parabolas",
    "parabolanus": "parabolani",
    "parador": "paradors",
    "paraganglioma": "paragangliomas",
    "paraganglion": "paraganglia",
    "paraglossa": "paraglossae",
    "paramaribo": "paramaribos",
    "paranucleus": "paranuclei",
    "parapleura": "parapleura",
    "parapleurum": "parapleura",
    "parapsis": "parapsides",
    "paraselene": "paraselenae",
    "parc ferme": "parcs fermes",
    "pardusco": "parduscos",
    "parenchyma": "parenchymata",
    "parergon": "parerga",
    "parhelion": "parhelions",
    "paries": "parietes",
    "parison": "parisa",
    "parmese": "parmese",
    "parnas": "parnassim",
    "parnass": "parnassim",
    "parvenu": "parvenus",
    "pasteurella": "pasteurellas",
    "pastorale": "pastorales",
    "patagium": "patagia",
    "patera": "paterae",
    "path": "paths",
    "pathos": "pathoses",
    "paxilla": "paxillae",
    "pecia": "pecias",
    "pecten": "pectens",
    "pectus": "pectora",
    "pedalo": "pedalos",
    "pedicellaria": "pedicellariae",
    "pediluvium": "pediluvia",
    "pedion": "pedia",
    "pekingese": "pekingese",
    "pelta": "peltae",
    "pelvis": "pelvises",
    "pence": "pence",
    "pendulum": "pendulums",
    "pengo": "pengo",
    "penicillus": "penicilli",
    "peninsula": "peninsulas",
    "penis": "penises",
    "penni": "pennia",
    "penny": "pennies",
    "pensiero": "pensieri",
    "perceptum": "percepta",
    "peridium": "peridia",
    "perikaryon": "perikarya",
    "perithecium": "perithecia",
    "pernio": "perniones",
    "peronium": "peronia",
    "perradius": "perradii",
    "person": "people",
    "persona": "personas",
    "pes": "pedes",
    "phalangium": "phalangia",
    "phalanx": "phalanxes",
    "phallos": "phalloi",
    "phallus": "phalluses",
    "pharmakos": "pharmakoi",
    "phenomenon": "phenomena",
    "phlegmasia": "phlegmasiae",
    "phoca": "phocas",
    "pholas": "pholades",
    "photo": "photos",
    "phototaxis": "phototaxes",
    "phragma": "phragmata",
    "phren": "phrenes",
    "phyle": "phylae",
    "phyllobranchia": "phyllobranchiae",
    "phylum": "phylums",
    "phyma": "phymata",
    "phyogemmarium": "phyogemmaria",
    "piano": "pianos",
    "pickerel": "pickerels",
    "pied a terre": "pieds a terre",
    "pied noir": "pieds noirs",
    "piedmontese": "piedmontese",
    "pietra": "pietre",
    "pifferaro": "pifferari",
    "pike": "pikes",
    "pileus": "pilei",
    "pillie": "pilleis",
    "pilus": "pili",
    "pimento": "pimentos",
    "pina colada": "pina coladas",
    "pinax": "pinaces",
    "pincer": "pincers",
    "pincers": "pincers",
    "pinetum": "pinetums",
    "pingo": "pingos",
    "pinna": "pinnas",
    "pinon": "pinons",
    "pinto": "pintos",
    "pipkrake": "pipkrakes",
    "piscina": "piscinas",
    "pistillidium": "pistillidia",
    "pistoiese": "pistoiese",
    "pithos": "pithoi",
    "pizza": "pizzas",
    "placebo": "placebos",
    "placitum": "placita",
    "planctus": "planctus",
    "plankter": "plankton",
    "planula": "planulae",
    "plasmalemma": "plasmalemmas",
    "plasmodium": "plasmodia",
    "plasmolyticum": "plasmolytica",
    "plasmoma": "plasmomata",
    "plateau": "plateaus",
    "platypus": "platypuses",
    "pleco": "plecos",
    "plectrum": "plectra",
    "plenarium": "plenaria",
    "plesiosaurus": "plesiosauri",
    "plethron": "plethra",
    "pleura": "pleura",
    "pleuron": "pleura",
    "pleurum": "pleura",
    "plexus": "plexuses",
    "plica": "plicae",
    "pliers": "pliers",
    "ploshchadka": "ploshchadki",
    "plurale tantum": "pluralia tantum",
    "pluteus": "plutei",
    "pluto": "plutos",
    "pneumothorax": "pneumothoraxes",
    "po": "poes",
    "poblador": "pobladores",
    "poculum": "pocula",
    "podere": "poderi",
    "podetium": "podetia",
    "podion": "podia",
    "podium": "podiums",
    "poe": "poes",
    "pogo": "pogos",
    "poise": "poises",
    "polis": "polises",
    "politbureau": "politbureaus",
    "politico": "politicos",
    "pollex": "pollices",
    "pollinarium": "pollinaria",
    "pollinium": "pollinia",
    "polo": "polos",
    "polonese": "polonese",
    "polythecium": "polythecia",
    "pomme de terre": "pommes de terre",
    "poncho": "ponchos",
    "pontifex": "pontifexes",
    "porcino": "porcini",
    "porion": "poria",
    "porron": "porrons",
    "port a beul": "puirt a beul",
    "port de bras": "ports de bras",
    "port-a-beul": "puirt-a-beul",
    "portico": "porticos",
    "portiforium": "portiforia",
    "portmanteau": "portmanteaus",
    "porto": "portos",
    "porto novo": "porto novos",
    "porto-novo": "porto-novos",
    "portrait parle": "portraits parles",
    "portuguese": "portuguese",
    "postnatus": "postnati",
    "postpubis": "postpubes",
    "postulatum": "postulata",
    "potato": "potatoes",
    "potestas": "potestates",
    "pox": "pox",
    "praecognitum": "praecognita",
    "pragma": "pragmas",
    "precis": "precis",
    "precognitum": "precognita",
    "premier danseur": "premiers danseurs",
    "prepenna": "prepennae",
    "prima ballerina": "prima ballerinas",
    "prima donna": "prima donnas",
    "primigravida": "primigravidae",
    "primordium": "primordia",
    "principium": "principia",
    "pro": "pros",
    "proavis": "proaves",
    "proboscis": "proboscises",
    "proceedings": "proceedings",
    "procoelia": "procoeliae",
    "prodromus": "prodromi",
    "proferens": "proferentes",
    "profluvium": "profluvia",
    "proglottis": "proglottides",
    "programma": "programmata",
    "progymnasium": "progymnasia",
    "progymnasma": "progymnasmata",
    "prokaryon": "prokarya",
    "prolabium": "prolabia",
    "prolegomenon": "prolegomena",
    "promerops": "promeropes",
    "propleuron": "propleura",
    "propodium": "propodia",
    "propositus": "propositi",
    "proprium": "propria",
    "propylaeum": "propylaea",
    "propylon": "propylons",
    "prora": "prorae",
    "prosa": "prosae",
    "proscenium": "proscenia",
    "proscolla": "proscollae",
    "prosecutrix": "prosecutrices",
    "prosencephalon": "prosencephala",
    "proseucha": "proseuchae",
    "prosodiencephalon": "prosodiencephala",
    "prosopium": "prosopia",
    "prospectus": "prospectuses",
    "prosula": "prosulae",
    "protarsus": "protarsi",
    "prothallium": "prothallia",
    "prothallus": "prothalli",
    "protococcus": "protococci",
    "protoma": "protomas",
    "protonema": "protonemata",
    "prototypon": "prototypons",
    "protypon": "protypa",
    "proviso": "provisos",
    "prytanis": "prytanes",
    "psammoma": "psammomata",
    "pseudopodium": "pseudopodia",
    "pseudostoma": "pseudostomata",
    "psycho": "psychos",
    "psydracium": "psydracia",
    "pteroma": "pteromata",
    "pteropus": "pteropi",
    "pteryla": "pterylae",
    "pueblo": "pueblos",
    "pul": "puls",
    "pullus": "pulli",
    "pulmonarium": "pulmonaria",
    "pulsellum": "pulsella",
    "pulveratrix": "pulveratrices",
    "punctum": "puncta",
    "punto blanco": "puntos blancos",
    "pupa": "pupae",
    "purlieu": "purlieus",
    "purum": "purums",
    "putto": "putti",
    "puttony": "puttonys",
    "pycnidium": "pycnidia",
    "pycnis": "pycnides",
    "pycnium": "pycnia",
    "pyjamas": "pyjamas",
    "pyla": "pylae",
    "pyralis": "pyralides",
    "pyramidion": "pyramidions",
    "pyrexia": "pyrexiae",
    "pyrites": "pyritae",
    "pyrophorus": "pyrophori",
    "pyxidium": "pyxidia",
    "quadratrix": "quadratrices",
    "quaesitum": "quaesita",
    "quantum": "quantums",
    "quarto": "quartos",
    "quartz": "quartzes",
    "quid": "quid",
    "quinarius": "quinarii",
    "quinquennium": "quinquennia",
    "quito": "quitos",
    "rabies": "rabies",
    "rad": "rads",
    "radiale": "radialia",
    "radius": "radiuses",
    "radix": "radixes",
    "ragazza": "ragazze",
    "ragazzo": "ragazzi",
    "raggare": "raggares",
    "ramulus": "ramuli",
    "ramus": "rami",
    "ranunculus": "ranunculuses",
    "rara avis": "rara avises",
    "razz": "razzes",
    "rectus": "recti",
    "redia": "rediae",
    "reductio": "reductiones",
    "referendum": "referendums",
    "refugium": "refugia",
    "regidor": "regidors",
    "regma": "regmata",
    "regulus": "reguli",
    "reich": "reiche",
    "rejon": "rejones",
    "rejoneador": "rejoneadores",
    "relatum": "relata",
    "rem": "rems",
    "remex": "remexes",
    "remous": "remous",
    "replum": "repla",
    "representamen": "representamens",
    "res": "res",
    "residuum": "residuums",
    "responsum": "responsa",
    "rete": "retia",
    "retinaculum": "retinacula",
    "retinula": "retinulae",
    "rezident": "rezidenty",
    "rezidentsia": "rezidentsii",
    "rhino": "rhinos",
    "rhinoceros": "rhinoceroses",
    "rhinophyma": "rhinophymas",
    "rhizoma": "rhizomas",
    "rhododendron": "rhododendrons",
    "rhombohedron": "rhombohedrons",
    "rhomboideus": "rhomboidei",
    "rhopalium": "rhopalia",
    "rhynchodaeum": "rhynchodaea",
    "rhythmus": "rhythmi",
    "ri": "ri",
    "ricercar": "ricercars",
    "ricercata": "ricercate",
    "ricordo": "ricordi",
    "rideau": "rideaus",
    "rilievo": "rilievos",
    "rimon": "rimonim",
    "rione": "rioni",
    "risotto": "risottos",
    "ritardando": "ritardandos",
    "rite de passage": "rites de passage",
    "rococo": "rococos",
    "roe": "roes",
    "rom": "roma",
    "romagnese": "romagnese",
    "roman": "romans",
    "romanese": "romanese",
    "romany": "romanies",
    "romanze": "romanzen",
    "rond de jambe": "rond de jambes",
    "rondeau": "rondeaus",
    "rondo": "rondos",
    "rostrum": "rostrums",
    "rotula": "rotulae",
    "route nationale": "routes nationales",
    "rubashka": "rubashkas",
    "ruga": "rugae",
    "ruse de guerre": "ruses de guerre",
    "rype": "ryper",
    "sabertooth": "sabertooths",
    "sacculus": "sacculi",
    "sacellum": "sacella",
    "sachverhalt": "sachverhalte",
    "sacramento": "sacramentos",
    "sacrarium": "sacraria",
    "sacrum": "sacrums",
    "saddo": "saddos",
    "saffo": "saffi",
    "sago": "sagos",
    "sagum": "saga",
    "salita": "salitas",
    "salivarium": "salivaria",
    "salmon": "salmon",
    "salpa": "salpas",
    "salvo": "salvos",
    "sambo": "sambos",
    "samurai": "samurai",
    "sanatorium": "sanatoriums",
    "sanctum sanctorum": "sancta sanctorum",
    "sandr": "sandrs",
    "sangirese": "sangirese",
    "santiago": "santiagos",
    "sapporo": "sapporos",
    "sarajevo": "sarajevos",
    "sarawakese": "sarawakese",
    "sarcina": "sarcinae",
    "sarcoma": "sarcomas",
    "sarcophagus": "sarcophagi",
    "sarissa": "sarissae",
    "sassafras": "sassafrases",
    "satelles": "satellites",
    "saxicava": "saxicavae",
    "sbirro": "sbirri",
    "scalenus": "scaleni",
    "scandalum magnatum": "scandala magnatum",
    "scapula": "scapulas",
    "scapus": "scapi",
    "scarabaeus": "scarabaei",
    "scarus": "scari",
    "scazon": "scazontes",
    "sceat": "sceattas",
    "schema": "schemas",
    "scherzando": "scherzandos",
    "scherzo": "scherzos",
    "schisma": "schismata",
    "schistosomulum": "schistosomula",
    "schizostylis": "schizostylis",
    "scholium": "scholia",
    "sciolus": "scioli",
    "scirrhus": "scirrhusses",
    "scissors": "scissors",
    "scolex": "scoleces",
    "scolopidium": "scolopidia",
    "scomber": "scombri",
    "scoria": "scorias",
    "scotoma": "scotomas",
    "scriptorium": "scriptoriums",
    "scudo": "scudi",
    "scungille": "scungilli",
    "scuola": "scuole",
    "scutellum": "scutella",
    "scutulum": "scutula",
    "scutum": "scuta",
    "scyphus": "scyphi",
    "seaquarium": "seaquaria",
    "secans": "secantes",
    "secretum": "secreta",
    "secundagravida": "secundagravidas",
    "secundigravida": "secundigravidas",
    "secundipara": "secundiparas",
    "sedes": "sedes",
    "sedile": "sedilia",
    "seecatch": "seecatchie",
    "seed": "seeds",
    "selman": "selmans",
    "sema": "semas",
    "seminium": "seminia",
    "semuncia": "semunciae",
    "senarius": "senarii",
    "senegalese": "senegalese",
    "senex": "senes",
    "sensedatum": "sensedata",
    "sensillum": "sensilla",
    "sensoriolum": "sensoriola",
    "sensum": "sensa",
    "sententia": "sententiae",
    "separatrix": "separatrices",
    "sephardi": "sephardim",
    "septarium": "septaria",
    "septemvir": "septemviri",
    "septenarius": "septenarii",
    "septum": "septa",
    "sequela": "sequelae",
    "sequestrum": "sequestra",
    "seraph": "seraphs",
    "sericterium": "sericteria",
    "series": "series",
    "serir": "serirs",
    "serpigo": "serpigoes",
    "serpula": "serpulae",
    "serra": "serrae",
    "sertularia": "sertularias",
    "serum": "serums",
    "sestertium": "sestertia",
    "sestiere": "sestieri",
    "set off": "set offs",
    "set out": "set outs",
    "set to": "set tos",
    "set-off": "set-offs",
    "set-out": "set-outs",
    "set-to": "set-tos",
    "seta": "setae",
    "sezession": "sezessionen",
    "sforzato": "sforzatos",
    "sgraffiato": "sgraffiati",
    "shad": "shads",
    "shaman": "shamans",
    "shauri": "shauris",
    "shavese": "shavese",
    "shawabti": "shawabtis",
    "shawanese": "shawanese",
    "she": "they",
    "shear": "shears",
    "shears": "shears",
    "sheep": "sheep",
    "shelf": "shelves",
    "shifta": "shiftas",
    "shochet": "shochetim",
    "shomer": "shomrim",
    "short": "shorts",
    "shorts": "shorts",
    "siamese": "siamese",
    "siemens": "siemens",
    "sienese": "sienese",
    "siglos": "sigli",
    "signum": "signa",
    "sikkimese": "sikkimese",
    "silbador": "silbadors",
    "silenus": "sileni",
    "silex": "silices",
    "siliqua": "siliquae",
    "silo": "silos",
    "silurus": "siluri",
    "simia": "simiae",
    "simplex": "simplexes",
    "simpulum": "simpula",
    "simulacrum": "simulacrums",
    "sinfonia": "sinfonias",
    "singhalese": "singhalese",
    "sinhalese": "sinhalese",
    "sinus": "sinuses",
    "siphunculus": "siphunculi",
    "sipunculus": "sipunculi",
    "sirocco": "siroccos",
    "sistrum": "sistrums",
    "situla": "situlas",
    "sizz": "sizzes",
    "ski": "skis",
    "snekkja": "snekkjur",
    "snipe": "snipes",
    "snooze": "snoozes",
    "so": "soes",
    "soe": "soes",
    "sogdianese": "sogdianese",
    "sol": "soles",
    "solarium": "solaria",
    "soldo": "soldi",
    "solenium": "solenia",
    "solfeggio": "solfeggios",
    "solidus": "solidi",
    "solo": "solos",
    "solstitium": "solstitia",
    "solvus": "solvi",
    "soma": "somas",
    "somaten": "somatenes",
    "sombrero": "sombreros",
    "sonaman": "sonamans",
    "soprano": "sopranos",
    "soralium": "soralia",
    "soredium": "soredia",
    "sorus": "sori",
    "soshi": "soshi",
    "sotch": "sotches",
    "souvlaki": "souvlakis",
    "spadix": "spadixes",
    "spaetlese": "spaetlesen",
    "sparus": "spari",
    "spasmus": "spasmi",
    "spatha": "spathae",
    "species": "species",
    "spectrum": "spectrums",
    "speculum": "speculums",
    "sperma": "spermata",
    "spermarium": "spermaria",
    "spermatheca": "spermathecae",
    "spermatium": "spermatia",
    "spermatoon": "spermatoa",
    "spermogonium": "spermogonia",
    "sphaeridium": "sphaeridia",
    "sphagnum": "sphagnums",
    "sphex": "spheges",
    "sphinx": "sphinxes",
    "spiculum": "spicula",
    "spiraculum": "spiracula",
    "spirillum": "spirilla",
    "spondylus": "spondyli",
    "sporodochium": "sporodochia",
    "sputum": "sputa",
    "squadra": "squadre",
    "squadrist": "squadrists",
    "squalus": "squali",
    "squama": "squamae",
    "squilla": "squillae",
    "stabilimentum": "stabilimenta",
    "staccato": "staccatos",
    "stadium": "stadiums",
    "stalactite": "stalactites",
    "stamen": "stamen",
    "staminidium": "staminidia",
    "staphylococcus": "staphylococci",
    "star fish": "star fishes",
    "star-fish": "star-fishes",
    "starets": "startsy",
    "staretz": "startzy",
    "starfish": "starfishes",
    "starosta": "starosti",
    "starover": "starovers",
    "stasimon": "stasimons",
    "statera": "staterae",
    "statua": "statuas",
    "status": "statuses",
    "stela": "stelas",
    "stella": "stellae",
    "stemma": "stemmata",
    "stephanion": "stephanions",
    "sterigma": "sterigmata",
    "sternebra": "sternebrae",
    "sterno": "sternos",
    "sternum": "sternums",
    "stichidium": "stichidia",
    "stichos": "stichoi",
    "stigma": "stigmas",
    "stigmaria": "stigmariae",
    "stiletto": "stilettoes",
    "stillicidium": "stillicidia",
    "stimulus": "stimuli",
    "stipes": "stipites",
    "stipula": "stipulas",
    "stirps": "stirpes",
    "stoa": "stoas",
    "stolo": "stolones",
    "stoma": "stomas",
    "stomach": "stomachs",
    "stomatium": "stomatia",
    "stomion": "stomia",
    "stomodaeum": "stomodea",
    "storey": "storeys",
    "stornello": "stornelli",
    "strambotto": "strambotti",
    "strategos": "strategoi",
    "strategus": "strategi",
    "stratum": "stratums",
    "streptococcus": "streptococci",
    "streptothrix": "streptothrices",
    "stria": "striae",
    "striatum": "striata",
    "striga": "strigae",
    "striola": "striolae",
    "strobila": "strobilae",
    "strobilus": "strobili",
    "stroma": "stromata",
    "stromodeum": "stromodea",
    "struma": "strumae",
    "stryddag": "stryddae",
    "stucco": "stuccos",
    "studium generale": "studia generali",
    "stylo": "stylos",
    "stylopodium": "stylopodia",
    "stylus": "styluses",
    "subbotnik": "subbotniks",
    "subsellium": "subsellia",
    "subsidium": "subsidia",
    "substratum": "substratums",
    "succinea": "succineas",
    "succubus": "succubuses",
    "succus": "succi",
    "sudanese": "sudanese",
    "suffrutex": "suffrutices",
    "suggestio falsi": "suggestiones falsi",
    "suggestum": "suggestums",
    "sulcus": "sulci",
    "sum total": "sum totals",
    "sum-total": "sum-totals",
    "summa": "summae",
    "summum bonum": "summa bona",
    "summum genus": "summa genera",
    "sumo": "sumos",
    "sundanese": "sundanese",
    "supercilium": "supercilia",
    "superficies": "superficies",
    "supernova": "supernovas",
    "superstratum": "superstrata",
    "suppositum": "supposita",
    "supracleithrum": "supracleithra",
    "suprascapula": "suprascapulae",
    "surculus": "surculi",
    "surimono": "surimonos",
    "sustentaculum": "sustentacula",
    "swahilese": "swahilese",
    "swine": "swine",
    "switchfoot": "switchfoots",
    "syllabarium": "syllabaria",
    "syllabus": "syllabuses",
    "symplasma": "symplasmata",
    "sympodium": "sympodia",
    "symposium": "symposiums",
    "synaesthesia": "synaesthesiae",
    "synangium": "synangia",
    "synapticula": "synapticulas",
    "synaxarion": "synaxaria",
    "synaxarium": "synaxaria",
    "synaxis": "synaxes",
    "syncellus": "syncelli",
    "syncytium": "syncytia",
    "syndendrium": "syndendria",
    "synechia": "synechiae",
    "synedrion": "synedria",
    "synedrium": "synedria",
    "synesthesia": "synesthesiae",
    "syntagma": "syntagmas",
    "syntaxis": "syntaxes",
    "synthronus": "synthroni",
    "synusia": "synusiae",
    "syphiloma": "syphilomata",
    "syringe": "syringes",
    "syrinx": "syrinxes",
    "syrphus": "syrphi",
    "syrtis": "syrtes",
    "tabella": "tabellae",
    "tableau": "tableaus",
    "tablinum": "tablina",
    "tabula": "tabulae",
    "tacoman": "tacomans",
    "taenia": "taenias",
    "tagma": "tagmata",
    "taiko": "taikos",
    "talea": "taleae",
    "talisman": "talismans",
    "talouse": "talouses",
    "talus": "tali",
    "tango": "tangos",
    "tarantata": "tarantate",
    "tarantato": "tarantati",
    "tarsus": "tarsi",
    "tax": "taxes",
    "taxis": "taxes",
    "taxon": "taxa",
    "tazza": "tazze",
    "teal": "teals",
    "techno": "technos",
    "tegmen": "tegmina",
    "tegmentum": "tegmenta",
    "tegula": "tegulae",
    "telamon": "telamones",
    "telium": "telia",
    "temblor": "temblores",
    "tempietto": "tempietti",
    "tempo": "tempos",
    "tenaculum": "tenacula",
    "tenderfoot": "tenderfoots",
    "tenggerese": "tenggerese",
    "tenia": "tenias",
    "tenorino": "tenorini",
    "tentaculum": "tentacula",
    "tentamen": "tentamina",
    "tenuis": "tenues",
    "tepidarium": "tepidaria",
    "teratoma": "teratomata",
    "terebellum": "terebella",
    "terebratula": "terebratulas",
    "teredo": "teredos",
    "tergum": "terga",
    "termes": "termites",
    "terminus": "terminuses",
    "terp": "terps",
    "terrae filius": "terrae filii",
    "terramara": "terremare",
    "terrarium": "terrariums",
    "terrazzo": "terrazzos",
    "territorium": "territoria",
    "terzetto": "terzetti",
    "tessella": "tessellas",
    "tessera": "tesserae",
    "testes": "testes",
    "testis": "testes",
    "testudo": "testudos",
    "tetanothrum": "tetanothra",
    "tete a tete": "tete a tetes",
    "tete-a-tete": "tete-a-tetes",
    "tetracolon": "tetracola",
    "tetragrammaton": "tetragrammata",
    "tetrahedron": "tetrahedrons",
    "tetrapleuron": "tetrapleurons",
    "tetrasporangium": "tetrasporangia",
    "thalamos": "thalamoi",
    "thalamus": "thalami",
    "thanatorium": "thanatoria",
    "that": "those",
    "thaumaturgus": "thaumaturgi",
    "thebe": "thebe",
    "theca": "thecae",
    "theirs": "theirs",
    "thelycum": "thelyca",
    "them": "them",
    "thema": "themata",
    "themself": "themselves",
    "theologoumenon": "theologoumena",
    "thesaurus": "thesauruses",
    "thesis": "theses",
    "they": "they",
    "thief": "thieves",
    "this": "these",
    "tholos": "tholoi",
    "tholus": "tholi",
    "thorax": "thoraxes",
    "thos": "thoes",
    "thraso": "thrasoes",
    "thymiaterion": "thymiateria",
    "thymus": "thymi",
    "thyridium": "thyridia",
    "thyrsus": "thyrsi",
    "tibia": "tibiae",
    "timorese": "timorese",
    "timpano": "timpanos",
    "tintinnabulum": "tintinnabula",
    "tipula": "tipulae",
    "tiro": "tiros",
    "tirolese": "tirolese",
    "to": "toes",
    "tobacco": "tobaccos",
    "tocharese": "tocharese",
    "toe": "toes",
    "togo": "togos",
    "tokyo": "tokyos",
    "tolkach": "tolkachi",
    "tomato": "tomatoes",
    "tombarolo": "tombaroli",
    "tomium": "tomia",
    "tondo": "tondi",
    "tong": "tongs",
    "tongs": "tongs",
    "tonkinese": "tonkinese",
    "tonneau": "tonneaus",
    "tooth": "teeth",
    "topaz": "topazes",
    "topos": "topoi",
    "torero": "toreros",
    "toril": "toriles",
    "torinese": "torinese",
    "torino": "torinos",
    "torma": "tormas",
    "tornese": "tornesi",
    "tornus": "torni",
    "toronto": "torontos",
    "torso": "torsos",
    "tortilla espanola": "tortillas espanolas",
    "tortrix": "tortrices",
    "torula": "torulae",
    "torulus": "toruli",
    "torus": "toruses",
    "tour de force": "tours de force",
    "toxicum": "toxica",
    "toxius": "toxii",
    "toxoplasma": "toxoplasmata",
    "trabea": "trabeae",
    "trabecula": "trabeculae",
    "trachea": "tracheae",
    "tractatrix": "tractatrices",
    "tractellum": "tractella",
    "tractrix": "tractrices",
    "tragus": "tragi",
    "trapezium": "trapeziums",
    "trapezius": "trapezii",
    "trapezohedron": "trapezohedrons",
    "trattoria": "trattorias",
    "trauma": "traumas",
    "trellis": "trellises",
    "tremolo": "tremolos",
    "treponema": "treponemata",
    "trichidium": "trichidia",
    "trichina": "trichine",
    "trichoma": "trichomata",
    "triclinium": "triclinia",
    "triens": "trientes",
    "triforium": "triforia",
    "trilby": "trilbys",
    "triplum": "tripla",
    "triradius": "triradii",
    "triumvir": "triumvirs",
    "trochus": "trochuses",
    "trockenbeerenauslese": "trockenbeerenauslesen",
    "tropaeolum": "tropaeolums",
    "troparion": "troparia",
    "tropis": "tropides",
    "trousers": "trousers",
    "trousseau": "trousseaus",
    "trout": "trout",
    "trullo": "trulli",
    "trumeau": "trumeaux",
    "tsuba": "tsubas",
    "tsubo": "tsubos",
    "tuber": "tuberes",
    "tuberculoma": "tuberculomas",
    "tuberculum": "tubercula",
    "tubulus": "tubuli",
    "tumulus": "tumuli",
    "tuna": "tuna",
    "turbot": "turbots",
    "turf": "turfs",
    "turron": "turrons",
    "tympanum": "tympana",
    "typo": "typos",
    "tyro": "tyros",
    "tyrolese": "tyrolese",
    "udarnik": "udarniki",
    "ulpan": "ulpanim",
    "ultimatum": "ultimatums",
    "umbilicus": "umbilicuses",
    "umbo": "umbos",
    "umbra": "umbras",
    "umwelt": "umwelten",
    "uncia": "unciae",
    "uncinus": "uncini",
    "uncus": "unci",
    "unguis": "ungues",
    "unio": "unios",
    "uomo": "uomini",
    "uomo universale": "uomini universali",
    "uraeus": "uraei",
    "uredinium": "uredinia",
    "uredium": "uredia",
    "uredosorus": "uredosori",
    "urodaeum": "urodaea",
    "urodeum": "urodea",
    "urus": "uruses",
    "usnea": "usneas",
    "ustilago": "ustilagines",
    "uterus": "uteruses",
    "utricularia": "utriculariae",
    "uva": "uvae",
    "vaalpens": "vaalpens",
    "vacuum": "vacuums",
    "vagina": "vaginas",
    "vaginula": "vaginulae",
    "vagus": "vagi",
    "vallecula": "valleculae",
    "valuta": "valute",
    "valvula": "valvulae",
    "vaporetto": "vaporettos",
    "vaquero": "vaqueros",
    "varella": "varellaes",
    "varia lectio": "variae lectiones",
    "varix": "varices",
    "vas": "vasa",
    "vasculitis": "vasculitises",
    "vasculum": "vasculums",
    "veduta": "vedutas",
    "veg": "veges",
    "vehiculum": "vehicula",
    "velamen": "velamina",
    "velarium": "velaria",
    "velum": "velums",
    "vena": "venae",
    "venus": "venuses",
    "vermicello": "vermicellos",
    "vermiculus": "vermiculi",
    "vermontese": "vermontese",
    "veronese": "veronese",
    "verruca": "verrucae",
    "vers libre": "vers libres",
    "versiculus": "versiculi",
    "verso": "versos",
    "vertebra": "vertebrae",
    "vertex": "vertexes",
    "verticillus": "verticilli",
    "vesicula": "vesiculae",
    "vestigium": "vestigia",
    "vetturino": "vetturini",
    "viale": "viali",
    "viaticum": "viatica",
    "vibex": "vibices",
    "vibraculum": "vibracula",
    "vibrato": "vibratos",
    "vibrio": "vibrios",
    "viennese": "viennese",
    "vietnamese": "vietnamese",
    "vila": "vilas",
    "villanella": "villanelle",
    "villino": "villini",
    "villus": "villi",
    "vinculum": "vincula",
    "violetta": "violette",
    "violoncello": "violoncellos",
    "virga": "virgae",
    "virgo": "virgos",
    "virtuoso": "virtuosos",
    "virus": "viruses",
    "vis": "vires",
    "viscus": "viscera",
    "vitellarium": "vitellaria",
    "vitta": "vittae",
    "vivarium": "vivariums",
    "vlakte": "vlaktes",
    "voeu": "voeux",
    "volcanello": "volcanelli",
    "volcano": "volcanos",
    "vomica": "vomicas",
    "vomitorium": "vomitoriums",
    "voorbok": "voorbokke",
    "vorspiel": "vorspiele",
    "vorstellung": "vorstellungen",
    "vortex": "vortexes",
    "wadi": "wadies",
    "wady": "wadies",
    "waldo": "waldos",
    "waltz": "waltzes",
    "wanderjahr": "wanderjahre",
    "wandervogel": "wandervogel",
    "washer up": "washers up",
    "washer-up": "washers-up",
    "weirdo": "weirdos",
    "wenchowese": "wenchowese",
    "wharf": "wharfs",
    "what": "what",
    "whatever": "whatever",
    "whatsoever": "whatsoever",
    "white staff": "white staves",
    "whiting": "whiting",
    "whitmanese": "whitmanese",
    "who": "who",
    "whoever": "whoever",
    "whom": "whom",
    "whomever": "whomever",
    "whomsoever": "whomsoever",
    "whose": "whose",
    "whosesoever": "whosesoever",
    "whosever": "whosever",
    "whosoever": "whosoever",
    "wife": "wives",
    "wildebeest": "wildebeests",
    "wolf": "wolves",
    "woman child": "women children",
    "woman servant": "women servants",
    "woman-child": "women-children",
    "woman-servant": "women-servants",
    "womanchild": "womenchildren",
    "wreath": "wreaths",
    "xanthoma": "xanthomas",
    "xiphiplastron": "xiphiplastra",
    "xiphopagus": "xiphopagi",
    "xiphoplastron": "xiphoplastra",
    "xoanon": "xoana",
    "xystus": "xysti",
    "yakiman": "yakimans",
    "yamoussoukro": "yamoussoukros",
    "yarraman": "yarramen",
    "yengeese": "yengeese",
    "yes": "yeses",
    "yo": "yoes",
    "yo yo": "yo yos",
    "yo-yo": "yo-yos",
    "yoe": "yoes",
    "yokohaman": "yokohamans",
    "you": "you",
    "yours": "yours",
    "yourself": "yourselves",
    "yuman": "yumans",
    "yunnanese": "yunnanese",
    "yuo": "yuon",
    "zadruga": "zadrugas",
    "zeitgeber": "zeitgebers",
    "zemirah": "zemiroth",
    "zeppole": "zeppoli",
    "zero": "zeros",
    "zibo": "zibos",
    "zingano": "zingani",
    "zita": "ziti",
    "zitella": "zitelle",
    "zizz": "zizzes",
    "zloty": "zlotys",
    "zoaea": "zoaeae",
    "zoarium": "zoaria",
    "zoea": "zoeas",
    "zooea": "zooeae",
    "zooecium": "zooecia",
    "zoophorus": "zoophori",
    "zoosporangium": "zoosporangia",
    "zophorus": "zophori",
    "zucca": "zucche",
    "zucchini": "zucchinis",
    "zuppa": "zuppe",
    "zuz": "zuzzes",
    "zygantrum": "zygantra",
    "zygoma": "zygomas",
    "zygon": "zyga"
}

classical_plural_of = {
    "Achinese": "Achinese",
    "Alfurese": "Alfurese",
    "Amboinese": "Amboinese",
    "Americanese": "Americanese",
    "Amoyese": "Amoyese",
    "Andamanese": "Andamanese",
    "Angolese": "Angolese",
    "Annamese": "Annamese",
    "Antananarivo": "Antananarivos",
    "Aragonese": "Aragonese",
    "Arakanese": "Arakanese",
    "Asherah": "Asherim",
    "Assamese": "Assamese",
    "Atlas": "Atlantes",
    "Auslese": "Auslesen",
    "Avignonese": "Avignonese",
    "Azerbaijanese": "Azerbaijanese",
    "Bahaman": "Bahamans",
    "Balinese": "Balinese",
    "Bamako": "Bamakos",
    "Barquisimeto": "Barquisimetos",
    "Bengalese": "Bengalese",
    "Bernese": "Bernese",
    "Bhutanese": "Bhutanese",
    "Biro": "Biros",
    "Bolognese": "Bolognese",
    "Bolzano": "Bolzanos",
    "Borghese": "Borghese",
    "Bostonese": "Bostonese",
    "Boto": "Botos",
    "Buginese": "Buginese",
    "Burman": "Burmans",
    "Burmese": "Burmese",
    "Cairo": "Cairos",
    "Calabrese": "Calabrese",
    "Camaldolese": "Camaldolese",
    "Cantonese": "Cantonese",
    "Carlylese": "Carlylese",
    "Cassinese": "Cassinese",
    "Celanese": "Celanese",
    "Ceylonese": "Ceylonese",
    "Chicago": "Chicagos",
    "Chimango": "Chimangos",
    "Chinese": "Chinese",
    "Colombo": "Colombos",
    "Colorado": "Colorados",
    "Congoese": "Congoese",
    "Congolese": "Congolese",
    "Czech": "Czechs",
    "Dongolese": "Dongolese",
    "Draco": "Dracos",
    "Esperanto": "Esperantos",
    "Faro": "Faros",
    "Faroese": "Faroese",
    "Ferrarese": "Ferrarese",
    "Filipino": "Filipinos",
    "Foochowese": "Foochowese",
    "Gabunese": "Gabunese",
    "Galla": "Galla",
    "Genevese": "Genevese",
    "Genoese": "Genoese",
    "German": "Germans",
    "Gestapo": "Gestapos",
    "Gilbertese": "Gilbertese",
    "Goanese": "Goanese",
    "Greensboro": "Greensboros",
    "Guaiabero": "Guaiaberos",
    "Guianese": "Guianese",
    "Hainanese": "Hainanese",
    "Hararese": "Hararese",
    "Harlemese": "Harlemese",
    "Havanese": "Havanese",
    "Heavenese": "Heavenese",
    "Hiroshiman": "Hiroshimans",
    "Hoosierese": "Hoosierese",
    "Hottentotese": "Hottentotese",
    "Hunanese": "Hunanese",
    "I": "we",
    "Idaho": "Idahos",
    "Indochinese": "Indochinese",
    "Iquico": "Iquicos",
    "Japanese": "Japanese",
    "Javanese": "Javanese",
    "Jerry": "Jerrys",
    "Johnsonese": "Johnsonese",
    "Kakapo": "Kakapos",
    "Kanarese": "Kanarese",
    "Kinkimavo": "Kinkimavos",
    "Kiplingese": "Kiplingese",
    "Kokako": "Kokakos",
    "Kongoese": "Kongoese",
    "Kongolese": "Kongolese",
    "Kosovo": "Kosovos",
    "Lapith": "Lapithae",
    "Lapponese": "Lapponese",
    "Lebanese": "Lebanese",
    "Leonese": "Leonese",
    "Lesotho": "Lesothos",
    "Lilo": "Lilos",
    "Liman": "Limans",
    "Londonese": "Londonese",
    "Lorrainese": "Lorrainese",
    "Lucchese": "Lucchese",
    "Lyonese": "Lyonese",
    "Macanese": "Macanese",
    "Macassarese": "Macassarese",
    "Madurese": "Madurese",
    "Majuro": "Majuros",
    "Malabarese": "Malabarese",
    "Malabo": "Malabos",
    "Maltese": "Maltese",
    "Maputo": "Maputos",
    "Maracaibo": "Maracaibos",
    "Mary": "Marys",
    "Messinese": "Messinese",
    "Mexico": "Mexicos",
    "Milanese": "Milanese",
    "Milano": "Milanos",
    "Modenese": "Modenese",
    "Monaco": "Monacos",
    "Montenegro": "Montenegros",
    "Morocco": "Moroccos",
    "Muqdisho": "Muqdishos",
    "Muranese": "Muranese",
    "Nakayaman": "Nakayamans",
    "Nankingese": "Nankingese",
    "Navarrese": "Navarrese",
    "Nenets": "Nentsi",
    "Nepalese": "Nepalese",
    "Niasese": "Niasese",
    "Nicobarese": "Nicobarese",
    "Ningbo": "Ningbos",
    "Nipponese": "Nipponese",
    "Norman": "Normans",
    "Oklahoman": "Oklahomans",
    "Orinoco": "Orinocos",
    "Orlando": "Orlandos",
    "Oslo": "Oslos",
    "Otomi": "Otomi",
    "Panaman": "Panamans",
    "Pangasinan": "Pangasinanes",
    "Paramaribo": "Paramaribos",
    "Pardusco": "Parduscos",
    "Parmese": "Parmese",
    "Pekingese": "Pekingese",
    "Piedmontese": "Piedmontese",
    "Pistoiese": "Pistoiese",
    "Pluto": "Plutos",
    "Polonese": "Polonese",
    "Porto": "Portos",
    "Porto Novo": "Porto Novos",
    "Porto-Novo": "Porto-Novos",
    "Portuguese": "Portuguese",
    "Purum": "Purum",
    "Quito": "Quitos",
    "Rom": "Roma",
    "Romagnese": "Romagnese",
    "Roman": "Romans",
    "Romanese": "Romanese",
    "Romany": "Romanies",
    "Sacramento": "Sacramentos",
    "Sangirese": "Sangirese",
    "Santiago": "Santiagos",
    "Sapporo": "Sapporos",
    "Sarajevo": "Sarajevos",
    "Sarawakese": "Sarawakese",
    "Selman": "Selmans",
    "Senegalese": "Senegalese",
    "Shavese": "Shavese",
    "Shawanese": "Shawanese",
    "Siamese": "Siamese",
    "Sienese": "Sienese",
    "Sikkimese": "Sikkimese",
    "Silurus": "Siluri",
    "Singhalese": "Singhalese",
    "Sinhalese": "Sinhalese",
    "Sogdianese": "Sogdianese",
    "Sonaman": "Sonamans",
    "Spaetlese": "Spaetlesen",
    "Sudanese": "Sudanese",
    "Sundanese": "Sundanese",
    "Swahilese": "Swahilese",
    "Tacoman": "Tacomans",
    "Taiko": "Taikos",
    "Tenggerese": "Tenggerese",
    "Timorese": "Timorese",
    "Tirolese": "Tirolese",
    "Tocharese": "Tocharese",
    "Togo": "Togos",
    "Tokyo": "Tokyos",
    "Tonkinese": "Tonkinese",
    "Torinese": "Torinese",
    "Torino": "Torinos",
    "Tornese": "Tornesi",
    "Toronto": "Torontos",
    "Tyrolese": "Tyrolese",
    "Vermontese": "Vermontese",
    "Veronese": "Veronese",
    "Viennese": "Viennese",
    "Vietnamese": "Vietnamese",
    "Virgo": "Virgos",
    "Wenchowese": "Wenchowese",
    "Whitmanese": "Whitmanese",
    "Yakiman": "Yakimans",
    "Yamoussoukro": "Yamoussoukros",
    "Yengeese": "Yengeese",
    "Yokohaman": "Yokohamans",
    "Yuman": "Yumans",
    "Yunnanese": "Yunnanese",
    "Zibo": "Zibos",
    "abacus": "abacuses",
    "aboiteau": "aboiteaux",
    "abortorium": "abortoria",
    "abortus": "abortuses",
    "abscissa": "abscissae",
    "abstractum": "abstracta",
    "acarus": "acari",
    "achinese": "achinese",
    "acicula": "aciculae",
    "acinus": "acini",
    "acroama": "acroamata",
    "acropolis": "acropolises",
    "actinia": "actiniae",
    "aculeus": "aculei",
    "addendum": "addenda",
    "adieu": "adieux",
    "adjutant general": "adjutant generals",
    "adminiculum": "adminicula",
    "ado": "ados",
    "adviso": "advisoes",
    "adytum": "adyta",
    "aecidium": "aecidia",
    "aegis": "aegises",
    "aelf": "aelfe",
    "aficionado": "aficionados",
    "afreet": "afreeti",
    "afrit": "afriti",
    "agape": "agapae",
    "agend": "agenda",
    "agendum": "agenda",
    "agent provocateur": "agents provocateurs",
    "aggro": "aggros",
    "agon": "agones",
    "agora": "agorae",
    "agrogorod": "agrogoroda",
    "aide de camp": "aides de camp",
    "aide-de-camp": "aides-de-camp",
    "ala": "alae",
    "alan": "alani",
    "albergo": "alberghi",
    "albino": "albinos",
    "ale wife": "ale wives",
    "ale-wife": "ale-wives",
    "alfurese": "alfurese",
    "alga": "algae",
    "alguacil": "alguaciles",
    "alias": "aliases",
    "allargando": "allargandi",
    "allegro": "allegros",
    "alluvium": "alluvia",
    "alto": "alti",
    "alto relievo": "alto relievos",
    "alto-relievo": "alto-relievos",
    "altorelievo": "altorelievos",
    "alula": "alulae",
    "alumna": "alumnae",
    "alumnus": "alumni",
    "alveolus": "alveoli",
    "ambo": "ambones",
    "amboinese": "amboinese",
    "ambulacrum": "ambulacra",
    "ambuscado": "ambuscadoes",
    "amentum": "amenta",
    "americanese": "americanese",
    "amicus curiae": "amici curiae",
    "ammo": "ammos",
    "amoeba": "amoebae",
    "amoretto": "amoretti",
    "amorino": "amorini",
    "amoyese": "amoyese",
    "amphipneust": "amphipneusta",
    "amphora": "amphorae",
    "anacoluthon": "anacolutha",
    "anathema": "anathemata",
    "ancona": "ancone",
    "andamanese": "andamanese",
    "angioma": "angiomata",
    "angolese": "angolese",
    "angustia": "angustiae",
    "anlage": "anlagen",
    "annamese": "annamese",
    "anta": "antae",
    "antananarivo": "antananarivos",
    "antefix": "antefixa",
    "antenna": "antennae",
    "anthemion": "anthemia",
    "anthodium": "anthodia",
    "anti serum": "anti sera",
    "anti-serum": "anti-sera",
    "antiserum": "antisera",
    "antrum": "antra",
    "apex": "apices",
    "aphelion": "aphelia",
    "aphis": "aphides",
    "apothecium": "apothecia",
    "apparatchik": "apparatchiki",
    "apparatus": "apparatus",
    "appendix": "appendices",
    "appressorium": "appressoria",
    "apsis": "apsides",
    "apterium": "apteria",
    "aptychus": "aptychi",
    "aquarium": "aquaria",
    "aragonese": "aragonese",
    "arakanese": "arakanese",
    "arango": "arangoes",
    "arboretum": "arboreta",
    "arcanum": "arcana",
    "archegonium": "archegonia",
    "archipelago": "archipelagoes",
    "arcosolium": "arcosolia",
    "areola": "areolae",
    "argumentum": "argumenta",
    "arista": "aristae",
    "armadillo": "armadillos",
    "arrha": "arrhae",
    "asbestos": "asbestoses",
    "ascidium": "ascidia",
    "asherah": "asherim",
    "assamese": "assamese",
    "asterias": "asteriae",
    "astrocytoma": "astrocytomata",
    "asylum": "asyla",
    "asyndeton": "asyndeta",
    "ataman": "atamans",
    "atlas": "atlantes",
    "atman": "atmas",
    "aureus": "aurei",
    "aurora": "aurorae",
    "auslese": "auslesen",
    "auspex": "auspices",
    "auto": "autos",
    "auto de fe": "autos de fes",
    "auto-de-fe": "autos-de-fes",
    "autobahn": "autobahnen",
    "autochthon": "autochthones",
    "autostrada": "autostrade",
    "avicularium": "avicularia",
    "avignonese": "avignonese",
    "aviso": "avisos",
    "avocado": "avocados",
    "ax": "axes",
    "axe": "axes",
    "axilla": "axillae",
    "axis": "axes",
    "axon": "axones",
    "axopodium": "axopodia",
    "azerbaijanese": "azerbaijanese",
    "baal": "baalim",
    "bacillus": "bacilli",
    "bacterium": "bacteria",
    "bahaman": "bahamans",
    "bajocco": "bajocchi",
    "balinese": "balinese",
    "ballabile": "ballabili",
    "ballata": "ballate",
    "ballerina": "ballerine",
    "ballista": "ballistae",
    "bamako": "bamakos",
    "bandeau": "bandeaux",
    "bandit": "banditti",
    "barquisimeto": "barquisimetos",
    "barrico": "barricoes",
    "base": "bases",
    "basilica": "basilicae",
    "basis": "bases",
    "bass": "bass",
    "basso": "bassi",
    "bateau": "bateaux",
    "bath": "baths",
    "bathos": "bathoses",
    "batz": "batzes",
    "beau": "beaux",
    "beau geste": "beaux gestes",
    "bechuana": "bechuana",
    "beef": "beeves",
    "beerenauslese": "beerenauslesen",
    "bel esprit": "beaux esprits",
    "bel homme": "beaux hommes",
    "belle epoque": "belles epoques",
    "bema": "bemata",
    "bengalese": "bengalese",
    "bernese": "bernese",
    "bersagliere": "bersaglieri",
    "bhikshu": "bhikshook",
    "bhutanese": "bhutanese",
    "bias": "biases",
    "bierstube": "bierstuben",
    "bifolium": "bifolia",
    "bijou": "bijoux",
    "bilbo": "bilboes",
    "bimbo": "bimbos",
    "bingo": "bingos",
    "biro": "biros",
    "biscotto": "biscotti",
    "bison": "bison",
    "blastema": "blastemata",
    "blemya": "blemyae",
    "blin": "blini",
    "blitz": "blitzes",
    "blouse": "blouses",
    "blue": "blues",
    "blues": "blues",
    "bodegon": "bodegones",
    "bolero": "boleros",
    "bolognese": "bolognese",
    "bolus": "boluses",
    "bolzano": "bolzanos",
    "bon mot": "bons mots",
    "bon vivant": "bons vivants",
    "bona fides": "bonis fidebus",
    "bongo": "bongoes",
    "bordereau": "bordereaux",
    "borghese": "borghese",
    "bostonese": "bostonese",
    "boto": "botos",
    "bozzetto": "bozzetti",
    "braccio": "braccia",
    "brachiopod": "brachiopoda",
    "branchiopod": "branchiopoda",
    "bratwurst": "bratwurste",
    "bravo": "bravoes",
    "break away": "breaks away",
    "break-away": "breaks-away",
    "breakaway": "breaksaway",
    "bream": "bream",
    "breech": "breeches",
    "breeches": "breeches",
    "bregma": "bregmata",
    "brigadier general": "brigadier generals",
    "britches": "britches",
    "bronchitis": "bronchitises",
    "bronchus": "bronchi",
    "brother": "brethren",
    "brother german": "brethren german",
    "brother-german": "brethren-german",
    "bruschetta": "bruschette",
    "bubo": "buboes",
    "buffalo": "buffalo",
    "buginese": "buginese",
    "bulimus": "bulimi",
    "bulla": "bullae",
    "bureau": "bureaux",
    "burman": "burmans",
    "burmese": "burmese",
    "burro": "burros",
    "bursa": "bursae",
    "bursch": "burschen",
    "bursitis": "bursitises",
    "busto": "bustoes",
    "buzz": "buzzes",
    "bylina": "byliny",
    "cabaletta": "cabalette",
    "cactus": "cacti",
    "caddis": "caddises",
    "caduceus": "caducei",
    "caecum": "caeca",
    "caffe macchiato": "caffe macchiati",
    "caffe ristretto": "caffe ristretti",
    "caiman": "caimans",
    "cairo": "cairos",
    "calabrese": "calabrese",
    "calamistrum": "calamistra",
    "calathus": "calathi",
    "calix": "calices",
    "callus": "calluses",
    "calypso": "calypsos",
    "calzone": "calzoni",
    "camaldolese": "camaldolese",
    "camera obscura": "camerae obscurae",
    "canaliculus": "canaliculi",
    "candelabrum": "candelabra",
    "cannabis": "cannabises",
    "cantharus": "canthari",
    "canto": "cantos",
    "cantonese": "cantonese",
    "cantus": "cantus",
    "canvas": "canvases",
    "canzonetta": "canzonette",
    "capataz": "capataces",
    "capo": "capi",
    "cappuccino": "cappuccinos",
    "carabiniere": "carabinieri",
    "carcinoma": "carcinomata",
    "cardo": "cardines",
    "cardophagus": "cardophagi",
    "carex": "carices",
    "cargador": "cargadores",
    "caribou": "caribou",
    "carlylese": "carlylese",
    "carmen": "carmina",
    "carp": "carp",
    "carte de visite": "cartes de visite",
    "carte-de-visite": "cartes-de-visite",
    "caryatid": "caryatides",
    "caryopsis": "caryopsides",
    "casino": "casinos",
    "cassinese": "cassinese",
    "cassone": "cassoni",
    "castrato": "castrati",
    "castrum": "castra",
    "casus belli": "casus belli",
    "cathisma": "cathismata",
    "caudex": "caudices",
    "caulis": "caules",
    "cavea": "caveae",
    "cayman": "caymans",
    "cecidium": "cecidia",
    "celanese": "celanese",
    "celia": "celiae",
    "cello": "cellos",
    "cementoma": "cementomata",
    "centesimo": "centesimi",
    "centumvir": "centumviri",
    "cercus": "cerci",
    "ceriman": "cerimans",
    "ceylonese": "ceylonese",
    "chaeta": "chaetae",
    "chaos": "chaoses",
    "chapeau": "chapeaux",
    "chapplis": "chapplies",
    "charisma": "charismata",
    "chassis": "chassis",
    "chateau": "chateaux",
    "cherub": "cherubim",
    "chervonetz": "chervontzi",
    "cheval": "chevaux",
    "cheval de bataille": "chevaux de bataille",
    "cheval-de-bataille": "chevaux-de-bataille",
    "chicago": "chicagos",
    "child": "children",
    "chimango": "chimangos",
    "chinese": "chinese",
    "chintz": "chintzes",
    "chitarrone": "chitarroni",
    "chlamydia": "chlamydiae",
    "choana": "choanae",
    "chorda": "chordae",
    "chorus": "choruses",
    "chou": "choux",
    "chromidium": "chromidia",
    "chromonema": "chromonemata",
    "chrysalis": "chrysalides",
    "cicatrix": "cicatrices",
    "cilantro": "cilantros",
    "cimex": "cimices",
    "cinclis": "cinclides",
    "cistophorus": "cistophorori",
    "cladus": "cladi",
    "claustrum": "claustra",
    "clausula": "clausulae",
    "clavula": "clavulae",
    "cleithrum": "cleithra",
    "clepsydra": "clepsydrae",
    "clipper": "clippers",
    "clippers": "clippers",
    "clitoris": "clitorides",
    "clostridium": "clostridia",
    "co": "coes",
    "coagulum": "coagula",
    "coccidium": "coccidia",
    "cochito": "cochitos",
    "coco": "cocos",
    "cod": "cod",
    "codex": "codices",
    "coe": "coes",
    "coelia": "coeliae",
    "cognoscente": "cognoscenti",
    "coitus": "coitus",
    "collegium": "collegia",
    "collegium musicum": "collegia musica",
    "colombo": "colombos",
    "colorado": "colorados",
    "columbarium": "columbaria",
    "columna": "columnae",
    "coma": "comae",
    "comatula": "comatulae",
    "commando": "commandos",
    "compendium": "compendia",
    "competent": "competentes",
    "concertino": "concertinos",
    "concerto grosso": "concerti grossi",
    "concetto": "concetti",
    "condo": "condos",
    "condottiere": "condottieri",
    "conductus": "conducti",
    "coney": "conies",
    "conferva": "confervae",
    "confluvium": "confluvia",
    "congius": "congii",
    "congoese": "congoese",
    "congolese": "congolese",
    "conidium": "conidia",
    "conoscente": "conoscenti",
    "consortium": "consortia",
    "contadina": "contadine",
    "contadino": "contadini",
    "contagium": "contagia",
    "contango": "contangos",
    "contrafactum": "contrafacta",
    "contralto": "contralti",
    "contretemps": "contretemps",
    "conversazione": "conversazioni",
    "convolvulus": "convolvuli",
    "cony": "conies",
    "corgi": "corgwn",
    "corgy": "corgwn",
    "cornetto": "cornetti",
    "corno": "corni",
    "cornu": "cornua",
    "corona": "coronae",
    "corps": "corps",
    "corpus": "corpora",
    "corpus vile": "corpora vilia",
    "corpusculum": "corpuscula",
    "corrigendum": "corrigenda",
    "cortex": "cortices",
    "cosmos": "cosmoses",
    "costa": "costae",
    "coup d'etat": "coups d'etat",
    "court martial": "courts martial",
    "court-martial": "courts-martial",
    "cow": "kine",
    "coxendix": "coxendices",
    "coz": "cozes",
    "craft": "craft",
    "crafts child": "crafts childer",
    "crafts-child": "crafts-childer",
    "cranium": "crania",
    "craspedon": "craspeda",
    "craspedum": "craspeda",
    "crazia": "crazie",
    "credo": "credos",
    "cremaster": "cremasteres",
    "crescendo": "crescendos",
    "cri de coeur": "cris de coeur",
    "crista": "cristae",
    "criterion": "criteria",
    "crowfoot": "crowfoots",
    "crus": "crura",
    "crusta": "crustae",
    "cubiculum": "cubicula",
    "cul de four": "culs de four",
    "cul de lampe": "culs de lampe",
    "cul de sac": "culs de sac",
    "cul-de-four": "culs-de-four",
    "cul-de-lampe": "culs-de-lampe",
    "cul-de-sac": "culs-de-sac",
    "cultellus": "cultelli",
    "cumulus": "cumuli",
    "cuniculus": "cuniculi",
    "curioso": "curiosi",
    "curriculum": "curricula",
    "curriculum vitae": "curricula vitae",
    "cuspis": "cuspides",
    "cyano": "cyanos",
    "cyathus": "cyathi",
    "cyclops": "cyclopes",
    "cynocephalus": "cynocephali",
    "cyphella": "cyphellae",
    "cysticercus": "cysticerci",
    "cystidium": "cystidia",
    "cystoma": "cystomata",
    "czech": "czechs",
    "dace": "dace",
    "dais": "daises",
    "datum": "data",
    "debris": "debris",
    "decennium": "decennia",
    "deer": "deer",
    "delirium": "deliria",
    "demo": "demos",
    "denarius": "denarii",
    "denkmal": "denkmaler",
    "denotatum": "denotata",
    "dentalium": "dentalia",
    "depositum": "deposita",
    "descriptum": "descripta",
    "desideratum": "desiderata",
    "designatum": "designata",
    "desma": "desmata",
    "desman": "desmans",
    "devoto": "devoti",
    "diabetes": "diabetes",
    "diastema": "diastemata",
    "dichasium": "dichasia",
    "dictum": "dicta",
    "didymis": "didymes",
    "die": "dies",
    "differentia": "differentiae",
    "digitalis": "digitalises",
    "dilettante": "dilettanti",
    "diplococcus": "diplococci",
    "diploma": "diplomata",
    "diporpa": "diporpae",
    "dipsas": "dipsades",
    "directrix": "directrices",
    "ditto": "dittos",
    "diverticulum": "diverticula",
    "divertimento": "divertimenti",
    "djinn": "djinn",
    "do": "does",
    "doe": "does",
    "dogma": "dogmata",
    "dolman": "dolmans",
    "dolphinarium": "dolphinaria",
    "dongolese": "dongolese",
    "draco": "dracos",
    "dragoman": "dragomans",
    "drama": "dramas",
    "dreikanter": "dreikanter",
    "dumka": "dumky",
    "duumvir": "duumviri",
    "dynamo": "dynamos",
    "edema": "edemata",
    "editio princeps": "editiones principes",
    "efreet": "efreeti",
    "eland": "eland",
    "elf": "elves",
    "elk": "elk",
    "ellipse": "ellipses",
    "ellipsis": "ellipses",
    "emblema": "emblememata",
    "embryo": "embryos",
    "embryoma": "embryomata",
    "emporium": "emporia",
    "enconium": "enconia",
    "endothelioma": "endotheliomata",
    "enema": "enemata",
    "enigma": "enigmata",
    "ens rationis": "entia rationis",
    "ens reale": "entia realia",
    "entente cordiale": "ententes cordiales",
    "enteron": "entera",
    "entrochus": "entrochi",
    "ephelis": "ephelides",
    "ephemera": "ephemerae",
    "ephemeris": "ephemerides",
    "epicardium": "epicardia",
    "epidermis": "epidermises",
    "epididymis": "epididymides",
    "epilimnion": "epilimnia",
    "epimeron": "epimera",
    "epiphenomenon": "epiphenomena",
    "epithelioma": "epitheliomata",
    "epyllion": "epyllia",
    "equinoctium": "equinoctia",
    "equisetum": "equiseta",
    "erf": "erven",
    "ergastulum": "ergastula",
    "eros": "erotes",
    "erratum": "errata",
    "espanol": "espanoles",
    "esperanto": "esperantos",
    "espresso": "espressos",
    "ethos": "ethoses",
    "eucalyptus": "eucalypti",
    "eunuch": "eunuchs",
    "euporiston": "euporista",
    "euro": "euros",
    "evirato": "evirati",
    "ex": "exes",
    "exanthema": "exanthemata",
    "exemplum": "exempla",
    "exencephalus": "exencephali",
    "explanandum": "explananda",
    "explicandum": "explicanda",
    "extispex": "extispices",
    "extremum": "extrema",
    "eyas": "eyases",
    "fabella": "fabellae",
    "fabliau": "fabliaux",
    "faciendum": "facienda",
    "faecula": "faeculae",
    "fait accompli": "faits accomplis",
    "falsetto": "falsettos",
    "falx": "falces",
    "familia": "familiae",
    "famulus": "famuli",
    "fanfoot": "fanfoots",
    "farman": "farmans",
    "faro": "faros",
    "faroese": "faroese",
    "fascia": "fasciae",
    "fasciculus": "fasciculi",
    "fauna": "faunae",
    "favella": "favellae",
    "faveolus": "faveoli",
    "favissa": "favissae",
    "fecula": "feculae",
    "fedai": "fedai",
    "feis": "feiseanna",
    "felo": "felones",
    "felo de se": "felones de se",
    "felo-de-se": "felones-de-se",
    "felsenmeer": "felsenmeere",
    "femur": "femora",
    "fenestra": "fenestrae",
    "fermata": "fermate",
    "ferrarese": "ferrarese",
    "fiasco": "fiascos",
    "fibra": "fibrae",
    "fibrilla": "fibrillae",
    "fibroma": "fibromata",
    "fibula": "fibulae",
    "fidei defensor": "fidei defensores",
    "figurante": "figuranti",
    "filaria": "filariae",
    "filipino": "filipinos",
    "finger end": "fingers ends",
    "finger-end": "fingers-ends",
    "fioritura": "fioriture",
    "fish": "fish",
    "fizz": "fizzes",
    "flabellum": "flabella",
    "flag staff": "flag staves",
    "flag-staff": "flag-staves",
    "flagellum": "flagella",
    "flamenco": "flamencos",
    "flatfoot": "flatfoots",
    "flatus": "flatuses",
    "flocculus": "flocculi",
    "floccus": "flocci",
    "flora": "florae",
    "flounder": "flounder",
    "fly": "flys",
    "focus": "foci",
    "follis": "folles",
    "fomes": "fomites",
    "foochowese": "foochowese",
    "foot": "feet",
    "foramen": "foramina",
    "formula": "formulae",
    "forum": "fora",
    "fossa": "fossae",
    "fowl": "fowl",
    "fraenum": "fraena",
    "franc archer": "francs archers",
    "franc-archer": "francs-archers",
    "frate": "frati",
    "frenum": "fraena",
    "fricandeau": "fricandeaux",
    "frizz": "frizzes",
    "fro": "froes",
    "froe": "froes",
    "frottola": "frottole",
    "frustulum": "frustula",
    "frustum": "frusta",
    "fucus": "fuci",
    "fulcrum": "fulcra",
    "funambulus": "funambuli",
    "fundamentum divisionis": "fundamenta divisionis",
    "fungo porcino": "funghi porchini",
    "fungus": "fungi",
    "furfur": "furfures",
    "furioso": "furiosos",
    "fuzz": "fuzzes",
    "gabunese": "gabunese",
    "galla": "galla",
    "gallows": "gallows",
    "gametangium": "gametangia",
    "ganglion": "ganglia",
    "gaon": "gaonim",
    "gas": "gases",
    "gastrocnemius": "gastrocnemii",
    "geisha": "geisha",
    "geisteswissenschaft": "geisteswissenschaften",
    "gemma": "gemmae",
    "gena": "genae",
    "generalissimo": "generalissimos",
    "genevese": "genevese",
    "genie": "genii",
    "genius": "genii",
    "genoese": "genoese",
    "gens": "gentes",
    "genus": "genera",
    "german": "germans",
    "germarium": "germaria",
    "gestapo": "gestapos",
    "ghetto": "ghettos",
    "gigolo": "gigolos",
    "gilbertese": "gilbertese",
    "giuncus": "giunchi",
    "gizmo": "gizmos",
    "glacis": "glacis",
    "gladiolus": "gladioli",
    "glass": "glasses",
    "glasses": "glasses",
    "glioma": "gliomata",
    "gliosarcoma": "gliosarcomata",
    "glissando": "glissandi",
    "globigerina": "globigerinae",
    "glochidium": "glochidia",
    "glomus": "glomera",
    "glottis": "glottises",
    "glutaeus": "glutaei",
    "gluteus": "glutaei",
    "gmina": "gminy",
    "go": "goes",
    "goanese": "goanese",
    "gonangium": "gonangia",
    "gonidium": "gonidia",
    "gonimium": "gonimia",
    "goose": "geese",
    "gorgonia": "gorgoniae",
    "gousblom": "gousblomme",
    "goy": "goyim",
    "graben": "graben",
    "graffiti": "graffiti",
    "graffito": "graffiti",
    "granita": "granite",
    "grano": "grani",
    "granula": "granulae",
    "gravamen": "gravamina",
    "greensboro": "greensboros",
    "gringo": "gringos",
    "grissino": "grissini",
    "grotto": "grottoes",
    "grouse": "grouse",
    "guaiabero": "guaiaberos",
    "guano": "guanos",
    "guardia civil": "guardias civiles",
    "gubernaculum": "gubernacula",
    "guglio": "guglioes",
    "guianese": "guianese",
    "gumbo": "gumbos",
    "gumma": "gummata",
    "guttus": "gutti",
    "guy": "guys",
    "gymnasium": "gymnasia",
    "gymnotus": "gymnoti",
    "gyro": "gyros",
    "gyrus": "gyri",
    "haddock": "haddock",
    "haemangioma": "haemangiomata",
    "haggis": "haggis",
    "hainanese": "hainanese",
    "hairdo": "hairdos",
    "hake": "hake",
    "halibut": "halibut",
    "hallux": "halluces",
    "halteridium": "halteridia",
    "hamadryad": "hamadryades",
    "hamulus": "hamuli",
    "hapax legomenon": "hapax legomena",
    "hapteron": "haptera",
    "hararese": "hararese",
    "harlemese": "harlemese",
    "harman": "harmans",
    "haruspex": "haruspices",
    "haustellum": "haustella",
    "haustorium": "haustoria",
    "haustrum": "haustra",
    "havanese": "havanese",
    "he": "they",
    "headquarters": "headquarters",
    "heavenese": "heavenese",
    "hedron": "hedra",
    "heft": "hefte",
    "helion": "helia",
    "helix": "helices",
    "hemangeioma": "hemangeiomata",
    "hemangioma": "hemangiomata",
    "hemelytrum": "hemelytra",
    "hemera": "hemerae",
    "hemi elytrum": "hemi elytra",
    "hemi-elytrum": "hemi-elytra",
    "her": "them",
    "hernia": "herniae",
    "hero": "heroes",
    "heroon": "heroa",
    "heroum": "heroa",
    "herpes": "herpes",
    "herr": "herren",
    "herring": "herring",
    "hers": "theirs",
    "herself": "themselves",
    "hertz": "hertz",
    "hetaera": "hetairai",
    "hetaira": "hetairai",
    "hetman": "hetmans",
    "hiatus": "hiatus",
    "hijinks": "hijinks",
    "him": "them",
    "himself": "themselves",
    "hinaki": "hinaki",
    "hippo": "hippos",
    "hippocampus": "hippocampi",
    "hippopotamus": "hippopotami",
    "hiroshiman": "hiroshimans",
    "his": "theirs",
    "ho": "hoes",
    "hoe": "hoes",
    "holothuria": "holothuriae",
    "honorarium": "honoraria",
    "hoof": "hooves",
    "hoosierese": "hoosierese",
    "hordeolum": "hordeola",
    "hormogonium": "hormogonia",
    "hoti": "hoties",
    "hottentotese": "hottentotese",
    "hsien": "heen",
    "hubris": "hubrises",
    "human": "humans",
    "humerus": "humeri",
    "hunanese": "hunanese",
    "hydra": "hydrae",
    "hydraulicon": "hydraulica",
    "hydria": "hydriae",
    "hymenium": "hymenia",
    "hymnarium": "hymnaria",
    "hyperbaton": "hyperbata",
    "hyperbola": "hyperbolae",
    "hypha": "hyphae",
    "hypnum": "hypna",
    "hypoderma": "hypodermata",
    "hypolimnion": "hypolimnia",
    "i": "we",
    "ibis": "ibises",
    "ichthyosaurus": "ichthyosauri",
    "idaho": "idahos",
    "idee fixe": "idees fixes",
    "idolon": "idola",
    "idolum": "idola",
    "ignoramus": "ignoramus",
    "ilium": "ilia",
    "imago": "imagines",
    "imbrex": "imbrices",
    "impetigo": "impetigos",
    "impetus": "impetus",
    "improperium": "improperia",
    "improvisatrice": "improvisatrici",
    "incubus": "incubi",
    "index": "indices",
    "indicium": "indicia",
    "indigena": "indigenae",
    "individuum": "individua",
    "indochinese": "indochinese",
    "indumentum": "indumenta",
    "indusium": "indusia",
    "inferno": "infernos",
    "infima species": "infimae species",
    "info": "infos",
    "inhalatorium": "inhalatoria",
    "innings": "innings",
    "inoculum": "inocula",
    "insula": "insulae",
    "intaglio": "intaglios",
    "interambulacrum": "interambulacra",
    "interaxis": "interaxes",
    "intercalarium": "intercalare",
    "interleaf": "leaves",
    "intermaxilla": "intermaxillae",
    "intermedium": "intermedia",
    "intermezzo": "intermezzos",
    "internodium": "internodia",
    "interphenomenon": "interphenomena",
    "interradius": "interradii",
    "interregnum": "interregna",
    "interrex": "interreges",
    "interstitium": "interstitia",
    "intertrigo": "intertrigos",
    "intervallum": "intervalla",
    "intichiuma": "intichiuma",
    "inuk": "inuit",
    "inukshuk": "inukshuit",
    "involucrum": "involucra",
    "ipse dixit": "ipse dixits",
    "iquico": "iquicos",
    "iris": "irides",
    "ischium": "ischia",
    "isidium": "isidia",
    "ispravnik": "ispravniki",
    "isthmus": "isthmi",
    "it": "they",
    "iter": "itinera",
    "its": "theirs",
    "itself": "themselves",
    "ixodid": "ixodides",
    "jackanapes": "jackanapes",
    "jacobus": "jacobus",
    "jalapeno": "jalapenos",
    "jambeau": "jambeaux",
    "japanese": "japanese",
    "javanese": "javanese",
    "jazz": "jazzes",
    "jerry": "jerries",
    "jeu": "jeux",
    "jo": "joes",
    "joe": "joes",
    "johnsonese": "johnsonese",
    "judicatum": "judicata",
    "juelet": "iueles",
    "jugum": "juga",
    "jumbo": "jumbos",
    "junto": "juntos",
    "kakapo": "kakapos",
    "kanarese": "kanarese",
    "kantharos": "kantharoi",
    "kar": "kare",
    "karakia": "karakia",
    "karrenfeld": "karrenfelder",
    "kernos": "kernoi",
    "kibbutz": "kibbutzim",
    "kilo": "kilos",
    "kimono": "kimonos",
    "kinkimavo": "kinkimavos",
    "kioko": "kioko",
    "kiplingese": "kiplingese",
    "kletterschuh": "kletterschuhe",
    "knife": "knives",
    "knopper": "knoppern",
    "kobo": "kobo",
    "kokako": "kokakos",
    "konditorei": "konditoreien",
    "kongoese": "kongoese",
    "kongolese": "kongolese",
    "kosovo": "kosovos",
    "kouros": "kouroi",
    "kovsh": "kovshi",
    "kuvasz": "kuvaszok",
    "kylix": "kylikes",
    "labrum": "labra",
    "lacinia": "laciniae",
    "lactobacillus": "lactobacilli",
    "lacuna": "lacunae",
    "lacunar": "lacunaria",
    "laika": "laiki",
    "lamella": "lamellae",
    "lamina": "laminae",
    "lapith": "lapithae",
    "lapponese": "lapponese",
    "lar": "lares",
    "larnax": "larnakes",
    "larynx": "larynges",
    "latex": "latices",
    "leaf": "leaves",
    "lebanese": "lebanese",
    "lebensform": "lebensformen",
    "legionella": "legionellae",
    "legomenon": "legomena",
    "legumen": "legumina",
    "leishmania": "leishmaniae",
    "leman": "lemans",
    "lemma": "lemmata",
    "lemniscus": "lemnisci",
    "lemur": "lemures",
    "lens": "lenses",
    "lentigo": "lentigines",
    "lentiscus": "lentisci",
    "leonese": "leonese",
    "leproma": "lepromata",
    "leprosarium": "leprosaria",
    "leptomonas": "leptomonas",
    "lepton": "lepta",
    "leptospira": "leptospirae",
    "lesotho": "lesothos",
    "leu": "lei",
    "lex humana": "leges humanae",
    "libero": "liberos",
    "libido": "libidos",
    "libretto": "libretti",
    "lido": "lidos",
    "lied": "lieder",
    "lieu": "lieux",
    "lieutenant general": "lieutenant generals",
    "life": "lives",
    "ligamentum": "ligamenta",
    "likuta": "makuta",
    "lilo": "lilos",
    "liman": "limans",
    "limax": "limaces",
    "limbo": "limbos",
    "limes": "limites",
    "limo": "limos",
    "limulus": "limuli",
    "linctus": "linctuses",
    "line out": "lines out",
    "line-out": "lines-out",
    "linea": "lineae",
    "lineno": "linenos",
    "lingo": "lingos",
    "lingula": "lingulae",
    "lino": "linos",
    "lipoma": "lipomata",
    "lis": "lis",
    "listerella": "listerellae",
    "listeria": "listeria",
    "lithodomus": "lithodomi",
    "lithophyton": "lithophyta",
    "livedo": "livedos",
    "lixivium": "lixivia",
    "loa": "loa",
    "loaf": "loaves",
    "lobulus": "lobuli",
    "lochus": "lochi",
    "loco": "locos",
    "loculus": "loculi",
    "locus": "loci",
    "loggia": "loggie",
    "logion": "logia",
    "logo": "logos",
    "logodaedalus": "logodaedale",
    "loma": "lomata",
    "lomentum": "lomenta",
    "londonese": "londonese",
    "lord lieutenant": "lords lieutenant",
    "lord-lieutenant": "lords-lieutenant",
    "lorrainese": "lorrainese",
    "lorum": "lora",
    "louis": "louis",
    "louse": "lice",
    "lowlife": "lowlifes",
    "lox": "lox",
    "lucchese": "lucchese",
    "lumbago": "lumbagos",
    "lumbricus": "lumbrici",
    "lumen": "lumina",
    "lustrum": "lustra",
    "lymphangioma": "lymphangiomata",
    "lymphoma": "lymphomata",
    "lyonese": "lyonese",
    "lyssa": "lyssae",
    "maar": "maare",
    "macacus": "macaci",
    "macanese": "macanese",
    "macaroni": "macaronies",
    "macassarese": "macassarese",
    "macho": "machos",
    "mackerel": "mackerel",
    "macro": "macros",
    "macula": "maculae",
    "madurese": "madurese",
    "maecenas": "maecenates",
    "maestro": "maestri",
    "mafiosa": "mafiose",
    "mafioso": "mafiosi",
    "magilus": "magili",
    "magma": "magmata",
    "magneto": "magnetos",
    "magnifico": "magnificos",
    "magnum opus": "magna opera",
    "magus": "magi",
    "major general": "major generals",
    "majuro": "majuros",
    "malabarese": "malabarese",
    "malabo": "malabos",
    "malagma": "malagmata",
    "malleolus": "malleoli",
    "malleus": "mallei",
    "maltese": "maltese",
    "malum in se": "mala in se",
    "man": "men",
    "man child": "men children",
    "man midwife": "men midwives",
    "man milliner": "men milliners",
    "man servant": "men servants",
    "man-child": "men-children",
    "man-midwife": "men-midwives",
    "man-milliner": "men-milliners",
    "man-servant": "men-servants",
    "manchild": "menchildren",
    "mango": "mangoes",
    "manifesto": "manifestos",
    "manservant": "menservants",
    "manteau": "manteaux",
    "mantelletta": "mantellette",
    "mantis": "mantises",
    "manubrium": "manubria",
    "maputo": "maputos",
    "maquereau": "maquereaux",
    "maracaibo": "maracaibos",
    "mare": "maria",
    "maremma": "maremme",
    "marquis": "marquises",
    "mary": "maries",
    "mater lectionis": "matres lectionis",
    "math": "maths",
    "matin": "matins",
    "matinee musicale": "matinees musicales",
    "matzah": "matzot",
    "mausoleum": "mausolea",
    "maxilla": "maxillae",
    "maximum": "maxima",
    "me": "us",
    "measles": "measles",
    "meatus": "meatus",
    "mediastinum": "mediastina",
    "mediatrix": "mediatrices",
    "medico": "medicoes",
    "medium": "media",
    "medulloblastoma": "medulloblastomata",
    "medusa": "medusae",
    "megalosaurus": "megalosauri",
    "megatherium": "megatheria",
    "melanoma": "melanomata",
    "melastoma": "melastomae",
    "meliceris": "melicerides",
    "melisma": "melismata",
    "memento": "mementoes",
    "memo": "memos",
    "memorandum": "memoranda",
    "menage a quatre": "menages a quatre",
    "menage a trois": "menages a trois",
    "meningioma": "meningiomata",
    "meniscus": "menisci",
    "menu": "menus",
    "meretrix": "meretrices",
    "meris": "merides",
    "meshumad": "meshumadim",
    "meshummad": "meshummadim",
    "mesothelioma": "mesotheliomata",
    "mesothelium": "mesothelia",
    "messa di voce": "messe di voce",
    "messinese": "messinese",
    "meta": "metae",
    "metalimnion": "metalimnia",
    "metapleuron": "metapleura",
    "metatarsus": "metatarsi",
    "metro": "metros",
    "metropolis": "metropolises",
    "mew": "mews",
    "mews": "mews",
    "mexico": "mexicos",
    "mezuza": "mezuzoth",
    "mezuzah": "mezuzoth",
    "mezzo rilievo": "mezzo rilievos",
    "mezzo-rilievo": "mezzo-rilievos",
    "miasma": "miasmata",
    "micro": "micros",
    "microbion": "microbia",
    "microcephalus": "microcephali",
    "micrococcus": "micrococci",
    "microfiche": "microfiche",
    "micronucleus": "micronuclei",
    "milanese": "milanese",
    "milano": "milanos",
    "miles gloriosus": "milites gloriosi",
    "milieu": "milieux",
    "milieu interieur": "milieux interieurs",
    "miliola": "miliolae",
    "millennium": "millennia",
    "mimosa": "mimosae",
    "mine": "ours",
    "minimum": "minima",
    "minnelied": "minnelieder",
    "minutia": "minutiae",
    "minyan": "minyanim",
    "miracidium": "miracidia",
    "mithraeum": "mithraea",
    "mitochondrion": "mitochondria",
    "mixolimnion": "mixolimnia",
    "mo": "moes",
    "moaks": "moakses",
    "modelletto": "modelletti",
    "modello": "modelli",
    "modenese": "modenese",
    "modernus": "moderni",
    "modius": "modii",
    "modulus": "moduli",
    "modus": "modi",
    "moe": "moes",
    "molimen": "molimina",
    "momentum": "momenta",
    "monaco": "monacos",
    "monas": "monades",
    "mondo": "mondi",
    "money": "monies",
    "mongoose": "mongooses",
    "monimolimnion": "monimolimnia",
    "mono": "monos",
    "monochasium": "monochasia",
    "monorchis": "monorchides",
    "monosyllabon": "monosyllaba",
    "monota": "monotae",
    "monoxylon": "monoxyla",
    "mons": "montes",
    "monseigneur": "messeigneurs",
    "monsignor": "monsignori",
    "monsignore": "monsignori",
    "monstre sacre": "monstres sacres",
    "montenegro": "montenegros",
    "moose": "moose",
    "moratorium": "moratoria",
    "morceau": "morceaux",
    "morocco": "moroccos",
    "moroso": "morosi",
    "mortadella": "mortadelle",
    "mortarium": "mortaria",
    "moscardino": "moscardini",
    "motoscafo": "motoscafi",
    "motto": "mottoes",
    "mouse": "mice",
    "mousetrap": "micetraps",
    "mouth": "mouths",
    "mucro": "mucrones",
    "mukim": "mukim",
    "multigravida": "multigravidae",
    "mumps": "mumps",
    "municipium": "municipia",
    "muqdisho": "muqdishos",
    "muranese": "muranese",
    "murex": "murices",
    "musa": "musae",
    "musee imaginaire": "musees imaginaires",
    "musique concrete": "musiques concretes",
    "mustachio": "mustachios",
    "muzz": "muzzes",
    "mya": "myae",
    "mycoplasma": "mycoplasmata",
    "mycorhiza": "mycorhizae",
    "mycorrhiza": "mycorrhizae",
    "myeloma": "myelomata",
    "myo": "myos",
    "myself": "ourselves",
    "mythos": "mythoi",
    "myxoma": "myxomata",
    "nachtlokal": "nachtlokale",
    "nagaika": "nagaiki",
    "nakayaman": "nakayamans",
    "nankingese": "nankingese",
    "nasutus": "nasuti",
    "naumachia": "naumachiae",
    "nauplius": "nauplii",
    "nautilus": "nautili",
    "navarrese": "navarrese",
    "naya paisa": "naye paise",
    "nebenkern": "nebenkerne",
    "nebula": "nebulae",
    "nectocalyx": "nectocalyces",
    "nemathecium": "nemathecia",
    "nenets": "nentsy",
    "nepalese": "nepalese",
    "nephridium": "nephridia",
    "nerita": "neritae",
    "neuma": "neumata",
    "neurinoma": "neurinomata",
    "neuroma": "neuromata",
    "neurula": "neurulae",
    "neutrino": "neutrinos",
    "news": "news",
    "nexus": "nexus",
    "ngwee": "ngwee",
    "niasese": "niasese",
    "nicobarese": "nicobarese",
    "nidus": "nidi",
    "niello": "nielli",
    "nimbus": "nimbi",
    "ningbo": "ningbos",
    "nipponese": "nipponese",
    "nockerl": "nockerln",
    "noctiluca": "noctilucae",
    "nodum": "noda",
    "nodus": "nodi",
    "noema": "noemata",
    "nomisma": "nomismata",
    "non-ens": "nonentia",
    "norman": "normans",
    "noumenon": "noumena",
    "nouveau pauvre": "nouveaux pauvres",
    "nouveau riche": "nouveaux riches",
    "nouvel arrive": "nouveaux arrives",
    "nova": "novae",
    "novus homo": "novi homines",
    "noxa": "noxae",
    "nubecula": "nubeculae",
    "nucleolonema": "nucleolonemata",
    "nucleolus": "nucleoli",
    "nucleus": "nuclei",
    "numen": "numina",
    "nuntius": "nuntii",
    "nympha": "nymphae",
    "nymphaeum": "nymphaea",
    "nympheum": "nymphaea",
    "oath": "oaths",
    "obelus": "obeli",
    "obolus": "oboli",
    "occiput": "occipita",
    "oceanarium": "oceanaria",
    "ocellus": "ocelli",
    "octavo": "octavos",
    "octohedron": "octohedra",
    "octopus": "octopuses",
    "oculus": "oculi",
    "odeum": "odea",
    "oedema": "oedemata",
    "offspring": "offspring",
    "ogle": "oglys",
    "oklahoman": "oklahomans",
    "olfactorium": "olfactoria",
    "omentum": "omenta",
    "ommatidium": "ommatidia",
    "onager": "onagri",
    "one": "some",
    "operculum": "opercula",
    "ophiophagus": "ophiophagi",
    "optimum": "optima",
    "opus": "opuses",
    "opus magnum": "opera magna",
    "opusculum": "opuscula",
    "oratrix": "oratrices",
    "oregano": "oreganos",
    "organon": "organa",
    "orinoco": "orinocos",
    "orlando": "orlandos",
    "orthoceras": "orthocerata",
    "osculum": "oscula",
    "oslo": "oslos",
    "osmaterium": "osmateria",
    "osmeterium": "osmeteria",
    "ossiculum": "ossicula",
    "osteoclastoma": "osteoclastomata",
    "osteoma": "osteomata",
    "ostium": "ostia",
    "oto": "otos",
    "otomi": "otomi",
    "ottoman": "ottomans",
    "ovarium": "ovaria",
    "ovolo": "ovoli",
    "ovulum": "ovula",
    "ovum": "ova",
    "ox": "oxen",
    "oxybaphon": "oxybapha",
    "oxymoron": "oxymora",
    "oxyuris": "oxyurides",
    "paha": "paha",
    "paisa": "paise",
    "pajamas": "pajamas",
    "pala": "palae",
    "palea": "paleae",
    "pallium": "pallia",
    "palpebra": "palpebrae",
    "palpus": "palpi",
    "palus": "pali",
    "panaman": "panamans",
    "pangasinan": "pangasinanes",
    "panino": "panini",
    "panorpa": "panorpae",
    "pant": "pants",
    "panto": "pantos",
    "pants": "pants",
    "paparrazo": "paparrazi",
    "papilla": "papillae",
    "papilloma": "papillomata",
    "papula": "papulae",
    "papyrus": "papyri",
    "parabola": "parabolae",
    "parabolanus": "parabolani",
    "parador": "paradores",
    "paraganglioma": "paragangliomata",
    "paraganglion": "paraganglia",
    "paraglossa": "paraglossae",
    "paramaribo": "paramaribos",
    "paranucleus": "paranuclei",
    "parapleura": "parapleura",
    "parapleurum": "parapleura",
    "parapsis": "parapsides",
    "paraselene": "paraselenae",
    "parc ferme": "parcs fermes",
    "pardusco": "parduscos",
    "parenchyma": "parenchymata",
    "parergon": "parerga",
    "parhelion": "parhelia",
    "paries": "parietes",
    "parison": "parisa",
    "parmese": "parmese",
    "parnas": "parnassim",
    "parnass": "parnassim",
    "parvenu": "parvenus",
    "pasteurella": "pasteurellae",
    "pastorale": "pastorali",
    "patagium": "patagia",
    "patera": "paterae",
    "path": "paths",
    "pathos": "pathoses",
    "paxilla": "paxillae",
    "pecia": "peciae",
    "pecten": "pectines",
    "pectus": "pectora",
    "pedalo": "pedalos",
    "pedicellaria": "pedicellariae",
    "pediluvium": "pediluvia",
    "pedion": "pedia",
    "pekingese": "pekingese",
    "pelta": "peltae",
    "pelvis": "pelves",
    "pence": "pence",
    "pendulum": "pendula",
    "pengo": "pengoes",
    "penicillus": "penicilli",
    "peninsula": "peninsulae",
    "penis": "penes",
    "penni": "pennia",
    "penny": "pence",
    "pensiero": "pensieri",
    "perceptum": "percepta",
    "peridium": "peridia",
    "perikaryon": "perikarya",
    "perithecium": "perithecia",
    "pernio": "perniones",
    "peronium": "peronia",
    "perradius": "perradii",
    "person": "persons",
    "persona": "personae",
    "pes": "pedes",
    "phalangium": "phalangia",
    "phalanx": "phalanges",
    "phallos": "phalloi",
    "phallus": "phalli",
    "pharmakos": "pharmakoi",
    "phenomenon": "phenomena",
    "phlegmasia": "phlegmasiae",
    "phoca": "phocae",
    "pholas": "pholades",
    "photo": "photos",
    "phototaxis": "phototaxes",
    "phragma": "phragmata",
    "phren": "phrenes",
    "phyle": "phylae",
    "phyllobranchia": "phyllobranchiae",
    "phylum": "phyla",
    "phyma": "phymata",
    "phyogemmarium": "phyogemmaria",
    "piano": "piani",
    "pickerel": "pickerel",
    "pied a terre": "pieds a terre",
    "pied noir": "pieds noirs",
    "piedmontese": "piedmontese",
    "pietra": "pietre",
    "pifferaro": "pifferari",
    "pike": "pike",
    "pileus": "pilei",
    "pillie": "pilleis",
    "pilus": "pili",
    "pimento": "pimentos",
    "pina colada": "pinas coladas",
    "pinax": "pinaces",
    "pincer": "pincers",
    "pincers": "pincers",
    "pinetum": "pineta",
    "pingo": "pingoes",
    "pinna": "pinnae",
    "pinon": "pinones",
    "pinto": "pintos",
    "pipkrake": "pipkraker",
    "piscina": "piscinae",
    "pistillidium": "pistillidia",
    "pistoiese": "pistoiese",
    "pithos": "pithoi",
    "pizza": "pizze",
    "placebo": "placeboes",
    "placitum": "placita",
    "planctus": "planctus",
    "plankter": "plankton",
    "planula": "planulae",
    "plasmalemma": "plasmalemmae",
    "plasmodium": "plasmodia",
    "plasmolyticum": "plasmolytica",
    "plasmoma": "plasmomata",
    "plateau": "plateaux",
    "platypus": "platypodes",
    "pleco": "plecos",
    "plectrum": "plectra",
    "plenarium": "plenaria",
    "plesiosaurus": "plesiosauri",
    "plethron": "plether",
    "pleura": "pleura",
    "pleuron": "pleura",
    "pleurum": "pleura",
    "plexus": "plexus",
    "plica": "plicae",
    "pliers": "pliers",
    "ploshchadka": "ploshchadki",
    "plurale tantum": "pluralia tantum",
    "pluteus": "plutei",
    "pluto": "plutos",
    "pneumothorax": "pneumothoraces",
    "po": "poes",
    "poblador": "pobladores",
    "poculum": "pocula",
    "podere": "poderi",
    "podetium": "podetia",
    "podion": "podia",
    "podium": "podia",
    "poe": "poes",
    "pogo": "pogos",
    "poise": "poise",
    "polis": "polises",
    "politbureau": "politbureaux",
    "politico": "politicoes",
    "pollex": "pollices",
    "pollinarium": "pollinaria",
    "pollinium": "pollinia",
    "polo": "polos",
    "polonese": "polonese",
    "polythecium": "polythecia",
    "pomme de terre": "pommes de terre",
    "poncho": "ponchos",
    "pontifex": "pontifices",
    "porcino": "porcini",
    "porion": "poria",
    "porron": "porrones",
    "port a beul": "puirt a beul",
    "port de bras": "ports de bras",
    "port-a-beul": "puirt-a-beul",
    "portico": "porticoes",
    "portiforium": "portiforia",
    "portmanteau": "portmanteaux",
    "porto": "portos",
    "porto novo": "porto novos",
    "porto-novo": "porto-novos",
    "portrait parle": "portraits parles",
    "portuguese": "portuguese",
    "postnatus": "postnati",
    "postpubis": "postpubes",
    "postulatum": "postulata",
    "potato": "potatoes",
    "potestas": "potestates",
    "pox": "pox",
    "praecognitum": "praecognita",
    "pragma": "pragmata",
    "precis": "precis",
    "precognitum": "precognita",
    "premier danseur": "premiers danseurs",
    "prepenna": "prepennae",
    "prima ballerina": "prime ballerine",
    "prima donna": "prime donne",
    "primigravida": "primigravidae",
    "primordium": "primordia",
    "principium": "principia",
    "pro": "pros",
    "proavis": "proaves",
    "proboscis": "proboscides",
    "proceedings": "proceedings",
    "procoelia": "procoeliae",
    "prodromus": "prodromi",
    "proferens": "proferentes",
    "profluvium": "profluvia",
    "proglottis": "proglottides",
    "programma": "programmata",
    "progymnasium": "progymnasia",
    "progymnasma": "progymnasmata",
    "prokaryon": "prokarya",
    "prolabium": "prolabia",
    "prolegomenon": "prolegomena",
    "promerops": "promeropes",
    "propleuron": "propleura",
    "propodium": "propodia",
    "propositus": "propositi",
    "proprium": "propria",
    "propylaeum": "propylaea",
    "propylon": "propyla",
    "prora": "prorae",
    "prosa": "prosae",
    "proscenium": "proscenia",
    "proscolla": "proscollae",
    "prosecutrix": "prosecutrices",
    "prosencephalon": "prosencephala",
    "proseucha": "proseuchae",
    "prosodiencephalon": "prosodiencephala",
    "prosopium": "prosopia",
    "prospectus": "prospectus",
    "prosula": "prosulae",
    "protarsus": "protarsi",
    "prothallium": "prothallia",
    "prothallus": "prothalli",
    "protococcus": "protococci",
    "protoma": "protomae",
    "protonema": "protonemata",
    "prototypon": "prototypa",
    "protypon": "protypa",
    "proviso": "provisoes",
    "prytanis": "prytan",
    "psammoma": "psammomata",
    "pseudopodium": "pseudopodia",
    "pseudostoma": "pseudostomata",
    "psycho": "psychos",
    "psydracium": "psydracia",
    "pteroma": "pteromata",
    "pteropus": "pteropi",
    "pteryla": "pterylae",
    "pueblo": "pueblos",
    "pul": "puli",
    "pullus": "pulli",
    "pulmonarium": "pulmonaria",
    "pulsellum": "pulsella",
    "pulveratrix": "pulveratrices",
    "punctum": "puncta",
    "punto blanco": "punti blanchi",
    "pupa": "pupae",
    "purlieu": "purlieux",
    "purum": "purum",
    "putto": "putti",
    "puttony": "puttonys",
    "pycnidium": "pycnidia",
    "pycnis": "pycnides",
    "pycnium": "pycnia",
    "pyjamas": "pyjamas",
    "pyla": "pylae",
    "pyralis": "pyralides",
    "pyramidion": "pyramidia",
    "pyrexia": "pyrexiae",
    "pyrites": "pirrites",
    "pyrophorus": "pyrophori",
    "pyxidium": "pyxidia",
    "quadratrix": "quadratrices",
    "quaesitum": "quaesita",
    "quantum": "quanta",
    "quarto": "quartos",
    "quartz": "quartzes",
    "quid": "quid",
    "quinarius": "quinarii",
    "quinquennium": "quinquennia",
    "quito": "quitos",
    "rabies": "rabies",
    "rad": "rad",
    "radiale": "radialia",
    "radius": "radii",
    "radix": "radices",
    "ragazza": "ragazze",
    "ragazzo": "ragazzi",
    "raggare": "raggare",
    "ramulus": "ramuli",
    "ramus": "rami",
    "ranunculus": "ranunculi",
    "rara avis": "rarae aves",
    "razz": "razzes",
    "rectus": "recti",
    "redia": "rediae",
    "reductio": "reductiones",
    "referendum": "referenda",
    "refugium": "refugia",
    "regidor": "regidores",
    "regma": "regmata",
    "regulus": "reguli",
    "reich": "reiche",
    "rejon": "rejones",
    "rejoneador": "rejoneadores",
    "relatum": "relata",
    "rem": "rem",
    "remex": "remiges",
    "remous": "remou",
    "replum": "repla",
    "representamen": "representamina",
    "res": "res",
    "residuum": "residua",
    "responsum": "responsa",
    "rete": "retia",
    "retinaculum": "retinacula",
    "retinula": "retinulae",
    "rezident": "rezidenty",
    "rezidentsia": "rezidentsii",
    "rhino": "rhinos",
    "rhinoceros": "rhinoceros",
    "rhinophyma": "rhinophymata",
    "rhizoma": "rhizomata",
    "rhododendron": "rhododendra",
    "rhombohedron": "rhombohedra",
    "rhomboideus": "rhomboidei",
    "rhopalium": "rhopalia",
    "rhynchodaeum": "rhynchodaea",
    "rhythmus": "rhythmi",
    "ri": "ri",
    "ricercar": "ricercari",
    "ricercata": "ricercate",
    "ricordo": "ricordi",
    "rideau": "rideaux",
    "rilievo": "rilievos",
    "rimon": "rimmonim",
    "rione": "rioni",
    "risotto": "risottos",
    "ritardando": "ritardandi",
    "rite de passage": "rites de passage",
    "rococo": "rococos",
    "roe": "roe",
    "rom": "roma",
    "romagnese": "romagnese",
    "roman": "romans",
    "romanese": "romanese",
    "romany": "romanies",
    "romanze": "romanzen",
    "rond de jambe": "ronds de jambes",
    "rondeau": "rondeaux",
    "rondo": "rondos",
    "rostrum": "rostra",
    "rotula": "rotulae",
    "route nationale": "routes nationales",
    "rubashka": "rubashki",
    "ruga": "rugae",
    "ruse de guerre": "ruses de guerre",
    "rype": "ryper",
    "sabertooth": "sabertooths",
    "sacculus": "sacculi",
    "sacellum": "sacella",
    "sachverhalt": "sachverhalte",
    "sacramento": "sacramentos",
    "sacrarium": "sacraria",
    "sacrum": "sacra",
    "saddo": "saddos",
    "saffo": "saffi",
    "sago": "sagos",
    "sagum": "saga",
    "salita": "salite",
    "salivarium": "salivaria",
    "salmon": "salmon",
    "salpa": "salpae",
    "salvo": "salvos",
    "sambo": "samboes",
    "samurai": "samurai",
    "sanatorium": "sanatoria",
    "sanctum sanctorum": "sancta sanctorum",
    "sandr": "sandr",
    "sangirese": "sangirese",
    "santiago": "santiagos",
    "sapporo": "sapporos",
    "sarajevo": "sarajevos",
    "sarawakese": "sarawakese",
    "sarcina": "sarcinae",
    "sarcoma": "sarcomata",
    "sarcophagus": "sarcophagi",
    "sarissa": "sarissae",
    "sassafras": "sassafrases",
    "satelles": "satellites",
    "saxicava": "saxicavae",
    "sbirro": "sbirri",
    "scalenus": "scaleni",
    "scandalum magnatum": "scandala magnatum",
    "scapula": "scapulae",
    "scapus": "scapi",
    "scarabaeus": "scarabaei",
    "scarus": "scari",
    "scazon": "scazons",
    "sceat": "sceattas",
    "schema": "schemata",
    "scherzando": "scherzandos",
    "scherzo": "scherzos",
    "schisma": "schismata",
    "schistosomulum": "schistosomula",
    "schizostylis": "schizostylis",
    "scholium": "scholia",
    "sciolus": "scioli",
    "scirrhus": "scirrhi",
    "scissors": "scissors",
    "scolex": "scoleces",
    "scolopidium": "scolopidia",
    "scomber": "scombri",
    "scoria": "scoriae",
    "scotoma": "scotomata",
    "scriptorium": "scriptoria",
    "scudo": "scudi",
    "scungille": "scungilli",
    "scuola": "scuole",
    "scutellum": "scutella",
    "scutulum": "scutula",
    "scutum": "scuta",
    "scyphus": "scyphi",
    "seaquarium": "seaquaria",
    "secans": "secantes",
    "secretum": "secreta",
    "secundagravida": "secundagravidae",
    "secundigravida": "secundigravidae",
    "secundipara": "secundiparae",
    "sedes": "sedes",
    "sedile": "sedilia",
    "seecatch": "seecatchie",
    "seed": "seed",
    "selman": "selmans",
    "sema": "semata",
    "seminium": "seminia",
    "semuncia": "semunciae",
    "senarius": "senarii",
    "senegalese": "senegalese",
    "senex": "senes",
    "sensedatum": "sensedata",
    "sensillum": "sensilla",
    "sensoriolum": "sensoriola",
    "sensum": "sensa",
    "sententia": "sententiae",
    "separatrix": "separatrices",
    "sephardi": "sephardin",
    "septarium": "septaria",
    "septemvir": "septemviri",
    "septenarius": "septenarii",
    "septum": "septa",
    "sequela": "sequelae",
    "sequestrum": "sequestra",
    "seraph": "seraphim",
    "sericterium": "sericteria",
    "series": "series",
    "serir": "serir",
    "serpigo": "serpigines",
    "serpula": "serpulae",
    "serra": "serrae",
    "sertularia": "sertulariae",
    "serum": "sera",
    "sestertium": "sestertia",
    "sestiere": "sestieri",
    "set off": "sets off",
    "set out": "sets out",
    "set to": "sets to",
    "set-off": "sets-off",
    "set-out": "sets-out",
    "set-to": "sets-to",
    "seta": "setae",
    "sezession": "sezessionen",
    "sforzato": "sforzati",
    "sgraffiato": "sgraffiati",
    "shad": "shad",
    "shaman": "shamans",
    "shauri": "shauries",
    "shavese": "shavese",
    "shawabti": "shawabtiu",
    "shawanese": "shawanese",
    "she": "they",
    "shear": "shears",
    "shears": "shears",
    "sheep": "sheep",
    "shelf": "shelves",
    "shifta": "shifta",
    "shochet": "shochetim",
    "shomer": "shomrim",
    "short": "shorts",
    "shorts": "shorts",
    "siamese": "siamese",
    "siemens": "siemens",
    "sienese": "sienese",
    "siglos": "sigloi",
    "signum": "signa",
    "sikkimese": "sikkimese",
    "silbador": "silbadores",
    "silenus": "sileni",
    "silex": "silices",
    "siliqua": "siliquae",
    "silo": "silos",
    "silurus": "siluri",
    "simia": "simiae",
    "simplex": "simplices",
    "simpulum": "simpula",
    "simulacrum": "simulacra",
    "sinfonia": "sinfonie",
    "singhalese": "singhalese",
    "sinhalese": "sinhalese",
    "sinus": "sinus",
    "siphunculus": "siphunculi",
    "sipunculus": "sipunculi",
    "sirocco": "siroccos",
    "sistrum": "sistra",
    "situla": "situlae",
    "sizz": "sizzes",
    "ski": "skis",
    "snekkja": "snekkjur",
    "snipe": "snipe",
    "snooze": "snoozes",
    "so": "soes",
    "soe": "soes",
    "sogdianese": "sogdianese",
    "sol": "soles",
    "solarium": "solaria",
    "soldo": "soldi",
    "solenium": "solenia",
    "solfeggio": "solfeggi",
    "solidus": "solidos",
    "solo": "soli",
    "solstitium": "solstitia",
    "solvus": "solvi",
    "soma": "somata",
    "somaten": "somatenes",
    "sombrero": "sombreros",
    "sonaman": "sonamans",
    "soprano": "soprani",
    "soralium": "soralia",
    "soredium": "soredia",
    "sorus": "sori",
    "soshi": "soshi",
    "sotch": "sotchs",
    "souvlaki": "souvlakia",
    "spadix": "spadices",
    "spaetlese": "spaetlesen",
    "sparus": "spari",
    "spasmus": "spasmi",
    "spatha": "spathae",
    "species": "species",
    "spectrum": "spectra",
    "speculum": "specula",
    "sperma": "spermata",
    "spermarium": "spermaria",
    "spermatheca": "spermathecae",
    "spermatium": "spermatia",
    "spermatoon": "spermatoa",
    "spermogonium": "spermogonia",
    "sphaeridium": "sphaerideum",
    "sphagnum": "sphagna",
    "sphex": "spheges",
    "sphinx": "sphinges",
    "spiculum": "spicula",
    "spiraculum": "spiracula",
    "spirillum": "spirilla",
    "spondylus": "spondyli",
    "sporodochium": "sporodochia",
    "sputum": "sputa",
    "squadra": "squadre",
    "squadrist": "squadristi",
    "squalus": "squali",
    "squama": "squamae",
    "squilla": "squillae",
    "stabilimentum": "stabilimenta",
    "staccato": "staccatos",
    "stadium": "stadia",
    "stalactite": "stalactitae",
    "stamen": "stamina",
    "staminidium": "staminidia",
    "staphylococcus": "staphylococci",
    "star fish": "star fish",
    "star-fish": "star-fish",
    "starets": "startsy",
    "staretz": "startzy",
    "starfish": "starfish",
    "starosta": "starosti",
    "starover": "starovery",
    "stasimon": "stasima",
    "statera": "staterae",
    "statua": "statuae",
    "status": "status",
    "stela": "stelae",
    "stella": "stellae",
    "stemma": "stemmata",
    "stephanion": "stephania",
    "sterigma": "sterigmata",
    "sternebra": "sternebrae",
    "sterno": "sternos",
    "sternum": "sterna",
    "stichidium": "stichidia",
    "stichos": "stichoi",
    "stigma": "stigmata",
    "stigmaria": "stigmariae",
    "stiletto": "stilettoes",
    "stillicidium": "stillicidia",
    "stimulus": "stimuli",
    "stipes": "stipites",
    "stipula": "stipulae",
    "stirps": "stirpes",
    "stoa": "stoai",
    "stolo": "stolones",
    "stoma": "stomata",
    "stomach": "stomachs",
    "stomatium": "stomatia",
    "stomion": "stomia",
    "stomodaeum": "stomodaea",
    "storey": "storeys",
    "stornello": "stornelli",
    "strambotto": "strambotti",
    "strategos": "strategoi",
    "strategus": "strategi",
    "stratum": "strata",
    "streptococcus": "streptococci",
    "streptothrix": "streptothrices",
    "stria": "striae",
    "striatum": "striata",
    "striga": "strigae",
    "striola": "striolae",
    "strobila": "strobilae",
    "strobilus": "strobili",
    "stroma": "stromata",
    "stromodeum": "stromodaea",
    "struma": "strumae",
    "stryddag": "stryddae",
    "stucco": "stuccos",
    "studium generale": "studia generali",
    "stylo": "stylos",
    "stylopodium": "stylopodia",
    "stylus": "styli",
    "subbotnik": "subbotniki",
    "subsellium": "subsellia",
    "subsidium": "subsidia",
    "substratum": "substrata",
    "succinea": "succineae",
    "succubus": "succubi",
    "succus": "succi",
    "sudanese": "sudanese",
    "suffrutex": "suffrutices",
    "suggestio falsi": "suggestiones falsi",
    "suggestum": "suggesta",
    "sulcus": "sulci",
    "sum total": "sums total",
    "sum-total": "sums-total",
    "summa": "summae",
    "summum bonum": "summa bona",
    "summum genus": "summa genera",
    "sumo": "sumos",
    "sundanese": "sundanese",
    "supercilium": "supercilia",
    "superficies": "superficies",
    "supernova": "supernovae",
    "superstratum": "superstrata",
    "suppositum": "supposita",
    "supracleithrum": "supracleithra",
    "suprascapula": "suprascapulae",
    "surculus": "surculi",
    "surimono": "surimonos",
    "sustentaculum": "sustentacula",
    "swahilese": "swahilese",
    "swine": "swines",
    "switchfoot": "switchfoots",
    "syllabarium": "syllabaria",
    "syllabus": "syllabi",
    "symplasma": "symplasmata",
    "sympodium": "sympodia",
    "symposium": "symposia",
    "synaesthesia": "synaesthesiae",
    "synangium": "synangia",
    "synapticula": "synapticulae",
    "synaxarion": "synaxaria",
    "synaxarium": "synaxaria",
    "synaxis": "synaxes",
    "syncellus": "syncelli",
    "syncytium": "syncytia",
    "syndendrium": "syndendria",
    "synechia": "synechiae",
    "synedrion": "synedria",
    "synedrium": "synedria",
    "synesthesia": "synesthesiae",
    "syntagma": "syntagmata",
    "syntaxis": "syntaxes",
    "synthronus": "synthroni",
    "synusia": "synusiae",
    "syphiloma": "syphilomata",
    "syringe": "syringes",
    "syrinx": "syringes",
    "syrphus": "syrphi",
    "syrtis": "syrtes",
    "tabella": "tabellae",
    "tableau": "tableaux",
    "tablinum": "tablina",
    "tabula": "tabulae",
    "tacoman": "tacomans",
    "taenia": "taeniae",
    "tagma": "tagmata",
    "taiko": "taikos",
    "talea": "taleae",
    "talisman": "talismans",
    "talouse": "talouses",
    "talus": "tali",
    "tango": "tangoes",
    "tarantata": "tarantate",
    "tarantato": "tarantati",
    "tarsus": "tarsi",
    "tax": "taxes",
    "taxis": "taxes",
    "taxon": "taxa",
    "tazza": "tazze",
    "teal": "teal",
    "techno": "technos",
    "tegmen": "tegmina",
    "tegmentum": "tegmenta",
    "tegula": "tegulae",
    "telamon": "telamones",
    "telium": "telia",
    "temblor": "temblores",
    "tempietto": "tempietti",
    "tempo": "tempi",
    "tenaculum": "tenacula",
    "tenderfoot": "tenderfoots",
    "tenggerese": "tenggerese",
    "tenia": "teniae",
    "tenorino": "tenorini",
    "tentaculum": "tentacula",
    "tentamen": "tentamina",
    "tenuis": "tenues",
    "tepidarium": "tepidaria",
    "teratoma": "teratomata",
    "terebellum": "terebella",
    "terebratula": "terebratulae",
    "teredo": "teredines",
    "tergum": "terga",
    "termes": "termites",
    "terminus": "termini",
    "terp": "terpen",
    "terrae filius": "terrae filii",
    "terramara": "terremare",
    "terrarium": "terraria",
    "terrazzo": "terrazzos",
    "territorium": "territoria",
    "terzetto": "terzetti",
    "tessella": "tessellae",
    "tessera": "tesserae",
    "testes": "testes",
    "testis": "testes",
    "testudo": "testudos",
    "tetanothrum": "tetanother",
    "tete a tete": "tete a tetes",
    "tete-a-tete": "tete-a-tetes",
    "tetracolon": "tetracola",
    "tetragrammaton": "tetragrammata",
    "tetrahedron": "tetrahedra",
    "tetrapleuron": "tetrapleura",
    "tetrasporangium": "tetrasporangia",
    "thalamos": "thalamoi",
    "thalamus": "thalami",
    "thanatorium": "thanatoria",
    "that": "those",
    "thaumaturgus": "thaumaturgi",
    "thebe": "thebe",
    "theca": "thecae",
    "theirs": "theirs",
    "thelycum": "thelyca",
    "them": "them",
    "thema": "themata",
    "themself": "themselves",
    "theologoumenon": "theologoumena",
    "thesaurus": "thesauri",
    "thesis": "theses",
    "they": "they",
    "thief": "thieves",
    "this": "these",
    "tholos": "tholoi",
    "tholus": "tholi",
    "thorax": "thoraces",
    "thos": "thoes",
    "thraso": "thrasones",
    "thymiaterion": "thymiateria",
    "thymus": "thymi",
    "thyridium": "thyridia",
    "thyrsus": "thyrsi",
    "tibia": "tibiae",
    "timorese": "timorese",
    "timpano": "timpanos",
    "tintinnabulum": "tintinnabula",
    "tipula": "tipulae",
    "tiro": "tiros",
    "tirolese": "tirolese",
    "to": "toes",
    "tobacco": "tobaccos",
    "tocharese": "tocharese",
    "toe": "toes",
    "togo": "togos",
    "tokyo": "tokyos",
    "tolkach": "tolkachi",
    "tomato": "tomatoes",
    "tombarolo": "tombaroli",
    "tomium": "tomia",
    "tondo": "tondi",
    "tong": "tongs",
    "tongs": "tongs",
    "tonkinese": "tonkinese",
    "tonneau": "tonneaux",
    "tooth": "teeth",
    "topaz": "topazes",
    "topos": "topoi",
    "torero": "toreros",
    "toril": "toriles",
    "torinese": "torinese",
    "torino": "torinos",
    "torma": "torma",
    "tornese": "tornesi",
    "tornus": "torni",
    "toronto": "torontos",
    "torso": "torsos",
    "tortilla espanola": "tortillas espanolas",
    "tortrix": "tortrices",
    "torula": "torulae",
    "torulus": "toruli",
    "torus": "tori",
    "tour de force": "tours de force",
    "toxicum": "toxica",
    "toxius": "toxii",
    "toxoplasma": "toxoplasmata",
    "trabea": "trabeae",
    "trabecula": "trabeculae",
    "trachea": "tracheae",
    "tractatrix": "tractatrices",
    "tractellum": "tractella",
    "tractrix": "tractrices",
    "tragus": "tragi",
    "trapezium": "trapezia",
    "trapezius": "trapezii",
    "trapezohedron": "trapezohedra",
    "trattoria": "trattorie",
    "trauma": "traumata",
    "trellis": "trellises",
    "tremolo": "tremolos",
    "treponema": "treponemata",
    "trichidium": "trichidia",
    "trichina": "trichinae",
    "trichoma": "trichomata",
    "triclinium": "triclinia",
    "triens": "trientes",
    "triforium": "triforia",
    "trilby": "trilbys",
    "triplum": "tripla",
    "triradius": "triradii",
    "triumvir": "triumviri",
    "trochus": "trochi",
    "trockenbeerenauslese": "trockenbeerenauslesen",
    "tropaeolum": "tropaeola",
    "troparion": "troparia",
    "tropis": "tropides",
    "trousers": "trousers",
    "trousseau": "trousseaux",
    "trout": "trout",
    "trullo": "trulli",
    "trumeau": "trumeaux",
    "tsuba": "tsuba",
    "tsubo": "tsubo",
    "tuber": "tuberes",
    "tuberculoma": "tuberculomata",
    "tuberculum": "tubercula",
    "tubulus": "tubuli",
    "tumulus": "tumuli",
    "tuna": "tuna",
    "turbot": "turbot",
    "turf": "turves",
    "turron": "turrones",
    "tympanum": "tympana",
    "typo": "typos",
    "tyro": "tyros",
    "tyrolese": "tyrolese",
    "udarnik": "udarniki",
    "ulpan": "ulpanim",
    "ultimatum": "ultimata",
    "umbilicus": "umbilici",
    "umbo": "umbones",
    "umbra": "umbrae",
    "umwelt": "umwelten",
    "uncia": "unciae",
    "uncinus": "uncini",
    "uncus": "unci",
    "unguis": "ungues",
    "unio": "uniones",
    "uomo": "uomini",
    "uomo universale": "uomini universali",
    "uraeus": "uraei",
    "uredinium": "uredinia",
    "uredium": "uredia",
    "uredosorus": "uredosori",
    "urodaeum": "urodaea",
    "urodeum": "urodea",
    "urus": "uri",
    "usnea": "usneae",
    "ustilago": "ustilagines",
    "uterus": "uteri",
    "utricularia": "utriculariae",
    "uva": "uvae",
    "vaalpens": "vaalpensen",
    "vacuum": "vacua",
    "vagina": "vaginae",
    "vaginula": "vaginulae",
    "vagus": "vagi",
    "vallecula": "valleculae",
    "valuta": "valuten",
    "valvula": "valvulae",
    "vaporetto": "vaporetti",
    "vaquero": "vaqueros",
    "varella": "varely",
    "varia lectio": "variae lectiones",
    "varix": "varices",
    "vas": "vasa",
    "vasculitis": "vasculitides",
    "vasculum": "vascula",
    "veduta": "vedute",
    "veg": "veg",
    "vehiculum": "vehicula",
    "velamen": "velamina",
    "velarium": "velaria",
    "velum": "vela",
    "vena": "venae",
    "venus": "veneres",
    "vermicello": "vermicellos",
    "vermiculus": "vermiculi",
    "vermontese": "vermontese",
    "veronese": "veronese",
    "verruca": "verrucae",
    "vers libre": "vers libres",
    "versiculus": "versiculi",
    "verso": "versos",
    "vertebra": "vertebrae",
    "vertex": "vertices",
    "verticillus": "verticilli",
    "vesicula": "vesiculae",
    "vestigium": "vestigia",
    "vetturino": "vetturini",
    "viale": "viali",
    "viaticum": "viatica",
    "vibex": "vibices",
    "vibraculum": "vibracula",
    "vibrato": "vibratos",
    "vibrio": "vibriones",
    "viennese": "viennese",
    "vietnamese": "vietnamese",
    "vila": "vile",
    "villanella": "villanelle",
    "villino": "villini",
    "villus": "villi",
    "vinculum": "vincula",
    "violetta": "violette",
    "violoncello": "violoncellos",
    "virga": "virgae",
    "virgo": "virgos",
    "virtuoso": "virtuosi",
    "virus": "viruses",
    "vis": "vires",
    "viscus": "viscera",
    "vitellarium": "vitellaria",
    "vitta": "vittae",
    "vivarium": "vivaria",
    "vlakte": "vlakten",
    "voeu": "voeux",
    "volcanello": "volcanelli",
    "volcano": "volcanoes",
    "vomica": "vomicae",
    "vomitorium": "vomitoria",
    "voorbok": "voorbokke",
    "vorspiel": "vorspiele",
    "vorstellung": "vorstellungen",
    "vortex": "vortices",
    "wadi": "wadis",
    "wady": "wadys",
    "waldo": "waldoes",
    "waltz": "waltzes",
    "wanderjahr": "wanderjahre",
    "wandervogel": "wandervogel",
    "washer up": "washers up",
    "washer-up": "washers-up",
    "weirdo": "weirdos",
    "wenchowese": "wenchowese",
    "wharf": "wharves",
    "what": "what",
    "whatever": "whatever",
    "whatsoever": "whatsoever",
    "white staff": "white staves",
    "whiting": "whiting",
    "whitmanese": "whitmanese",
    "who": "who",
    "whoever": "whoever",
    "whom": "whom",
    "whomever": "whomever",
    "whomsoever": "whomsoever",
    "whose": "whose",
    "whosesoever": "whosesoever",
    "whosever": "whosever",
    "whosoever": "whosoever",
    "wife": "wives",
    "wildebeest": "wildebeest",
    "wolf": "wolves",
    "woman child": "women children",
    "woman servant": "women servants",
    "woman-child": "women-children",
    "woman-servant": "women-servants",
    "womanchild": "womenchildren",
    "wreath": "wreaths",
    "xanthoma": "xanthomata",
    "xiphiplastron": "xiphiplastra",
    "xiphopagus": "xiphopagi",
    "xiphoplastron": "xiphoplastra",
    "xoanon": "xoana",
    "xystus": "xysti",
    "yakiman": "yakimans",
    "yamoussoukro": "yamoussoukros",
    "yarraman": "yarramans",
    "yengeese": "yengeese",
    "yes": "yeses",
    "yo": "yoes",
    "yo yo": "yo yos",
    "yo-yo": "yo-yos",
    "yoe": "yoes",
    "yokohaman": "yokohamans",
    "you": "you",
    "yours": "yours",
    "yourself": "yourselves",
    "yuman": "yumans",
    "yunnanese": "yunnanese",
    "yuo": "yuon",
    "zadruga": "zadruge",
    "zeitgeber": "zeitgebers",
    "zemirah": "zemirot",
    "zeppole": "zeppoli",
    "zero": "zeros",
    "zibo": "zibos",
    "zingano": "zingani",
    "zita": "zite",
    "zitella": "zitelle",
    "zizz": "zizzes",
    "zloty": "zloty",
    "zoaea": "zoaeae",
    "zoarium": "zoaria",
    "zoea": "zoeae",
    "zooea": "zooeae",
    "zooecium": "zooecia",
    "zoophorus": "zoophori",
    "zoosporangium": "zoosporangia",
    "zophorus": "zophori",
    "zucca": "zucche",
    "zucchini": "zucchini",
    "zuppa": "zuppe",
    "zuz": "zuzim",
    "zygantrum": "zygantra",
    "zygoma": "zygomata",
    "zygon": "zyga"
}

singular_of = {
    "Achinese": "Achinese",
    "Alfurese": "Alfurese",
    "Amboinese": "Amboinese",
    "Americanese": "Americanese",
    "Amoyese": "Amoyese",
    "Andamanese": "Andamanese",
    "Angolese": "Angolese",
    "Annamese": "Annamese",
    "Antananarivos": "Antananarivo",
    "Aragonese": "Aragonese",
    "Arakanese": "Arakanese",
    "Asherahs": "Asherah",
    "Asherim": "Asherah",
    "Assamese": "Assamese",
    "Atlantes": "Atlas",
    "Auslesen": "Auslese",
    "Avignonese": "Avignonese",
    "Azerbaijanese": "Azerbaijanese",
    "Bahamans": "Bahaman",
    "Balinese": "Balinese",
    "Bamakos": "Bamako",
    "Barquisimetos": "Barquisimeto",
    "Bengalese": "Bengalese",
    "Bernese": "Bernese",
    "Bhutanese": "Bhutanese",
    "Biros": "Biro",
    "Bolognese": "Bolognese",
    "Bolzanos": "Bolzano",
    "Borghese": "Borghese",
    "Bostonese": "Bostonese",
    "Botos": "Boto",
    "Buginese": "Buginese",
    "Burmans": "Burman",
    "Burmese": "Burmese",
    "Cairos": "Cairo",
    "Calabrese": "Calabrese",
    "Camaldolese": "Camaldolese",
    "Cantonese": "Cantonese",
    "Carlylese": "Carlylese",
    "Cassinese": "Cassinese",
    "Celanese": "Celanese",
    "Ceylonese": "Ceylonese",
    "Chicagos": "Chicago",
    "Chimangos": "Chimango",
    "Chinese": "Chinese",
    "Colombos": "Colombo",
    "Colorados": "Colorado",
    "Congoese": "Congoese",
    "Congolese": "Congolese",
    "Czechs": "Czech",
    "Dongolese": "Dongolese",
    "Dracos": "Draco",
    "Esperantos": "Esperanto",
    "Faroese": "Faroese",
    "Faros": "Faro",
    "Ferrarese": "Ferrarese",
    "Filipinos": "Filipino",
    "Foochowese": "Foochowese",
    "Gabunese": "Gabunese",
    "Galla": "Galla",
    "Gallas": "Galla",
    "Genevese": "Genevese",
    "Genoese": "Genoese",
    "Germans": "German",
    "Gestapos": "Gestapo",
    "Gilbertese": "Gilbertese",
    "Goanese": "Goanese",
    "Greensboros": "Greensboro",
    "Guaiaberos": "Guaiabero",
    "Guianese": "Guianese",
    "Hainanese": "Hainanese",
    "Hararese": "Hararese",
    "Harlemese": "Harlemese",
    "Havanese": "Havanese",
    "Heavenese": "Heavenese",
    "Hiroshimans": "Hiroshiman",
    "Hoosierese": "Hoosierese",
    "Hottentotese": "Hottentotese",
    "Hunanese": "Hunanese",
    "Idahos": "Idaho",
    "Indochinese": "Indochinese",
    "Iquicos": "Iquico",
    "Japanese": "Japanese",
    "Javanese": "Javanese",
    "Jerrys": "Jerry",
    "Johnsonese": "Johnsonese",
    "Kakapos": "Kakapo",
    "Kanarese": "Kanarese",
    "Kinkimavos": "Kinkimavo",
    "Kiplingese": "Kiplingese",
    "Kokakos": "Kokako",
    "Kongoese": "Kongoese",
    "Kongolese": "Kongolese",
    "Kosovos": "Kosovo",
    "Lapithae": "Lapith",
    "Lapiths": "Lapith",
    "Lapponese": "Lapponese",
    "Lebanese": "Lebanese",
    "Leonese": "Leonese",
    "Lesothos": "Lesotho",
    "Lilos": "Lilo",
    "Limans": "Liman",
    "Londonese": "Londonese",
    "Lorrainese": "Lorrainese",
    "Lucchese": "Lucchese",
    "Lyonese": "Lyonese",
    "Macanese": "Macanese",
    "Macassarese": "Macassarese",
    "Madurese": "Madurese",
    "Majuros": "Majuro",
    "Malabarese": "Malabarese",
    "Malabos": "Malabo",
    "Maltese": "Maltese",
    "Maputos": "Maputo",
    "Maracaibos": "Maracaibo",
    "Marys": "Mary",
    "Messinese": "Messinese",
    "Mexicos": "Mexico",
    "Milanese": "Milanese",
    "Milanos": "Milano",
    "Modenese": "Modenese",
    "Monacos": "Monaco",
    "Montenegros": "Montenegro",
    "Moroccos": "Morocco",
    "Muqdishos": "Muqdisho",
    "Muranese": "Muranese",
    "Nakayamans": "Nakayaman",
    "Nankingese": "Nankingese",
    "Navarrese": "Navarrese",
    "Nenets": "Nenets",
    "Nentsi": "Nenets",
    "Nepalese": "Nepalese",
    "Niasese": "Niasese",
    "Nicobarese": "Nicobarese",
    "Ningbos": "Ningbo",
    "Nipponese": "Nipponese",
    "Normans": "Norman",
    "Oklahomans": "Oklahoman",
    "Orinocos": "Orinoco",
    "Orlandos": "Orlando",
    "Oslos": "Oslo",
    "Otomi": "Otomi",
    "Otomies": "Otomi",
    "Panamans": "Panaman",
    "Pangasinanes": "Pangasinan",
    "Pangasinans": "Pangasinan",
    "Paramaribos": "Paramaribo",
    "Parduscos": "Pardusco",
    "Parmese": "Parmese",
    "Pekingese": "Pekingese",
    "Piedmontese": "Piedmontese",
    "Pistoiese": "Pistoiese",
    "Plutos": "Pluto",
    "Polonese": "Polonese",
    "Porto Novos": "Porto Novo",
    "Porto-Novos": "Porto-Novo",
    "Portos": "Porto",
    "Portuguese": "Portuguese",
    "Purum": "Purum",
    "Purums": "Purum",
    "Quitos": "Quito",
    "Roma": "Rom",
    "Romagnese": "Romagnese",
    "Romanese": "Romanese",
    "Romanies": "Romany",
    "Romans": "Roman",
    "Sacramentos": "Sacramento",
    "Sangirese": "Sangirese",
    "Santiagos": "Santiago",
    "Sapporos": "Sapporo",
    "Sarajevos": "Sarajevo",
    "Sarawakese": "Sarawakese",
    "Selmans": "Selman",
    "Senegalese": "Senegalese",
    "Shavese": "Shavese",
    "Shawanese": "Shawanese",
    "Siamese": "Siamese",
    "Sienese": "Sienese",
    "Sikkimese": "Sikkimese",
    "Siluri": "Silurus",
    "Singhalese": "Singhalese",
    "Sinhalese": "Sinhalese",
    "Sogdianese": "Sogdianese",
    "Sonamans": "Sonaman",
    "Spaetlesen": "Spaetlese",
    "Sudanese": "Sudanese",
    "Sundanese": "Sundanese",
    "Swahilese": "Swahilese",
    "Tacomans": "Tacoman",
    "Taikos": "Taiko",
    "Tenggerese": "Tenggerese",
    "Timorese": "Timorese",
    "Tirolese": "Tirolese",
    "Tocharese": "Tocharese",
    "Togos": "Togo",
    "Tokyos": "Tokyo",
    "Tonkinese": "Tonkinese",
    "Torinese": "Torinese",
    "Torinos": "Torino",
    "Tornesi": "Tornese",
    "Torontos": "Toronto",
    "Tyrolese": "Tyrolese",
    "Vermontese": "Vermontese",
    "Veronese": "Veronese",
    "Viennese": "Viennese",
    "Vietnamese": "Vietnamese",
    "Virgos": "Virgo",
    "Wenchowese": "Wenchowese",
    "Whitmanese": "Whitmanese",
    "Yakimans": "Yakiman",
    "Yamoussoukros": "Yamoussoukro",
    "Yengeese": "Yengeese",
    "Yokohamans": "Yokohaman",
    "Yumans": "Yuman",
    "Yunnanese": "Yunnanese",
    "Zibos": "Zibo",
    "abacuses": "abacus",
    "aboiteaus": "aboiteau",
    "aboiteaux": "aboiteau",
    "abortoria": "abortorium",
    "abortuses": "abortus",
    "abscissae": "abscissa",
    "abscissas": "abscissa",
    "abstracta": "abstractum",
    "acari": "acarus",
    "achinese": "achinese",
    "aciculae": "acicula",
    "acini": "acinus",
    "acroamata": "acroama",
    "acropolises": "acropolis",
    "actiniae": "actinia",
    "actinias": "actinia",
    "aculei": "aculeus",
    "addenda": "addendum",
    "addendums": "addendum",
    "adieus": "adieu",
    "adieux": "adieu",
    "adjutant generals": "adjutant general",
    "adminicula": "adminiculum",
    "ados": "ado",
    "advisoes": "adviso",
    "adyta": "adytum",
    "aecidia": "aecidium",
    "aegises": "aegis",
    "aelfe": "aelf",
    "aficionados": "aficionado",
    "afreeti": "afreet",
    "afreets": "afreet",
    "afriti": "afrit",
    "afrits": "afrit",
    "agapae": "agape",
    "agapes": "agape",
    "agenda": "agendum",
    "agends": "agend",
    "agents provocateurs": "agent provocateur",
    "aggros": "aggro",
    "agones": "agon",
    "agorae": "agora",
    "agrogoroda": "agrogorod",
    "aides de camp": "aide de camp",
    "aides-de-camp": "aide-de-camp",
    "alae": "ala",
    "alani": "alan",
    "alans": "alan",
    "alberghi": "albergo",
    "albinos": "albino",
    "ale wives": "ale wife",
    "ale-wives": "ale-wife",
    "alfurese": "alfurese",
    "algae": "alga",
    "alguaciles": "alguacil",
    "alguacils": "alguacil",
    "aliases": "alias",
    "allargandi": "allargando",
    "allegros": "allegro",
    "alluvia": "alluvium",
    "alluviums": "alluvium",
    "alti": "alto",
    "alto relievos": "alto relievo",
    "alto-relievos": "alto-relievo",
    "altorelievos": "altorelievo",
    "altos": "alto",
    "alulae": "alula",
    "alumnae": "alumna",
    "alumni": "alumnus",
    "alveoli": "alveolus",
    "amboinese": "amboinese",
    "ambones": "ambo",
    "ambos": "ambo",
    "ambulacra": "ambulacrum",
    "ambuscadoes": "ambuscado",
    "ambuscados": "ambuscado",
    "amenta": "amentum",
    "americanese": "americanese",
    "amici curiae": "amicus curiae",
    "ammos": "ammo",
    "amoebae": "amoeba",
    "amoebas": "amoeba",
    "amoretti": "amoretto",
    "amorini": "amorino",
    "amoyese": "amoyese",
    "amphipneusta": "amphipneust",
    "amphipneusts": "amphipneust",
    "amphorae": "amphora",
    "anacolutha": "anacoluthon",
    "anacoluthons": "anacoluthon",
    "anathemas": "anathema",
    "anathemata": "anathema",
    "ancone": "ancona",
    "andamanese": "andamanese",
    "angiomata": "angioma",
    "angolese": "angolese",
    "angustiae": "angustia",
    "anlagen": "anlage",
    "anlages": "anlage",
    "annamese": "annamese",
    "antae": "anta",
    "antananarivos": "antananarivo",
    "antefixa": "antefix",
    "antennae": "antenna",
    "antennas": "antenna",
    "anthemia": "anthemion",
    "anthodia": "anthodium",
    "anti sera": "anti serum",
    "anti serums": "anti serum",
    "anti-sera": "anti-serum",
    "anti-serums": "anti-serum",
    "antifixes": "antefix",
    "antisera": "antiserum",
    "antra": "antrum",
    "apexes": "apex",
    "aphelia": "aphelion",
    "aphides": "aphis",
    "apices": "apex",
    "apothecia": "apothecium",
    "apparatchiki": "apparatchik",
    "apparatchiks": "apparatchik",
    "apparatus": "apparatus",
    "apparatuses": "apparatus",
    "appendices": "appendix",
    "appendixes": "appendix",
    "appressoria": "appressorium",
    "apsides": "apsis",
    "apteria": "apterium",
    "aptychi": "aptychus",
    "aquaria": "aquarium",
    "aquariums": "aquarium",
    "aragonese": "aragonese",
    "arakanese": "arakanese",
    "arangoes": "arango",
    "arboreta": "arboretum",
    "arcana": "arcanum",
    "archegonia": "archegonium",
    "archipelagoes": "archipelago",
    "archipelagos": "archipelago",
    "arcosolia": "arcosolium",
    "areolae": "areola",
    "argumenta": "argumentum",
    "aristae": "arista",
    "armadillos": "armadillo",
    "arrhae": "arrha",
    "asbestoses": "asbestos",
    "ascidia": "ascidium",
    "asherahs": "asherah",
    "asherim": "asherah",
    "assamese": "assamese",
    "asteriae": "asterias",
    "astrocytomas": "astrocytoma",
    "astrocytomata": "astrocytoma",
    "asyla": "asylum",
    "asylums": "asylum",
    "asyndeta": "asyndeton",
    "atamans": "ataman",
    "atlantes": "atlas",
    "atlases": "atlas",
    "atmas": "atman",
    "aurei": "aureus",
    "aurorae": "aurora",
    "auroras": "aurora",
    "auslesen": "auslese",
    "ausleses": "auslese",
    "auspices": "auspex",
    "auto de fes": "auto de fe",
    "auto-de-fes": "auto-de-fe",
    "autobahnen": "autobahn",
    "autobahns": "autobahn",
    "autochthones": "autochthon",
    "autochthons": "autochthon",
    "autos": "auto",
    "autos de fes": "auto de fe",
    "autos-de-fes": "auto-de-fe",
    "autostradas": "autostrada",
    "autostrade": "autostrada",
    "avicularia": "avicularium",
    "avignonese": "avignonese",
    "avisos": "aviso",
    "avocados": "avocado",
    "axes": "axe",
    "axillae": "axilla",
    "axones": "axon",
    "axons": "axon",
    "axopodia": "axopodium",
    "azerbaijanese": "azerbaijanese",
    "baalim": "baal",
    "baals": "baal",
    "bacilli": "bacillus",
    "bacteria": "bacterium",
    "bahamans": "bahaman",
    "bajocchi": "bajocco",
    "balinese": "balinese",
    "ballabili": "ballabile",
    "ballate": "ballata",
    "ballerinas": "ballerina",
    "ballerine": "ballerina",
    "ballistae": "ballista",
    "ballistas": "ballista",
    "bamakos": "bamako",
    "bandeaus": "bandeau",
    "bandeaux": "bandeau",
    "bandits": "bandit",
    "banditti": "bandit",
    "barquisimetos": "barquisimeto",
    "barricoes": "barrico",
    "bases": "base",
    "basilicae": "basilica",
    "basilicas": "basilica",
    "bass": "bass",
    "basses": "bass",
    "bassi": "basso",
    "bassos": "basso",
    "bateaus": "bateau",
    "bateaux": "bateau",
    "bathoses": "bathos",
    "baths": "bath",
    "batzes": "batz",
    "beau gestes": "beau geste",
    "beaus": "beau",
    "beaux": "beau",
    "beaux esprits": "bel esprit",
    "beaux gestes": "beau geste",
    "beaux hommes": "bel homme",
    "bechuana": "bechuana",
    "bechuanas": "bechuana",
    "beefs": "beef",
    "beerenauslesen": "beerenauslese",
    "beeves": "beef",
    "belle epoques": "belle epoque",
    "belles epoques": "belle epoque",
    "bemas": "bema",
    "bemata": "bema",
    "bengalese": "bengalese",
    "bernese": "bernese",
    "bersaglieri": "bersagliere",
    "bhikshook": "bhikshu",
    "bhikshus": "bhikshu",
    "bhutanese": "bhutanese",
    "biases": "bias",
    "bierstuben": "bierstube",
    "bierstubes": "bierstube",
    "bifolia": "bifolium",
    "bijoux": "bijou",
    "bilboes": "bilbo",
    "bimbos": "bimbo",
    "bingos": "bingo",
    "biros": "biro",
    "biscotti": "biscotto",
    "bison": "bison",
    "bisons": "bison",
    "blastemata": "blastema",
    "blemyae": "blemya",
    "blini": "blin",
    "blinis": "blin",
    "blitzes": "blitz",
    "blouses": "blouse",
    "blues": "blue",
    "bodegones": "bodegon",
    "boleros": "bolero",
    "bolognese": "bolognese",
    "boluses": "bolus",
    "bolzanos": "bolzano",
    "bon mots": "bon mot",
    "bon vivants": "bon vivant",
    "bongoes": "bongo",
    "bongos": "bongo",
    "bonis fidebus": "bona fides",
    "bons mots": "bon mot",
    "bons vivants": "bon vivant",
    "bordereaux": "bordereau",
    "borghese": "borghese",
    "bostonese": "bostonese",
    "botos": "boto",
    "bozzetti": "bozzetto",
    "braccia": "braccio",
    "brachiopoda": "brachiopod",
    "brachiopods": "brachiopod",
    "branchiopoda": "branchiopod",
    "branchiopods": "branchiopod",
    "bratwurste": "bratwurst",
    "bratwursts": "bratwurst",
    "bravoes": "bravo",
    "bravos": "bravo",
    "break aways": "break away",
    "break-aways": "break-away",
    "breakaways": "breakaway",
    "breaks away": "break away",
    "breaks-away": "break-away",
    "breaksaway": "breakaway",
    "bream": "bream",
    "breeches": "breech",
    "bregmata": "bregma",
    "brethren": "brother",
    "brethren german": "brother german",
    "brethren-german": "brother-german",
    "brigadier generals": "brigadier general",
    "britches": "britches",
    "bronchi": "bronchus",
    "bronchitises": "bronchitis",
    "brothers": "brother",
    "brothers german": "brother german",
    "brothers-german": "brother-german",
    "bruschette": "bruschetta",
    "buboes": "bubo",
    "buffalo": "buffalo",
    "buffalos": "buffalo",
    "buginese": "buginese",
    "bulimi": "bulimus",
    "bullae": "bulla",
    "bureaus": "bureau",
    "bureaux": "bureau",
    "burmans": "burman",
    "burmese": "burmese",
    "burros": "burro",
    "bursae": "bursa",
    "burschen": "bursch",
    "bursitises": "bursitis",
    "bustoes": "busto",
    "bustos": "busto",
    "buzzes": "buzz",
    "bylinas": "bylina",
    "byliny": "bylina",
    "cabalettas": "cabaletta",
    "cabalette": "cabaletta",
    "cacti": "cactus",
    "cactuses": "cactus",
    "caddises": "caddis",
    "caducei": "caduceus",
    "caeca": "caecum",
    "caffe macchiati": "caffe macchiato",
    "caffe macchiatos": "caffe macchiato",
    "caffe ristretti": "caffe ristretto",
    "caffe ristrettos": "caffe ristretto",
    "caimans": "caiman",
    "cairos": "cairo",
    "calabrese": "calabrese",
    "calamistra": "calamistrum",
    "calathi": "calathus",
    "calices": "calix",
    "calluses": "callus",
    "calypsos": "calypso",
    "calzones": "calzone",
    "calzoni": "calzone",
    "camaldolese": "camaldolese",
    "camera obscuras": "camera obscura",
    "camerae obscurae": "camera obscura",
    "canaliculi": "canaliculus",
    "candelabra": "candelabrum",
    "cannabises": "cannabis",
    "canthari": "cantharus",
    "cantonese": "cantonese",
    "cantos": "canto",
    "cantus": "cantus",
    "cantuses": "cantus",
    "canvases": "canvas",
    "canzonettas": "canzonetta",
    "canzonette": "canzonetta",
    "capataces": "capataz",
    "capi": "capo",
    "capos": "capo",
    "cappuccinos": "cappuccino",
    "carabinieri": "carabiniere",
    "carcinomas": "carcinoma",
    "carcinomata": "carcinoma",
    "cardines": "cardo",
    "cardophagi": "cardophagus",
    "cargadores": "cargador",
    "caribou": "caribou",
    "caribous": "caribou",
    "carices": "carex",
    "carlylese": "carlylese",
    "carmens": "carmen",
    "carmina": "carmen",
    "carp": "carp",
    "cartes de visite": "carte de visite",
    "cartes-de-visite": "carte-de-visite",
    "caryatides": "caryatid",
    "caryatids": "caryatid",
    "caryopsides": "caryopsis",
    "casinos": "casino",
    "cassinese": "cassinese",
    "cassones": "cassone",
    "cassoni": "cassone",
    "castra": "castrum",
    "castrati": "castrato",
    "casus belli": "casus belli",
    "cathismata": "cathisma",
    "caudices": "caudex",
    "caules": "caulis",
    "caveae": "cavea",
    "caymans": "cayman",
    "cecidia": "cecidium",
    "celanese": "celanese",
    "celiae": "celia",
    "cellos": "cello",
    "cementomata": "cementoma",
    "centesimi": "centesimo",
    "centumviri": "centumvir",
    "cerci": "cercus",
    "cerimans": "ceriman",
    "ceylonese": "ceylonese",
    "chaetae": "chaeta",
    "chaoses": "chaos",
    "chapeaus": "chapeau",
    "chapeaux": "chapeau",
    "chapplies": "chapplis",
    "charismas": "charisma",
    "charismata": "charisma",
    "chassis": "chassis",
    "chateaus": "chateau",
    "chateaux": "chateau",
    "cherubim": "cherub",
    "cherubs": "cherub",
    "chervontzi": "chervonetz",
    "chevaux": "cheval",
    "chevaux de bataille": "cheval de bataille",
    "chevaux-de-bataille": "cheval-de-bataille",
    "chicagos": "chicago",
    "children": "child",
    "chimangos": "chimango",
    "chinese": "chinese",
    "chintzes": "chintz",
    "chitarroni": "chitarrone",
    "chlamydiae": "chlamydia",
    "chlamydias": "chlamydia",
    "choanae": "choana",
    "chordae": "chorda",
    "choruses": "chorus",
    "choux": "chou",
    "chromidia": "chromidium",
    "chromonemata": "chromonema",
    "chrysalides": "chrysalis",
    "chrysalises": "chrysalis",
    "cicatrices": "cicatrix",
    "cilantros": "cilantro",
    "cimices": "cimex",
    "cinclides": "cinclis",
    "cistophorori": "cistophorus",
    "cladi": "cladus",
    "claustra": "claustrum",
    "clausulae": "clausula",
    "clavulae": "clavula",
    "cleithra": "cleithrum",
    "clepsydrae": "clepsydra",
    "clepsydras": "clepsydra",
    "clippers": "clippers",
    "clitorides": "clitoris",
    "clitorises": "clitoris",
    "clostridia": "clostridium",
    "coagula": "coagulum",
    "coccidia": "coccidium",
    "cochitos": "cochito",
    "cocos": "coco",
    "cod": "cod",
    "codices": "codex",
    "coeliae": "coelia",
    "coes": "co",
    "cognoscenti": "cognoscente",
    "coitus": "coitus",
    "coituses": "coitus",
    "collegia": "collegium",
    "collegia musica": "collegium musicum",
    "colombos": "colombo",
    "colorados": "colorado",
    "columbaria": "columbarium",
    "columnae": "columna",
    "comae": "coma",
    "comatulae": "comatula",
    "commandos": "commando",
    "compendia": "compendium",
    "compendiums": "compendium",
    "competentes": "competent",
    "concerti grossi": "concerto grosso",
    "concertinos": "concertino",
    "concetti": "concetto",
    "condos": "condo",
    "condottieri": "condottiere",
    "conducti": "conductus",
    "coneys": "coney",
    "confervae": "conferva",
    "confluvia": "confluvium",
    "congii": "congius",
    "congoese": "congoese",
    "congolese": "congolese",
    "conidia": "conidium",
    "conies": "coney",
    "conoscenti": "conoscente",
    "consortia": "consortium",
    "consortiums": "consortium",
    "contadine": "contadina",
    "contadini": "contadino",
    "contagia": "contagium",
    "contangos": "contango",
    "contrafacta": "contrafactum",
    "contralti": "contralto",
    "contraltos": "contralto",
    "contretemps": "contretemps",
    "conversaziones": "conversazione",
    "conversazioni": "conversazione",
    "convolvuli": "convolvulus",
    "convolvuluses": "convolvulus",
    "conys": "cony",
    "corgis": "corgi",
    "corgwn": "corgi",
    "corgys": "corgy",
    "cornetti": "cornetto",
    "corni": "corno",
    "cornua": "cornu",
    "coronae": "corona",
    "coronas": "corona",
    "corpora": "corpus",
    "corpora vilia": "corpus vile",
    "corps": "corps",
    "corpuscula": "corpusculum",
    "corpuses": "corpus",
    "corrigenda": "corrigendum",
    "cortexes": "cortex",
    "cortices": "cortex",
    "cosmoses": "cosmos",
    "costae": "costa",
    "coups d'etat": "coup d'etat",
    "court martials": "court martial",
    "court-martials": "court-martial",
    "courts martial": "court martial",
    "courts-martial": "court-martial",
    "cows": "cow",
    "coxendices": "coxendix",
    "cozes": "coz",
    "craft": "craft",
    "crafts childer": "crafts child",
    "crafts-childer": "crafts-child",
    "crania": "cranium",
    "craniums": "cranium",
    "craspeda": "craspedum",
    "crazie": "crazia",
    "credos": "credo",
    "cremasteres": "cremaster",
    "cremasters": "cremaster",
    "crescendos": "crescendo",
    "cris de coeur": "cri de coeur",
    "cristae": "crista",
    "criteria": "criterion",
    "crowfoots": "crowfoot",
    "crura": "crus",
    "crustae": "crusta",
    "cubicula": "cubiculum",
    "culs de four": "cul de four",
    "culs de lampe": "cul de lampe",
    "culs de sac": "cul de sac",
    "culs-de-four": "cul-de-four",
    "culs-de-lampe": "cul-de-lampe",
    "culs-de-sac": "cul-de-sac",
    "cultelli": "cultellus",
    "cumuli": "cumulus",
    "cuniculi": "cuniculus",
    "curiosi": "curioso",
    "curiosos": "curioso",
    "curricula": "curriculum",
    "curricula vitae": "curriculum vitae",
    "curriculums": "curriculum",
    "cuspides": "cuspis",
    "cyanos": "cyano",
    "cyathi": "cyathus",
    "cyclopes": "cyclops",
    "cyclopses": "cyclops",
    "cynocephali": "cynocephalus",
    "cyphellae": "cyphella",
    "cysticerci": "cysticercus",
    "cystidia": "cystidium",
    "cystomata": "cystoma",
    "czechs": "czech",
    "dace": "dace",
    "daces": "dace",
    "daises": "dais",
    "data": "datum",
    "debris": "debris",
    "decennia": "decennium",
    "deer": "deer",
    "deliria": "delirium",
    "deliriums": "delirium",
    "demos": "demo",
    "denarii": "denarius",
    "denkmaler": "denkmal",
    "denotata": "denotatum",
    "dentalia": "dentalium",
    "deposita": "depositum",
    "depositums": "depositum",
    "descripta": "descriptum",
    "desiderata": "desideratum",
    "designata": "designatum",
    "desmans": "desman",
    "desmas": "desma",
    "desmata": "desma",
    "devoti": "devoto",
    "devotos": "devoto",
    "diabetes": "diabetes",
    "diastemata": "diastema",
    "dice": "die",
    "dichasia": "dichasium",
    "dicta": "dictum",
    "dictums": "dictum",
    "didymes": "didymis",
    "dies": "die",
    "differentiae": "differentia",
    "digitalises": "digitalis",
    "dilettantes": "dilettante",
    "dilettanti": "dilettante",
    "diplococci": "diplococcus",
    "diplomas": "diploma",
    "diplomata": "diploma",
    "diporpae": "diporpa",
    "dipsades": "dipsas",
    "directrices": "directrix",
    "dittos": "ditto",
    "diverticula": "diverticulum",
    "divertimenti": "divertimento",
    "divertimentos": "divertimento",
    "djinn": "djinn",
    "djinns": "djinn",
    "does": "doe",
    "dogmas": "dogma",
    "dogmata": "dogma",
    "dolmans": "dolman",
    "dolphinaria": "dolphinarium",
    "dolphinariums": "dolphinarium",
    "dongolese": "dongolese",
    "dracos": "draco",
    "dragomans": "dragoman",
    "dragomen": "dragoman",
    "dramas": "drama",
    "dreikanter": "dreikanter",
    "dreikanters": "dreikanter",
    "dumkas": "dumka",
    "dumky": "dumka",
    "duumviri": "duumvir",
    "duumvirs": "duumvir",
    "dynamos": "dynamo",
    "edemas": "edema",
    "edemata": "edema",
    "editiones principes": "editio princeps",
    "efreeti": "efreet",
    "efreets": "efreet",
    "eland": "eland",
    "elands": "eland",
    "elk": "elk",
    "elks": "elk",
    "ellipses": "ellipse",
    "elves": "elf",
    "emblememata": "emblema",
    "embryomata": "embryoma",
    "embryos": "embryo",
    "emporia": "emporium",
    "emporiums": "emporium",
    "enconia": "enconium",
    "enconiums": "enconium",
    "endotheliomata": "endothelioma",
    "enemas": "enema",
    "enemata": "enema",
    "enigmas": "enigma",
    "enigmata": "enigma",
    "ententes cordiales": "entente cordiale",
    "entera": "enteron",
    "entia rationis": "ens rationis",
    "entia realia": "ens reale",
    "entrochi": "entrochus",
    "ephelides": "ephelis",
    "ephemerae": "ephemera",
    "ephemeras": "ephemera",
    "ephemerides": "ephemeris",
    "ephemerises": "ephemeris",
    "epicardia": "epicardium",
    "epidermises": "epidermis",
    "epididymides": "epididymis",
    "epididymises": "epididymis",
    "epilimnia": "epilimnion",
    "epimera": "epimeron",
    "epiphenomena": "epiphenomenon",
    "epitheliomata": "epithelioma",
    "epyllia": "epyllion",
    "equinoctia": "equinoctium",
    "equinoctiums": "equinoctium",
    "equiseta": "equisetum",
    "equisetums": "equisetum",
    "erfs": "erf",
    "ergastula": "ergastulum",
    "eroses": "eros",
    "erotes": "eros",
    "errata": "erratum",
    "erven": "erf",
    "espanoles": "espanol",
    "esperantos": "esperanto",
    "espressos": "espresso",
    "ethoses": "ethos",
    "eucalypti": "eucalyptus",
    "eucalyptuses": "eucalyptus",
    "eunuchs": "eunuch",
    "euporista": "euporiston",
    "euros": "euro",
    "evirati": "evirato",
    "exanthemata": "exanthema",
    "exempla": "exemplum",
    "exencephali": "exencephalus",
    "exes": "ex",
    "explananda": "explanandum",
    "explicanda": "explicandum",
    "extispices": "extispex",
    "extrema": "extremum",
    "extremums": "extremum",
    "eyases": "eyas",
    "fabellae": "fabella",
    "fabliaux": "fabliau",
    "facienda": "faciendum",
    "faeculae": "faecula",
    "faits accomplis": "fait accompli",
    "falces": "falx",
    "falsettos": "falsetto",
    "familiae": "familia",
    "famuli": "famulus",
    "fanfoots": "fanfoot",
    "farmans": "farman",
    "faroese": "faroese",
    "faros": "faro",
    "fasciae": "fascia",
    "fascias": "fascia",
    "fasciculi": "fasciculus",
    "faunae": "fauna",
    "faunas": "fauna",
    "favellae": "favella",
    "faveoli": "faveolus",
    "favissae": "favissa",
    "feculae": "fecula",
    "fedai": "fedai",
    "fedais": "fedai",
    "feet": "foot",
    "feiseanna": "feis",
    "felones": "felo",
    "felones de se": "felo de se",
    "felones-de-se": "felo-de-se",
    "felos de se": "felo de se",
    "felos-de-se": "felo-de-se",
    "felsenmeere": "felsenmeer",
    "femora": "femur",
    "femurs": "femur",
    "fenestrae": "fenestra",
    "fermatas": "fermata",
    "fermate": "fermata",
    "ferrarese": "ferrarese",
    "fess": "feis",
    "fiascos": "fiasco",
    "fibrae": "fibra",
    "fibrillae": "fibrilla",
    "fibromata": "fibroma",
    "fibulae": "fibula",
    "fibulas": "fibula",
    "fidei defensores": "fidei defensor",
    "figurantes": "figurante",
    "figuranti": "figurante",
    "filariae": "filaria",
    "filarias": "filaria",
    "filipinos": "filipino",
    "finger ends": "finger end",
    "finger-ends": "finger-end",
    "fingers ends": "finger end",
    "fingers-ends": "finger-end",
    "fioriture": "fioritura",
    "fish": "fish",
    "fizzes": "fizz",
    "flabella": "flabellum",
    "flag staffs": "flag staff",
    "flag staves": "flag staff",
    "flag-staffs": "flag-staff",
    "flag-staves": "flag-staff",
    "flagella": "flagellum",
    "flamencos": "flamenco",
    "flatfoots": "flatfoot",
    "flatuses": "flatus",
    "flies": "fly",
    "flocci": "floccus",
    "flocculi": "flocculus",
    "florae": "flora",
    "floras": "flora",
    "flounder": "flounder",
    "flys": "fly",
    "foci": "focus",
    "focuses": "focus",
    "folles": "follis",
    "fomites": "fomes",
    "foochowese": "foochowese",
    "fora": "forum",
    "foramens": "foramen",
    "foramina": "foramen",
    "formulae": "formula",
    "formulas": "formula",
    "forums": "forum",
    "fossae": "fossa",
    "fowl": "fowl",
    "fowls": "fowl",
    "fraena": "fraenum",
    "franc archers": "franc archer",
    "franc-archers": "franc-archer",
    "francs archers": "franc archer",
    "francs-archers": "franc-archer",
    "frati": "frate",
    "frena": "frenum",
    "fricandeaux": "fricandeau",
    "frizzes": "frizz",
    "froes": "fro",
    "frottole": "frottola",
    "frusta": "frustum",
    "frustula": "frustulum",
    "frustums": "frustum",
    "fuci": "fucus",
    "fulcra": "fulcrum",
    "funambuli": "funambulus",
    "fundamenta divisionis": "fundamentum divisionis",
    "funghi porchini": "fungo porcino",
    "fungi": "fungus",
    "funguses": "fungus",
    "furfures": "furfur",
    "furiosos": "furioso",
    "fuzzes": "fuzz",
    "gabunese": "gabunese",
    "galla": "galla",
    "gallas": "galla",
    "gallows": "gallows",
    "gametangia": "gametangium",
    "ganglia": "ganglion",
    "ganglions": "ganglion",
    "gaonim": "gaon",
    "gaons": "gaon",
    "gases": "gas",
    "gastrocnemii": "gastrocnemius",
    "geese": "goose",
    "geisha": "geisha",
    "geishas": "geisha",
    "geisteswissenschaften": "geisteswissenschaft",
    "gemmae": "gemma",
    "genae": "gena",
    "genera": "genus",
    "generalissimos": "generalissimo",
    "genevese": "genevese",
    "genies": "genie",
    "genii": "genie",
    "geniuses": "genius",
    "genoese": "genoese",
    "gentes": "gens",
    "germans": "german",
    "germaria": "germarium",
    "gestapos": "gestapo",
    "ghettos": "ghetto",
    "gigolos": "gigolo",
    "gilbertese": "gilbertese",
    "giunchi": "giuncus",
    "gizmos": "gizmo",
    "glacis": "glacis",
    "glacises": "glacis",
    "gladioli": "gladiolus",
    "gladioluses": "gladiolus",
    "glasses": "glass",
    "gliomata": "glioma",
    "gliosarcomata": "gliosarcoma",
    "glissandi": "glissando",
    "glissandos": "glissando",
    "globigerinae": "globigerina",
    "glochidia": "glochidium",
    "glomera": "glomus",
    "glottises": "glottis",
    "glutaei": "glutaeus",
    "glutei": "glutaeus",
    "gminy": "gmina",
    "goanese": "goanese",
    "goes": "go",
    "gonangia": "gonangium",
    "gonidia": "gonidium",
    "gonimia": "gonimium",
    "gorgoniae": "gorgonia",
    "gorgonias": "gorgonia",
    "gousblomme": "gousblom",
    "goyim": "goy",
    "goys": "goy",
    "graben": "graben",
    "grabens": "graben",
    "graffiti": "graffiti",
    "grani": "grano",
    "granite": "granita",
    "granulae": "granula",
    "gravamina": "gravamen",
    "greensboros": "greensboro",
    "gringos": "gringo",
    "grissini": "grissino",
    "grottoes": "grotto",
    "grottos": "grotto",
    "grouse": "grouse",
    "grouses": "grouse",
    "guaiaberos": "guaiabero",
    "guanos": "guano",
    "guardias civiles": "guardia civil",
    "gubernacula": "gubernaculum",
    "guglioes": "guglio",
    "guglios": "guglio",
    "guianese": "guianese",
    "gumbos": "gumbo",
    "gummas": "gumma",
    "gummata": "gumma",
    "gutti": "guttus",
    "guys": "guy",
    "gymnasia": "gymnasium",
    "gymnasiums": "gymnasium",
    "gymnoti": "gymnotus",
    "gyri": "gyrus",
    "gyros": "gyro",
    "haddock": "haddock",
    "haddocks": "haddock",
    "haemangiomas": "haemangioma",
    "haemangiomata": "haemangioma",
    "haggis": "haggis",
    "hainanese": "hainanese",
    "hairdos": "hairdo",
    "hake": "hake",
    "hakes": "hake",
    "halibut": "halibut",
    "halibuts": "halibut",
    "halluces": "hallux",
    "halteridia": "halteridium",
    "hamadryades": "hamadryad",
    "hamadryads": "hamadryad",
    "hamuli": "hamulus",
    "hapax legomena": "hapax legomenon",
    "hapax legomenons": "hapax legomenon",
    "haptera": "hapteron",
    "hararese": "hararese",
    "harlemese": "harlemese",
    "harmans": "harman",
    "haruspices": "haruspex",
    "haustella": "haustellum",
    "haustoria": "haustorium",
    "haustra": "haustrum",
    "havanese": "havanese",
    "headquarters": "headquarters",
    "heavenese": "heavenese",
    "hedra": "hedron",
    "hedrons": "hedron",
    "heen": "hsien",
    "hefte": "heft",
    "helia": "helion",
    "helices": "helix",
    "helions": "helion",
    "hemangeiomas": "hemangeioma",
    "hemangeiomata": "hemangeioma",
    "hemangiomas": "hemangioma",
    "hemangiomata": "hemangioma",
    "hemelytra": "hemelytrum",
    "hemerae": "hemera",
    "hemi elytra": "hemi elytrum",
    "hemi-elytra": "hemi-elytrum",
    "herniae": "hernia",
    "hernias": "hernia",
    "heroa": "heroon",
    "heroes": "hero",
    "herpes": "herpes",
    "herren": "herr",
    "herring": "herring",
    "herrings": "herring",
    "hertz": "hertz",
    "hetaerae": "hetaera",
    "hetairai": "hetaera",
    "hetmans": "hetman",
    "hiatus": "hiatus",
    "hiatuses": "hiatus",
    "hijinks": "hijinks",
    "hinaki": "hinaki",
    "hippocampi": "hippocampus",
    "hippopotami": "hippopotamus",
    "hippopotamuses": "hippopotamus",
    "hippos": "hippo",
    "hiroshimans": "hiroshiman",
    "hoes": "hoe",
    "holothuriae": "holothuria",
    "holothurias": "holothuria",
    "honoraria": "honorarium",
    "honorariums": "honorarium",
    "hoofs": "hoof",
    "hoosierese": "hoosierese",
    "hooves": "hoof",
    "hordeola": "hordeolum",
    "hormogonia": "hormogonium",
    "hoties": "hoti",
    "hottentotese": "hottentotese",
    "hubrises": "hubris",
    "humans": "human",
    "humeri": "humerus",
    "hunanese": "hunanese",
    "hydrae": "hydra",
    "hydras": "hydra",
    "hydraulica": "hydraulicon",
    "hydriae": "hydria",
    "hymenia": "hymenium",
    "hymnaria": "hymnarium",
    "hyperbata": "hyperbaton",
    "hyperbolae": "hyperbola",
    "hyperbolas": "hyperbola",
    "hyphae": "hypha",
    "hypna": "hypnum",
    "hypnums": "hypnum",
    "hypodermata": "hypoderma",
    "hypolimnia": "hypolimnion",
    "ibises": "ibis",
    "ichthyosauri": "ichthyosaurus",
    "idahos": "idaho",
    "idees fixes": "idee fixe",
    "idola": "idolum",
    "ignoramus": "ignoramus",
    "ignoramuses": "ignoramus",
    "ilia": "ilium",
    "imagines": "imago",
    "imagos": "imago",
    "imbrices": "imbrex",
    "impetigos": "impetigo",
    "impetus": "impetus",
    "impetuses": "impetus",
    "improperia": "improperium",
    "improvisatrici": "improvisatrice",
    "incubi": "incubus",
    "incubuses": "incubus",
    "indexes": "index",
    "indices": "index",
    "indicia": "indicium",
    "indigenae": "indigena",
    "individua": "individuum",
    "individuums": "individuum",
    "indochinese": "indochinese",
    "indumenta": "indumentum",
    "indusia": "indusium",
    "infernos": "inferno",
    "infimae species": "infima species",
    "infos": "info",
    "inhalatoria": "inhalatorium",
    "innings": "innings",
    "inocula": "inoculum",
    "insulae": "insula",
    "intaglios": "intaglio",
    "interambulacra": "interambulacrum",
    "interaxes": "interaxis",
    "intercalare": "intercalarium",
    "intercalaria": "intercalarium",
    "intermaxillae": "intermaxilla",
    "intermedia": "intermedium",
    "intermezzos": "intermezzo",
    "internodia": "internodium",
    "interphenomena": "interphenomenon",
    "interradii": "interradius",
    "interreges": "interrex",
    "interregna": "interregnum",
    "interregnums": "interregnum",
    "interstitia": "interstitium",
    "interstitiums": "interstitium",
    "intertrigos": "intertrigo",
    "intervalla": "intervallum",
    "intervallums": "intervallum",
    "intichiuma": "intichiuma",
    "inuit": "inuk",
    "inukshuit": "inukshuk",
    "involucra": "involucrum",
    "ipse dixits": "ipse dixit",
    "iquicos": "iquico",
    "irides": "iris",
    "irises": "iris",
    "ischia": "ischium",
    "isidia": "isidium",
    "ispravniki": "ispravnik",
    "ispravniks": "ispravnik",
    "isthmi": "isthmus",
    "isthmuses": "isthmus",
    "iters": "iter",
    "itinera": "iter",
    "iueles": "juelet",
    "ixodides": "ixodid",
    "jackanapes": "jackanapes",
    "jacobus": "jacobus",
    "jacobuses": "jacobus",
    "jalapenos": "jalapeno",
    "jambeaus": "jambeau",
    "jambeaux": "jambeau",
    "japanese": "japanese",
    "javanese": "javanese",
    "jazzes": "jazz",
    "jerries": "jerry",
    "jerrys": "jerry",
    "jeux": "jeu",
    "joes": "joe",
    "johnsonese": "johnsonese",
    "judicata": "judicatum",
    "juga": "jugum",
    "jumbos": "jumbo",
    "juntos": "junto",
    "kakapos": "kakapo",
    "kanarese": "kanarese",
    "kantharoi": "kantharos",
    "karakia": "karakia",
    "karakias": "karakia",
    "kare": "kar",
    "karrenfelder": "karrenfeld",
    "karrenfelds": "karrenfeld",
    "kars": "kar",
    "kernoi": "kernos",
    "kibbutzes": "kibbutz",
    "kibbutzim": "kibbutz",
    "kilos": "kilo",
    "kimonos": "kimono",
    "kine": "cow",
    "kinkimavos": "kinkimavo",
    "kioko": "kioko",
    "kiokos": "kioko",
    "kiplingese": "kiplingese",
    "kletterschuhe": "kletterschuh",
    "knives": "knife",
    "knoppern": "knopper",
    "knoppers": "knopper",
    "kobo": "kobo",
    "kokakos": "kokako",
    "konditorei": "konditorei",
    "konditoreien": "konditorei",
    "kongoese": "kongoese",
    "kongolese": "kongolese",
    "kosovos": "kosovo",
    "kouroi": "kouros",
    "kovshi": "kovsh",
    "kuvaszok": "kuvasz",
    "kylikes": "kylix",
    "kylixes": "kylix",
    "labra": "labrum",
    "laciniae": "lacinia",
    "lactobacilli": "lactobacillus",
    "lacunae": "lacuna",
    "lacunaria": "lacunar",
    "lacunars": "lacunar",
    "lacunas": "lacuna",
    "laiki": "laika",
    "lamellae": "lamella",
    "laminae": "lamina",
    "lapithae": "lapith",
    "lapiths": "lapith",
    "lapponese": "lapponese",
    "lares": "lar",
    "larnakes": "larnax",
    "lars": "lar",
    "larynges": "larynx",
    "latexes": "latex",
    "latices": "latex",
    "leaves": "leaf",
    "lebanese": "lebanese",
    "lebensformen": "lebensform",
    "leges humanae": "lex humana",
    "legionellae": "legionella",
    "legomena": "legomenon",
    "legumens": "legumen",
    "legumina": "legumen",
    "lei": "leu",
    "leishmaniae": "leishmania",
    "leishmanias": "leishmania",
    "lemans": "leman",
    "lemmas": "lemma",
    "lemmata": "lemma",
    "lemnisci": "lemniscus",
    "lemures": "lemur",
    "lemurs": "lemur",
    "lenses": "lens",
    "lentigines": "lentigo",
    "lentisci": "lentiscus",
    "leonese": "leonese",
    "lepromas": "leproma",
    "lepromata": "leproma",
    "leprosaria": "leprosarium",
    "lepta": "lepton",
    "leptomonas": "leptomonas",
    "leptospirae": "leptospira",
    "lesothos": "lesotho",
    "liberos": "libero",
    "libidos": "libido",
    "libretti": "libretto",
    "librettos": "libretto",
    "lice": "louse",
    "lidos": "lido",
    "lieder": "lied",
    "lieus": "lieu",
    "lieutenant generals": "lieutenant general",
    "lieux": "lieu",
    "ligamenta": "ligamentum",
    "lilos": "lilo",
    "limaces": "limax",
    "limans": "liman",
    "limbos": "limbo",
    "limites": "limes",
    "limos": "limo",
    "limuli": "limulus",
    "linctuses": "linctus",
    "line outs": "line out",
    "line-outs": "line-out",
    "lineae": "linea",
    "linenos": "lineno",
    "lines out": "line out",
    "lines-out": "line-out",
    "lingos": "lingo",
    "lingulae": "lingula",
    "linos": "lino",
    "lipomata": "lipoma",
    "lis": "lis",
    "lisses": "lis",
    "listerellae": "listerella",
    "listerellas": "listerella",
    "listeria": "listeria",
    "listerias": "listeria",
    "lithodomi": "lithodomus",
    "lithophyta": "lithophyton",
    "livedos": "livedo",
    "lives": "life",
    "lixivia": "lixivium",
    "loa": "loa",
    "loas": "loa",
    "loaves": "loaf",
    "lobuli": "lobulus",
    "lochi": "lochus",
    "loci": "locus",
    "locos": "loco",
    "loculi": "loculus",
    "loggias": "loggia",
    "loggie": "loggia",
    "logia": "logion",
    "logodaedale": "logodaedalus",
    "logodaedali": "logodaedalus",
    "logos": "logo",
    "lomata": "loma",
    "lomenta": "lomentum",
    "londonese": "londonese",
    "lora": "lorum",
    "lord lieutenants": "lord lieutenant",
    "lord-lieutenants": "lord-lieutenant",
    "lords lieutenant": "lord lieutenant",
    "lords-lieutenant": "lord-lieutenant",
    "lorrainese": "lorrainese",
    "louis": "louis",
    "lowlifes": "lowlife",
    "lox": "lox",
    "loxes": "lox",
    "lucchese": "lucchese",
    "lumbagos": "lumbago",
    "lumbrici": "lumbricus",
    "lumens": "lumen",
    "lumina": "lumen",
    "lustra": "lustrum",
    "lustrums": "lustrum",
    "lymphangiomata": "lymphangioma",
    "lymphomas": "lymphoma",
    "lymphomata": "lymphoma",
    "lyonese": "lyonese",
    "lyssae": "lyssa",
    "maare": "maar",
    "maars": "maar",
    "macaci": "macacus",
    "macanese": "macanese",
    "macaronies": "macaroni",
    "macassarese": "macassarese",
    "machos": "macho",
    "mackerel": "mackerel",
    "mackerels": "mackerel",
    "macros": "macro",
    "maculae": "macula",
    "maculas": "macula",
    "madurese": "madurese",
    "maecenases": "maecenas",
    "maecenates": "maecenas",
    "maestri": "maestro",
    "maestros": "maestro",
    "mafiose": "mafiosa",
    "mafiosi": "mafioso",
    "mafiosos": "mafioso",
    "magi": "magus",
    "magili": "magilus",
    "magmas": "magma",
    "magmata": "magma",
    "magna opera": "magnum opus",
    "magnetos": "magneto",
    "magnificos": "magnifico",
    "magnum opuses": "magnum opus",
    "major generals": "major general",
    "majuros": "majuro",
    "makuta": "likuta",
    "mala in se": "malum in se",
    "malabarese": "malabarese",
    "malabos": "malabo",
    "malagmata": "malagma",
    "mallei": "malleus",
    "malleoli": "malleolus",
    "maltese": "maltese",
    "man milliners": "man milliner",
    "man-milliners": "man-milliner",
    "mangoes": "mango",
    "mangos": "mango",
    "manifestos": "manifesto",
    "manteaus": "manteau",
    "manteaux": "manteau",
    "mantellette": "mantelletta",
    "mantises": "mantis",
    "manubria": "manubrium",
    "manubriums": "manubrium",
    "maputos": "maputo",
    "maquereaux": "maquereau",
    "maracaibos": "maracaibo",
    "maremme": "maremma",
    "mares": "mare",
    "maria": "mare",
    "maries": "mary",
    "marquises": "marquis",
    "marys": "mary",
    "maths": "math",
    "matinees musicales": "matinee musicale",
    "matins": "matin",
    "matres lectionis": "mater lectionis",
    "matzahs": "matzah",
    "matzot": "matzah",
    "mausolea": "mausoleum",
    "mausoleums": "mausoleum",
    "maxillae": "maxilla",
    "maxima": "maximum",
    "maximums": "maximum",
    "measles": "measles",
    "meatus": "meatus",
    "meatuses": "meatus",
    "media": "medium",
    "mediastina": "mediastinum",
    "mediatrices": "mediatrix",
    "medicoes": "medico",
    "medicos": "medico",
    "mediums": "medium",
    "medulloblastomas": "medulloblastoma",
    "medulloblastomata": "medulloblastoma",
    "medusae": "medusa",
    "medusas": "medusa",
    "megalosauri": "megalosaurus",
    "megatheria": "megatherium",
    "melanomas": "melanoma",
    "melanomata": "melanoma",
    "melastomae": "melastoma",
    "melastomas": "melastoma",
    "melicerides": "meliceris",
    "melismas": "melisma",
    "melismata": "melisma",
    "mementoes": "memento",
    "mementos": "memento",
    "memoranda": "memorandum",
    "memorandums": "memorandum",
    "memos": "memo",
    "men": "man",
    "men children": "man child",
    "men midwives": "man midwife",
    "men milliners": "man milliner",
    "men servants": "man servant",
    "men-children": "man-child",
    "men-midwives": "man-midwife",
    "men-milliners": "man-milliner",
    "men-servants": "man-servant",
    "menages a quatre": "menage a quatre",
    "menages a trois": "menage a trois",
    "menchildren": "manchild",
    "meningiomas": "meningioma",
    "meningiomata": "meningioma",
    "menisci": "meniscus",
    "menservants": "manservant",
    "menus": "menu",
    "meretrices": "meretrix",
    "merides": "meris",
    "meshumadim": "meshumad",
    "meshummadim": "meshummad",
    "mesothelia": "mesothelium",
    "mesotheliomas": "mesothelioma",
    "mesotheliomata": "mesothelioma",
    "messe di voce": "messa di voce",
    "messeigneurs": "monseigneur",
    "messinese": "messinese",
    "metae": "meta",
    "metalimnia": "metalimnion",
    "metapleura": "metapleuron",
    "metatarsi": "metatarsus",
    "metropolises": "metropolis",
    "metros": "metro",
    "mews": "mew",
    "mexicos": "mexico",
    "mezuzoth": "mezuza",
    "mezzo rilievos": "mezzo rilievo",
    "mezzo-rilievos": "mezzo-rilievo",
    "miasmas": "miasma",
    "miasmata": "miasma",
    "mice": "mouse",
    "micetraps": "mousetrap",
    "microbia": "microbion",
    "microcephali": "microcephalus",
    "micrococci": "micrococcus",
    "microfiche": "microfiche",
    "microfiches": "microfiche",
    "micronuclei": "micronucleus",
    "micros": "micro",
    "milanese": "milanese",
    "milanos": "milano",
    "milieus": "milieu",
    "milieux": "milieu",
    "milieux interieurs": "milieu interieur",
    "miliolae": "miliola",
    "milites gloriosi": "miles gloriosus",
    "millennia": "millennium",
    "millenniums": "millennium",
    "mimosae": "mimosa",
    "mimosas": "mimosa",
    "minima": "minimum",
    "minimums": "minimum",
    "minnelieder": "minnelied",
    "minutiae": "minutia",
    "minutias": "minutia",
    "minyanim": "minyan",
    "miracidia": "miracidium",
    "mithraea": "mithraeum",
    "mitochondria": "mitochondrion",
    "mixolimnia": "mixolimnion",
    "moakses": "moaks",
    "modelletti": "modelletto",
    "modelli": "modello",
    "modellos": "modello",
    "modenese": "modenese",
    "moderni": "modernus",
    "modi": "modus",
    "modii": "modius",
    "moduli": "modulus",
    "moduluses": "modulus",
    "moduses": "modus",
    "moes": "moe",
    "molimina": "molimen",
    "momenta": "momentum",
    "momentums": "momentum",
    "monacos": "monaco",
    "monades": "monas",
    "mondi": "mondo",
    "mongooses": "mongoose",
    "monies": "money",
    "monimolimnia": "monimolimnion",
    "monochasia": "monochasium",
    "monorchides": "monorchis",
    "monos": "mono",
    "monosyllaba": "monosyllabon",
    "monotae": "monota",
    "monoxyla": "monoxylon",
    "monsiegneurs": "monseigneur",
    "monsignori": "monsignore",
    "monsignors": "monsignor",
    "monstres sacres": "monstre sacre",
    "montenegros": "montenegro",
    "montes": "mons",
    "moose": "moose",
    "moratoria": "moratorium",
    "moratoriums": "moratorium",
    "morceaux": "morceau",
    "moroccos": "morocco",
    "morosi": "moroso",
    "mortadellas": "mortadella",
    "mortadelle": "mortadella",
    "mortaria": "mortarium",
    "moscardini": "moscardino",
    "motoscafi": "motoscafo",
    "mottoes": "motto",
    "mottos": "motto",
    "mousetraps": "mousetrap",
    "mouths": "mouth",
    "mucrones": "mucro",
    "mucros": "mucro",
    "mukim": "mukim",
    "mukims": "mukim",
    "multigravidae": "multigravida",
    "multigravidas": "multigravida",
    "mumps": "mumps",
    "municipia": "municipium",
    "muqdishos": "muqdisho",
    "muranese": "muranese",
    "murexes": "murex",
    "murices": "murex",
    "musae": "musa",
    "musas": "musa",
    "musees imaginaires": "musee imaginaire",
    "musiques concretes": "musique concrete",
    "mustachios": "mustachio",
    "muzzes": "muzz",
    "myae": "mya",
    "myas": "mya",
    "mycoplasmas": "mycoplasma",
    "mycoplasmata": "mycoplasma",
    "mycorhizae": "mycorhiza",
    "mycorhizas": "mycorhiza",
    "mycorrhizae": "mycorrhiza",
    "mycorrhizas": "mycorrhiza",
    "myelomas": "myeloma",
    "myelomata": "myeloma",
    "myos": "myo",
    "mythoi": "mythos",
    "myxomata": "myxoma",
    "nachtlokale": "nachtlokal",
    "nagaikas": "nagaika",
    "nagaiki": "nagaika",
    "nakayamans": "nakayaman",
    "nankingese": "nankingese",
    "nasuti": "nasutus",
    "naumachiae": "naumachia",
    "naumachias": "naumachia",
    "nauplii": "nauplius",
    "nautili": "nautilus",
    "nautiluses": "nautilus",
    "navarrese": "navarrese",
    "naye paise": "naya paisa",
    "nebenkerne": "nebenkern",
    "nebulae": "nebula",
    "nebulas": "nebula",
    "nectocalyces": "nectocalyx",
    "nemathecia": "nemathecium",
    "nenets": "nenets",
    "nentsi": "nenets",
    "nentsy": "nenets",
    "nepalese": "nepalese",
    "nephridia": "nephridium",
    "neritae": "nerita",
    "neritas": "nerita",
    "neumas": "neuma",
    "neumata": "neuma",
    "neurinomas": "neurinoma",
    "neurinomata": "neurinoma",
    "neuromata": "neuroma",
    "neurulae": "neurula",
    "neutrinos": "neutrino",
    "news": "news",
    "nexus": "nexus",
    "nexuses": "nexus",
    "ngwee": "ngwee",
    "niasese": "niasese",
    "nicobarese": "nicobarese",
    "nidi": "nidus",
    "niduses": "nidus",
    "nielli": "niello",
    "niellos": "niello",
    "nimbi": "nimbus",
    "nimbuses": "nimbus",
    "ningbos": "ningbo",
    "nipponese": "nipponese",
    "nockerln": "nockerl",
    "noctilucae": "noctiluca",
    "noda": "nodum",
    "nodi": "nodus",
    "noemata": "noema",
    "nomismata": "nomisma",
    "nonentia": "non-ens",
    "normans": "norman",
    "noumena": "noumenon",
    "nouveaux arrives": "nouvel arrive",
    "nouveaux pauvres": "nouveau pauvre",
    "nouveaux riches": "nouveau riche",
    "nouvelle pauvre": "nouveau pauvre",
    "nouvelle riche": "nouveau riche",
    "novae": "nova",
    "novas": "nova",
    "novi homines": "novus homo",
    "noxae": "noxa",
    "nubeculae": "nubecula",
    "nuclei": "nucleus",
    "nucleoli": "nucleolus",
    "nucleolonemas": "nucleolonema",
    "nucleolonemata": "nucleolonema",
    "nucleoluses": "nucleolus",
    "nucleuses": "nucleus",
    "numina": "numen",
    "nuntii": "nuntius",
    "nymphae": "nympha",
    "nymphaea": "nymphaeum",
    "nymphea": "nymphaeum",
    "oaths": "oath",
    "obeli": "obelus",
    "oboli": "obolus",
    "occipita": "occiput",
    "occiputs": "occiput",
    "oceanaria": "oceanarium",
    "ocelli": "ocellus",
    "octavos": "octavo",
    "octohedra": "octohedron",
    "octohedrons": "octohedron",
    "octopi": "octopus",
    "octopodes": "octopus",
    "octopuses": "octopus",
    "oculi": "oculus",
    "odea": "odeum",
    "oedemas": "oedema",
    "oedemata": "oedema",
    "offspring": "offspring",
    "oglys": "ogle",
    "oklahomans": "oklahoman",
    "olfactoria": "olfactorium",
    "omenta": "omentum",
    "ommatidia": "ommatidium",
    "onagers": "onager",
    "onagri": "onager",
    "opera magna": "opus magnum",
    "opercula": "operculum",
    "ophiophagi": "ophiophagus",
    "optima": "optimum",
    "optimums": "optimum",
    "opuscula": "opusculum",
    "opuses": "opus",
    "oratrices": "oratrix",
    "oreganos": "oregano",
    "organa": "organon",
    "orinocos": "orinoco",
    "orlandos": "orlando",
    "orthocerata": "orthoceras",
    "oscula": "osculum",
    "oslos": "oslo",
    "osmateria": "osmaterium",
    "osmeteria": "osmeterium",
    "ossicula": "ossiculum",
    "osteoclastomas": "osteoclastoma",
    "osteoclastomata": "osteoclastoma",
    "osteomata": "osteoma",
    "ostia": "ostium",
    "otomi": "otomi",
    "otomies": "otomi",
    "otos": "oto",
    "ottomans": "ottoman",
    "ours": "mine",
    "ourselves": "myself",
    "ova": "ovum",
    "ovaria": "ovarium",
    "ovoli": "ovolo",
    "ovula": "ovulum",
    "oxen": "ox",
    "oxybapha": "oxybaphon",
    "oxymora": "oxymoron",
    "oxymorons": "oxymoron",
    "oxyurides": "oxyuris",
    "paha": "paha",
    "paisas": "paisa",
    "paise": "paisa",
    "pajamas": "pajamas",
    "palae": "pala",
    "paleae": "palea",
    "pali": "palus",
    "pallia": "pallium",
    "palpebrae": "palpebra",
    "palpi": "palpus",
    "panamans": "panaman",
    "pangasinanes": "pangasinan",
    "pangasinans": "pangasinan",
    "panini": "panino",
    "panorpae": "panorpa",
    "pantos": "panto",
    "pants": "pant",
    "paparrazi": "paparrazo",
    "papillae": "papilla",
    "papillomata": "papilloma",
    "papulae": "papula",
    "papyri": "papyrus",
    "papyruses": "papyrus",
    "parabolae": "parabola",
    "parabolani": "parabolanus",
    "parabolas": "parabola",
    "paradores": "parador",
    "paradors": "parador",
    "paraganglia": "paraganglion",
    "paragangliomas": "paraganglioma",
    "paragangliomata": "paraganglioma",
    "paraglossae": "paraglossa",
    "paramaribos": "paramaribo",
    "paranuclei": "paranucleus",
    "parapleura": "parapleura",
    "parapsides": "parapsis",
    "paraselenae": "paraselene",
    "parcs fermes": "parc ferme",
    "parduscos": "pardusco",
    "parenchymata": "parenchyma",
    "parerga": "parergon",
    "parhelia": "parhelion",
    "parhelions": "parhelion",
    "parietes": "paries",
    "parisa": "parison",
    "parmese": "parmese",
    "parnassim": "parnas",
    "parvenus": "parvenu",
    "pasteurellae": "pasteurella",
    "pasteurellas": "pasteurella",
    "pastorales": "pastorale",
    "pastorali": "pastorale",
    "patagia": "patagium",
    "paterae": "patera",
    "pathoses": "pathos",
    "paths": "path",
    "paxillae": "paxilla",
    "peciae": "pecia",
    "pecias": "pecia",
    "pectens": "pecten",
    "pectines": "pecten",
    "pectora": "pectus",
    "pedalos": "pedalo",
    "pedes": "pes",
    "pedia": "pedion",
    "pedicellariae": "pedicellaria",
    "pediluvia": "pediluvium",
    "pekingese": "pekingese",
    "peltae": "pelta",
    "pelves": "pelvis",
    "pelvises": "pelvis",
    "pence": "penny",
    "pendula": "pendulum",
    "pendulums": "pendulum",
    "penes": "penis",
    "pengo": "pengo",
    "pengoes": "pengo",
    "penicilli": "penicillus",
    "peninsulae": "peninsula",
    "peninsulas": "peninsula",
    "penises": "penis",
    "pennia": "penni",
    "pennies": "penny",
    "pensieri": "pensiero",
    "people": "person",
    "percepta": "perceptum",
    "peridia": "peridium",
    "perikarya": "perikaryon",
    "perithecia": "perithecium",
    "perniones": "pernio",
    "peronia": "peronium",
    "perradii": "perradius",
    "personae": "persona",
    "personas": "persona",
    "persons": "person",
    "phalanges": "phalanx",
    "phalangia": "phalangium",
    "phalanxes": "phalanx",
    "phalli": "phallus",
    "phalloi": "phallos",
    "phalluses": "phallus",
    "pharmakoi": "pharmakos",
    "phenomena": "phenomenon",
    "phlegmasiae": "phlegmasia",
    "phocae": "phoca",
    "phocas": "phoca",
    "pholades": "pholas",
    "photos": "photo",
    "phototaxes": "phototaxis",
    "phragmata": "phragma",
    "phrenes": "phren",
    "phyla": "phylum",
    "phylae": "phyle",
    "phyllobranchiae": "phyllobranchia",
    "phylums": "phylum",
    "phymata": "phyma",
    "phyogemmaria": "phyogemmarium",
    "piani": "piano",
    "pianos": "piano",
    "pickerel": "pickerel",
    "pickerels": "pickerel",
    "piedmontese": "piedmontese",
    "pieds a terre": "pied a terre",
    "pieds noirs": "pied noir",
    "pietre": "pietra",
    "pifferari": "pifferaro",
    "pike": "pike",
    "pikes": "pike",
    "pilei": "pileus",
    "pili": "pilus",
    "pilleis": "pillie",
    "pimentos": "pimento",
    "pina coladas": "pina colada",
    "pinaces": "pinax",
    "pinas coladas": "pina colada",
    "pincers": "pincer",
    "pineta": "pinetum",
    "pinetums": "pinetum",
    "pingoes": "pingo",
    "pingos": "pingo",
    "pinnae": "pinna",
    "pinnas": "pinna",
    "pinones": "pinon",
    "pinons": "pinon",
    "pintos": "pinto",
    "pipkraker": "pipkrake",
    "pipkrakes": "pipkrake",
    "pirrites": "pyrites",
    "piscinae": "piscina",
    "piscinas": "piscina",
    "pistillidia": "pistillidium",
    "pistoiese": "pistoiese",
    "pithoi": "pithos",
    "pizzas": "pizza",
    "pizze": "pizza",
    "placeboes": "placebo",
    "placebos": "placebo",
    "placita": "placitum",
    "planctus": "planctus",
    "plankton": "plankter",
    "planulae": "planula",
    "plasmalemmae": "plasmalemma",
    "plasmalemmas": "plasmalemma",
    "plasmodia": "plasmodium",
    "plasmolytica": "plasmolyticum",
    "plasmomata": "plasmoma",
    "plateaus": "plateau",
    "plateaux": "plateau",
    "platypodes": "platypus",
    "platypuses": "platypus",
    "plecos": "pleco",
    "plectra": "plectrum",
    "plenaria": "plenarium",
    "plesiosauri": "plesiosaurus",
    "plether": "plethron",
    "plethra": "plethron",
    "pleura": "pleura",
    "plexus": "plexus",
    "plexuses": "plexus",
    "plicae": "plica",
    "pliers": "pliers",
    "ploshchadki": "ploshchadka",
    "pluralia tantum": "plurale tantum",
    "plutei": "pluteus",
    "plutos": "pluto",
    "pneumothoraces": "pneumothorax",
    "pneumothoraxes": "pneumothorax",
    "pobladores": "poblador",
    "pocula": "poculum",
    "poderi": "podere",
    "podetia": "podetium",
    "podia": "podium",
    "podiums": "podium",
    "poes": "po",
    "pogos": "pogo",
    "poise": "poise",
    "poises": "poise",
    "polises": "polis",
    "politbureaus": "politbureau",
    "politbureaux": "politbureau",
    "politicoes": "politico",
    "politicos": "politico",
    "pollices": "pollex",
    "pollinaria": "pollinarium",
    "pollinia": "pollinium",
    "polonese": "polonese",
    "polos": "polo",
    "polythecia": "polythecium",
    "pommes de terre": "pomme de terre",
    "ponchos": "poncho",
    "pontifexes": "pontifex",
    "pontifices": "pontifex",
    "porcini": "porcino",
    "poria": "porion",
    "porrones": "porron",
    "porrons": "porron",
    "porticoes": "portico",
    "porticos": "portico",
    "portiforia": "portiforium",
    "portmanteaus": "portmanteau",
    "portmanteaux": "portmanteau",
    "porto novos": "porto novo",
    "porto-novos": "porto-novo",
    "portos": "porto",
    "portraits parles": "portrait parle",
    "ports de bras": "port de bras",
    "portuguese": "portuguese",
    "pos": "po",
    "postnati": "postnatus",
    "postpubes": "postpubis",
    "postulata": "postulatum",
    "potatoes": "potato",
    "potestates": "potestas",
    "pox": "pox",
    "praecognita": "praecognitum",
    "pragmas": "pragma",
    "pragmata": "pragma",
    "precis": "precis",
    "precognita": "precognitum",
    "premiers danseurs": "premier danseur",
    "prepennae": "prepenna",
    "prima ballerinas": "prima ballerina",
    "prima donnas": "prima donna",
    "prime ballerine": "prima ballerina",
    "prime donne": "prima donna",
    "primigravidae": "primigravida",
    "primordia": "primordium",
    "principia": "principium",
    "proaves": "proavis",
    "proboscides": "proboscis",
    "proboscises": "proboscis",
    "proceedings": "proceedings",
    "procoeliae": "procoelia",
    "prodromi": "prodromus",
    "proferentes": "proferens",
    "profluvia": "profluvium",
    "proglottides": "proglottis",
    "programmata": "programma",
    "progymnasia": "progymnasium",
    "progymnasmata": "progymnasma",
    "prokarya": "prokaryon",
    "prolabia": "prolabium",
    "prolegomena": "prolegomenon",
    "promeropes": "promerops",
    "propleura": "propleuron",
    "propodia": "propodium",
    "propositi": "propositus",
    "propria": "proprium",
    "propyla": "propylon",
    "propylaea": "propylaeum",
    "propylons": "propylon",
    "prorae": "prora",
    "pros": "pro",
    "prosae": "prosa",
    "proscenia": "proscenium",
    "proscollae": "proscolla",
    "prosecutrices": "prosecutrix",
    "prosencephala": "prosencephalon",
    "proseuchae": "proseucha",
    "prosodiencephala": "prosodiencephalon",
    "prosopia": "prosopium",
    "prospectus": "prospectus",
    "prospectuses": "prospectus",
    "prosulae": "prosula",
    "protarsi": "protarsus",
    "prothalli": "prothallus",
    "prothallia": "prothallium",
    "protococci": "protococcus",
    "protomae": "protoma",
    "protomas": "protoma",
    "protonemata": "protonema",
    "prototypa": "prototypon",
    "prototypons": "prototypon",
    "protypa": "protypon",
    "provisoes": "proviso",
    "provisos": "proviso",
    "prytan": "prytanis",
    "prytanes": "prytanis",
    "psammomata": "psammoma",
    "pseudopodia": "pseudopodium",
    "pseudostomata": "pseudostoma",
    "psychos": "psycho",
    "psydracia": "psydracium",
    "pteromata": "pteroma",
    "pteropi": "pteropus",
    "pterylae": "pteryla",
    "pueblos": "pueblo",
    "puirt a beul": "port a beul",
    "puirt-a-beul": "port-a-beul",
    "puli": "pul",
    "pulli": "pullus",
    "pulmonaria": "pulmonarium",
    "puls": "pul",
    "pulsella": "pulsellum",
    "pulveratrices": "pulveratrix",
    "puncta": "punctum",
    "punti blanchi": "punto blanco",
    "puntos blancos": "punto blanco",
    "pupae": "pupa",
    "purlieus": "purlieu",
    "purlieux": "purlieu",
    "purum": "purum",
    "purums": "purum",
    "putti": "putto",
    "puttonys": "puttony",
    "pycnia": "pycnium",
    "pycnides": "pycnis",
    "pycnidia": "pycnidium",
    "pyjamas": "pyjamas",
    "pylae": "pyla",
    "pyralides": "pyralis",
    "pyramidia": "pyramidion",
    "pyramidions": "pyramidion",
    "pyrexiae": "pyrexia",
    "pyritae": "pyrites",
    "pyrophori": "pyrophorus",
    "pyxidia": "pyxidium",
    "quadratrices": "quadratrix",
    "quaesita": "quaesitum",
    "quanta": "quantum",
    "quantums": "quantum",
    "quartos": "quarto",
    "quartzes": "quartz",
    "quid": "quid",
    "quinarii": "quinarius",
    "quinquennia": "quinquennium",
    "quitos": "quito",
    "rabies": "rabies",
    "rad": "rad",
    "radialia": "radiale",
    "radices": "radix",
    "radii": "radius",
    "radiuses": "radius",
    "radixes": "radix",
    "rads": "rad",
    "ragazze": "ragazza",
    "ragazzi": "ragazzo",
    "raggare": "raggare",
    "raggares": "raggare",
    "rami": "ramus",
    "ramuli": "ramulus",
    "ranunculi": "ranunculus",
    "ranunculuses": "ranunculus",
    "rara avises": "rara avis",
    "rarae aves": "rara avis",
    "razzes": "razz",
    "recti": "rectus",
    "rediae": "redia",
    "reductiones": "reductio",
    "referenda": "referendum",
    "referendums": "referendum",
    "refugia": "refugium",
    "regidores": "regidor",
    "regidors": "regidor",
    "regmata": "regma",
    "reguli": "regulus",
    "reiche": "reich",
    "rejoneadores": "rejoneador",
    "rejones": "rejon",
    "relata": "relatum",
    "rem": "rem",
    "remexes": "remex",
    "remiges": "remex",
    "remou": "remous",
    "remous": "remous",
    "rems": "rem",
    "repla": "replum",
    "representamens": "representamen",
    "representamina": "representamen",
    "res": "res",
    "residua": "residuum",
    "residuums": "residuum",
    "responsa": "responsum",
    "retia": "rete",
    "retinacula": "retinaculum",
    "retinulae": "retinula",
    "rezidentsii": "rezidentsia",
    "rezidenty": "rezident",
    "rhinoceros": "rhinoceros",
    "rhinoceroses": "rhinoceros",
    "rhinophymas": "rhinophyma",
    "rhinophymata": "rhinophyma",
    "rhinos": "rhino",
    "rhizomas": "rhizoma",
    "rhizomata": "rhizoma",
    "rhododendra": "rhododendron",
    "rhododendrons": "rhododendron",
    "rhombohedra": "rhombohedron",
    "rhombohedrons": "rhombohedron",
    "rhomboidei": "rhomboideus",
    "rhopalia": "rhopalium",
    "rhynchodaea": "rhynchodaeum",
    "rhythmi": "rhythmus",
    "ri": "ri",
    "ricercari": "ricercar",
    "ricercars": "ricercar",
    "ricercate": "ricercata",
    "ricordi": "ricordo",
    "rideaus": "rideau",
    "rideaux": "rideau",
    "rilievos": "rilievo",
    "rimmonim": "rimon",
    "rimonim": "rimon",
    "rioni": "rione",
    "risottos": "risotto",
    "ritardandi": "ritardando",
    "ritardandos": "ritardando",
    "rites de passage": "rite de passage",
    "rococos": "rococo",
    "roe": "roe",
    "roes": "roe",
    "roma": "rom",
    "romagnese": "romagnese",
    "romanese": "romanese",
    "romanies": "romany",
    "romans": "roman",
    "romanzen": "romanze",
    "rond de jambes": "rond de jambe",
    "rondeaus": "rondeau",
    "rondeaux": "rondeau",
    "rondos": "rondo",
    "ronds de jambes": "rond de jambe",
    "rostra": "rostrum",
    "rostrums": "rostrum",
    "rotulae": "rotula",
    "routes nationales": "route nationale",
    "rubashkas": "rubashka",
    "rubashki": "rubashka",
    "rugae": "ruga",
    "ruses de guerre": "ruse de guerre",
    "ryper": "rype",
    "sabertooths": "sabertooth",
    "sacculi": "sacculus",
    "sacella": "sacellum",
    "sachverhalte": "sachverhalt",
    "sacra": "sacrum",
    "sacramentos": "sacramento",
    "sacraria": "sacrarium",
    "sacrums": "sacrum",
    "saddos": "saddo",
    "saffi": "saffo",
    "saga": "sagum",
    "sagos": "sago",
    "salitas": "salita",
    "salite": "salita",
    "salivaria": "salivarium",
    "salmon": "salmon",
    "salpae": "salpa",
    "salpas": "salpa",
    "salvos": "salvo",
    "samboes": "sambo",
    "sambos": "sambo",
    "samurai": "samurai",
    "sanatoria": "sanatorium",
    "sanatoriums": "sanatorium",
    "sancta sanctorum": "sanctum sanctorum",
    "sandr": "sandr",
    "sandrs": "sandr",
    "sangirese": "sangirese",
    "santiagos": "santiago",
    "sapporos": "sapporo",
    "sarajevos": "sarajevo",
    "sarawakese": "sarawakese",
    "sarcinae": "sarcina",
    "sarcomas": "sarcoma",
    "sarcomata": "sarcoma",
    "sarcophagi": "sarcophagus",
    "sarissae": "sarissa",
    "sassafrases": "sassafras",
    "satellites": "satelles",
    "saxicavae": "saxicava",
    "sbirri": "sbirro",
    "scaleni": "scalenus",
    "scandala magnatum": "scandalum magnatum",
    "scapi": "scapus",
    "scapulae": "scapula",
    "scapulas": "scapula",
    "scarabaei": "scarabaeus",
    "scari": "scarus",
    "scazons": "scazon",
    "scazontes": "scazon",
    "sceattas": "sceat",
    "schemas": "schema",
    "schemata": "schema",
    "scherzandos": "scherzando",
    "scherzos": "scherzo",
    "schismata": "schisma",
    "schistosomula": "schistosomulum",
    "schizostylis": "schizostylis",
    "scholia": "scholium",
    "scioli": "sciolus",
    "scirrhi": "scirrhus",
    "scirrhusses": "scirrhus",
    "scissors": "scissors",
    "scoleces": "scolex",
    "scolopidia": "scolopidium",
    "scombri": "scomber",
    "scoriae": "scoria",
    "scorias": "scoria",
    "scotomas": "scotoma",
    "scotomata": "scotoma",
    "scriptoria": "scriptorium",
    "scriptoriums": "scriptorium",
    "scudi": "scudo",
    "scungilli": "scungille",
    "scuole": "scuola",
    "scuta": "scutum",
    "scutella": "scutellum",
    "scutula": "scutulum",
    "scyphi": "scyphus",
    "seaquaria": "seaquarium",
    "secantes": "secans",
    "secreta": "secretum",
    "secundagravidae": "secundagravida",
    "secundagravidas": "secundagravida",
    "secundigravidae": "secundigravida",
    "secundigravidas": "secundigravida",
    "secundiparae": "secundipara",
    "secundiparas": "secundipara",
    "sedes": "sedes",
    "sedilia": "sedile",
    "seecatchie": "seecatch",
    "seed": "seed",
    "seeds": "seed",
    "selmans": "selman",
    "semas": "sema",
    "semata": "sema",
    "seminia": "seminium",
    "semunciae": "semuncia",
    "senarii": "senarius",
    "senegalese": "senegalese",
    "senes": "senex",
    "sensa": "sensum",
    "sensedata": "sensedatum",
    "sensilla": "sensillum",
    "sensoriola": "sensoriolum",
    "sententiae": "sententia",
    "separatrices": "separatrix",
    "sephardim": "sephardi",
    "sephardin": "sephardi",
    "septa": "septum",
    "septaria": "septarium",
    "septemviri": "septemvir",
    "septenarii": "septenarius",
    "sequelae": "sequela",
    "sequestra": "sequestrum",
    "sera": "serum",
    "seraphim": "seraph",
    "seraphs": "seraph",
    "sericteria": "sericterium",
    "series": "series",
    "serir": "serir",
    "serirs": "serir",
    "serpigines": "serpigo",
    "serpigoes": "serpigo",
    "serpulae": "serpula",
    "serrae": "serra",
    "sertulariae": "sertularia",
    "sertularias": "sertularia",
    "serums": "serum",
    "sestertia": "sestertium",
    "sestieri": "sestiere",
    "set offs": "set off",
    "set outs": "set out",
    "set tos": "set to",
    "set-offs": "set-off",
    "set-outs": "set-out",
    "set-tos": "set-to",
    "setae": "seta",
    "sets off": "set off",
    "sets out": "set out",
    "sets to": "set to",
    "sets-off": "set-off",
    "sets-out": "set-out",
    "sets-to": "set-to",
    "sezessionen": "sezession",
    "sforzati": "sforzato",
    "sforzatos": "sforzato",
    "sgraffiati": "sgraffiato",
    "shad": "shad",
    "shads": "shad",
    "shamans": "shaman",
    "shauries": "shauri",
    "shauris": "shauri",
    "shavese": "shavese",
    "shawabtis": "shawabti",
    "shawabtiu": "shawabti",
    "shawanese": "shawanese",
    "shears": "shears",
    "sheep": "sheep",
    "shelves": "shelf",
    "shifta": "shifta",
    "shiftas": "shifta",
    "shochetim": "shochet",
    "shomrim": "shomer",
    "shorts": "shorts",
    "siamese": "siamese",
    "siemens": "siemens",
    "sienese": "sienese",
    "sigli": "siglos",
    "sigloi": "siglos",
    "signa": "signum",
    "sikkimese": "sikkimese",
    "silbadores": "silbador",
    "silbadors": "silbador",
    "sileni": "silenus",
    "silices": "silex",
    "siliquae": "siliqua",
    "silos": "silo",
    "siluri": "silurus",
    "simiae": "simia",
    "simplexes": "simplex",
    "simplices": "simplex",
    "simpula": "simpulum",
    "simulacra": "simulacrum",
    "simulacrums": "simulacrum",
    "sinfonias": "sinfonia",
    "sinfonie": "sinfonia",
    "singhalese": "singhalese",
    "sinhalese": "sinhalese",
    "sinus": "sinus",
    "sinuses": "sinus",
    "siphunculi": "siphunculus",
    "sipunculi": "sipunculus",
    "siroccos": "sirocco",
    "sistra": "sistrum",
    "sistrums": "sistrum",
    "situlae": "situla",
    "situlas": "situla",
    "sizzes": "sizz",
    "skis": "ski",
    "snekkjur": "snekkja",
    "snipe": "snipe",
    "snipes": "snipe",
    "snoozes": "snooze",
    "soes": "so",
    "sogdianese": "sogdianese",
    "solaria": "solarium",
    "soldi": "soldo",
    "solenia": "solenium",
    "soles": "sol",
    "solfeggi": "solfeggio",
    "solfeggios": "solfeggio",
    "soli": "solo",
    "solidi": "solidus",
    "solidos": "solidus",
    "solos": "solo",
    "solstitia": "solstitium",
    "solvi": "solvus",
    "somas": "soma",
    "somata": "soma",
    "somatenes": "somaten",
    "sombreros": "sombrero",
    "some": "one",
    "sonamans": "sonaman",
    "soprani": "soprano",
    "sopranos": "soprano",
    "soralia": "soralium",
    "soredia": "soredium",
    "sori": "sorus",
    "soshi": "soshi",
    "sotches": "sotch",
    "sotchs": "sotch",
    "souvlakia": "souvlaki",
    "souvlakis": "souvlaki",
    "spadices": "spadix",
    "spadixes": "spadix",
    "spaetlesen": "spaetlese",
    "spari": "sparus",
    "spasmi": "spasmus",
    "spathae": "spatha",
    "species": "species",
    "spectra": "spectrum",
    "spectrums": "spectrum",
    "specula": "speculum",
    "speculums": "speculum",
    "spermaria": "spermarium",
    "spermata": "sperma",
    "spermathecae": "spermatheca",
    "spermatia": "spermatium",
    "spermatoa": "spermatoon",
    "spermogonia": "spermogonium",
    "sphaerideum": "sphaeridium",
    "sphaeridia": "sphaeridium",
    "sphagna": "sphagnum",
    "sphagnums": "sphagnum",
    "spheges": "sphex",
    "sphinges": "sphinx",
    "sphinxes": "sphinx",
    "spicula": "spiculum",
    "spiracula": "spiraculum",
    "spirilla": "spirillum",
    "spondyli": "spondylus",
    "sporodochia": "sporodochium",
    "sputa": "sputum",
    "squadre": "squadra",
    "squadristi": "squadrist",
    "squadrists": "squadrist",
    "squali": "squalus",
    "squamae": "squama",
    "squillae": "squilla",
    "stabilimenta": "stabilimentum",
    "staccatos": "staccato",
    "stadia": "stadium",
    "stadiums": "stadium",
    "stalactitae": "stalactite",
    "stalactites": "stalactite",
    "stamen": "stamen",
    "stamina": "stamen",
    "staminidia": "staminidium",
    "staphylococci": "staphylococcus",
    "star fish": "star fish",
    "star fishes": "star fish",
    "star-fish": "star-fish",
    "star-fishes": "star-fish",
    "starfish": "starfish",
    "starfishes": "starfish",
    "starosti": "starosta",
    "starovers": "starover",
    "starovery": "starover",
    "startsy": "starets",
    "startzy": "staretz",
    "stasima": "stasimon",
    "stasimons": "stasimon",
    "staterae": "statera",
    "statuae": "statua",
    "statuas": "statua",
    "status": "status",
    "statuses": "status",
    "stelae": "stela",
    "stelas": "stela",
    "stellae": "stella",
    "stemmata": "stemma",
    "stephania": "stephanion",
    "stephanions": "stephanion",
    "sterigmata": "sterigma",
    "sterna": "sternum",
    "sternebrae": "sternebra",
    "sternos": "sterno",
    "sternums": "sternum",
    "stichidia": "stichidium",
    "stichoi": "stichos",
    "stigmariae": "stigmaria",
    "stigmas": "stigma",
    "stigmata": "stigma",
    "stilettoes": "stiletto",
    "stillicidia": "stillicidium",
    "stimuli": "stimulus",
    "stipites": "stipes",
    "stipulae": "stipula",
    "stipulas": "stipula",
    "stirpes": "stirps",
    "stoai": "stoa",
    "stoas": "stoa",
    "stolones": "stolo",
    "stomachs": "stomach",
    "stomas": "stoma",
    "stomata": "stoma",
    "stomatia": "stomatium",
    "stomia": "stomion",
    "stomodaea": "stomodaeum",
    "stomodea": "stomodaeum",
    "storeys": "storey",
    "stornelli": "stornello",
    "strambotti": "strambotto",
    "strata": "stratum",
    "strategi": "strategus",
    "strategoi": "strategos",
    "stratums": "stratum",
    "streptococci": "streptococcus",
    "streptothrices": "streptothrix",
    "striae": "stria",
    "striata": "striatum",
    "strigae": "striga",
    "striolae": "striola",
    "strobilae": "strobila",
    "strobili": "strobilus",
    "stromata": "stroma",
    "stromodaea": "stromodeum",
    "stromodea": "stromodeum",
    "strumae": "struma",
    "stryddae": "stryddag",
    "stuccos": "stucco",
    "studia generali": "studium generale",
    "styli": "stylus",
    "stylopodia": "stylopodium",
    "stylos": "stylo",
    "styluses": "stylus",
    "subbotniki": "subbotnik",
    "subbotniks": "subbotnik",
    "subsellia": "subsellium",
    "subsidia": "subsidium",
    "substrata": "substratum",
    "substratums": "substratum",
    "succi": "succus",
    "succineae": "succinea",
    "succineas": "succinea",
    "succubi": "succubus",
    "succubuses": "succubus",
    "sudanese": "sudanese",
    "suffrutices": "suffrutex",
    "suggesta": "suggestum",
    "suggestiones falsi": "suggestio falsi",
    "suggestums": "suggestum",
    "sulci": "sulcus",
    "sum totals": "sum total",
    "sum-totals": "sum-total",
    "summa bona": "summum bonum",
    "summa genera": "summum genus",
    "summae": "summa",
    "sumos": "sumo",
    "sums total": "sum total",
    "sums-total": "sum-total",
    "sundanese": "sundanese",
    "supercilia": "supercilium",
    "superficies": "superficies",
    "supernovae": "supernova",
    "supernovas": "supernova",
    "superstrata": "superstratum",
    "supposita": "suppositum",
    "supracleithra": "supracleithrum",
    "suprascapulae": "suprascapula",
    "surculi": "surculus",
    "surimonos": "surimono",
    "sustentacula": "sustentaculum",
    "swahilese": "swahilese",
    "swine": "swine",
    "swines": "swine",
    "switchfoots": "switchfoot",
    "syllabaria": "syllabarium",
    "syllabi": "syllabus",
    "syllabuses": "syllabus",
    "symplasmata": "symplasma",
    "sympodia": "sympodium",
    "symposia": "symposium",
    "symposiums": "symposium",
    "synaesthesiae": "synaesthesia",
    "synangia": "synangium",
    "synapticulae": "synapticula",
    "synapticulas": "synapticula",
    "synaxaria": "synaxarium",
    "synaxes": "synaxis",
    "syncelli": "syncellus",
    "syncytia": "syncytium",
    "syndendria": "syndendrium",
    "synechiae": "synechia",
    "synedria": "synedrium",
    "synesthesiae": "synesthesia",
    "syntagmas": "syntagma",
    "syntagmata": "syntagma",
    "syntaxes": "syntaxis",
    "synthroni": "synthronus",
    "synusiae": "synusia",
    "syphilomata": "syphiloma",
    "syringes": "syringe",
    "syrinxes": "syrinx",
    "syrphi": "syrphus",
    "syrtes": "syrtis",
    "tabellae": "tabella",
    "tableaus": "tableau",
    "tableaux": "tableau",
    "tablina": "tablinum",
    "tabulae": "tabula",
    "tacomans": "tacoman",
    "taeniae": "taenia",
    "taenias": "taenia",
    "tagmata": "tagma",
    "taikos": "taiko",
    "taleae": "talea",
    "tali": "talus",
    "talismans": "talisman",
    "talouses": "talouse",
    "tangoes": "tango",
    "tangos": "tango",
    "tarantate": "tarantata",
    "tarantati": "tarantato",
    "tarsi": "tarsus",
    "taxa": "taxon",
    "taxes": "tax",
    "tazze": "tazza",
    "teal": "teal",
    "teals": "teal",
    "technos": "techno",
    "teeth": "tooth",
    "tegmenta": "tegmentum",
    "tegmina": "tegmen",
    "tegulae": "tegula",
    "telamones": "telamon",
    "telia": "telium",
    "temblores": "temblor",
    "tempi": "tempo",
    "tempietti": "tempietto",
    "tempos": "tempo",
    "tenacula": "tenaculum",
    "tenderfoots": "tenderfoot",
    "tenggerese": "tenggerese",
    "teniae": "tenia",
    "tenias": "tenia",
    "tenorini": "tenorino",
    "tentacula": "tentaculum",
    "tentamina": "tentamen",
    "tenues": "tenuis",
    "tepidaria": "tepidarium",
    "teratomata": "teratoma",
    "terebella": "terebellum",
    "terebratulae": "terebratula",
    "terebratulas": "terebratula",
    "teredines": "teredo",
    "teredos": "teredo",
    "terga": "tergum",
    "termini": "terminus",
    "terminuses": "terminus",
    "termites": "termes",
    "terpen": "terp",
    "terps": "terp",
    "terrae filii": "terrae filius",
    "terraria": "terrarium",
    "terrariums": "terrarium",
    "terrazzos": "terrazzo",
    "terremare": "terramara",
    "territoria": "territorium",
    "terzetti": "terzetto",
    "tessellae": "tessella",
    "tessellas": "tessella",
    "tesserae": "tessera",
    "testes": "testes",
    "testudos": "testudo",
    "tetanother": "tetanothrum",
    "tetanothra": "tetanothrum",
    "tete a tetes": "tete a tete",
    "tete-a-tetes": "tete-a-tete",
    "tetracola": "tetracolon",
    "tetragrammata": "tetragrammaton",
    "tetrahedra": "tetrahedron",
    "tetrahedrons": "tetrahedron",
    "tetrapleura": "tetrapleuron",
    "tetrapleurons": "tetrapleuron",
    "tetrasporangia": "tetrasporangium",
    "thalami": "thalamus",
    "thalamoi": "thalamos",
    "thanatoria": "thanatorium",
    "thaumaturgi": "thaumaturgus",
    "thebe": "thebe",
    "thecae": "theca",
    "theirs": "its",
    "thelyca": "thelycum",
    "them": "it",
    "themata": "thema",
    "themselves": "itself",
    "theologoumena": "theologoumenon",
    "thesauri": "thesaurus",
    "thesauruses": "thesaurus",
    "these": "this",
    "theses": "thesis",
    "they": "it",
    "thieves": "thief",
    "thoes": "thos",
    "tholi": "tholus",
    "tholoi": "tholos",
    "thoraces": "thorax",
    "thoraxes": "thorax",
    "those": "that",
    "thrasoes": "thraso",
    "thrasones": "thraso",
    "thymi": "thymus",
    "thymiateria": "thymiaterion",
    "thyridia": "thyridium",
    "thyrsi": "thyrsus",
    "tibiae": "tibia",
    "timorese": "timorese",
    "timpanos": "timpano",
    "tintinnabula": "tintinnabulum",
    "tipulae": "tipula",
    "tirolese": "tirolese",
    "tiros": "tiro",
    "tobaccos": "tobacco",
    "tocharese": "tocharese",
    "toes": "toe",
    "togos": "togo",
    "tokyos": "tokyo",
    "tolkachi": "tolkach",
    "tomatoes": "tomato",
    "tombaroli": "tombarolo",
    "tomia": "tomium",
    "tondi": "tondo",
    "tongs": "tongs",
    "tonkinese": "tonkinese",
    "tonneaus": "tonneau",
    "tonneaux": "tonneau",
    "topazes": "topaz",
    "topoi": "topos",
    "toreros": "torero",
    "tori": "torus",
    "toriles": "toril",
    "torinese": "torinese",
    "torinos": "torino",
    "torma": "torma",
    "tormas": "torma",
    "tornesi": "tornese",
    "torni": "tornus",
    "torontos": "toronto",
    "torsos": "torso",
    "tortillas espanolas": "tortilla espanola",
    "tortrices": "tortrix",
    "torulae": "torula",
    "toruli": "torulus",
    "toruses": "torus",
    "tours de force": "tour de force",
    "toxica": "toxicum",
    "toxii": "toxius",
    "toxoplasmata": "toxoplasma",
    "trabeae": "trabea",
    "trabeculae": "trabecula",
    "tracheae": "trachea",
    "tractatrices": "tractatrix",
    "tractella": "tractellum",
    "tractrices": "tractrix",
    "tragi": "tragus",
    "trapezia": "trapezium",
    "trapezii": "trapezius",
    "trapeziums": "trapezium",
    "trapezohedra": "trapezohedron",
    "trapezohedrons": "trapezohedron",
    "trattorias": "trattoria",
    "trattorie": "trattoria",
    "traumas": "trauma",
    "traumata": "trauma",
    "trellises": "trellis",
    "tremolos": "tremolo",
    "treponemata": "treponema",
    "trichidia": "trichidium",
    "trichinae": "trichina",
    "trichine": "trichina",
    "trichomata": "trichoma",
    "triclinia": "triclinium",
    "trientes": "triens",
    "triforia": "triforium",
    "trilbys": "trilby",
    "tripla": "triplum",
    "triradii": "triradius",
    "triumviri": "triumvir",
    "triumvirs": "triumvir",
    "trochi": "trochus",
    "trochuses": "trochus",
    "trockenbeerenauslesen": "trockenbeerenauslese",
    "tropaeola": "tropaeolum",
    "tropaeolums": "tropaeolum",
    "troparia": "troparion",
    "tropides": "tropis",
    "trousers": "trousers",
    "trousseaus": "trousseau",
    "trousseaux": "trousseau",
    "trout": "trout",
    "trulli": "trullo",
    "trumeaux": "trumeau",
    "tsuba": "tsuba",
    "tsubas": "tsuba",
    "tsubo": "tsubo",
    "tsubos": "tsubo",
    "tubercula": "tuberculum",
    "tuberculomas": "tuberculoma",
    "tuberculomata": "tuberculoma",
    "tuberes": "tuber",
    "tubuli": "tubulus",
    "tumuli": "tumulus",
    "tuna": "tuna",
    "turbot": "turbot",
    "turbots": "turbot",
    "turfs": "turf",
    "turrones": "turron",
    "turrons": "turron",
    "turves": "turf",
    "tympana": "tympanum",
    "typos": "typo",
    "tyrolese": "tyrolese",
    "tyros": "tyro",
    "udarniki": "udarnik",
    "ulpanim": "ulpan",
    "ultimata": "ultimatum",
    "ultimatums": "ultimatum",
    "umbilici": "umbilicus",
    "umbilicuses": "umbilicus",
    "umbones": "umbo",
    "umbos": "umbo",
    "umbrae": "umbra",
    "umbras": "umbra",
    "umwelten": "umwelt",
    "unci": "uncus",
    "unciae": "uncia",
    "uncini": "uncinus",
    "ungues": "unguis",
    "uniones": "unio",
    "unios": "unio",
    "uomini": "uomo",
    "uomini universali": "uomo universale",
    "uraei": "uraeus",
    "uredia": "uredium",
    "uredinia": "uredinium",
    "uredosori": "uredosorus",
    "uri": "urus",
    "urodaea": "urodaeum",
    "urodea": "urodeum",
    "uruses": "urus",
    "us": "me",
    "usneae": "usnea",
    "usneas": "usnea",
    "ustilagines": "ustilago",
    "uteri": "uterus",
    "uteruses": "uterus",
    "utriculariae": "utricularia",
    "uvae": "uva",
    "vaalpens": "vaalpens",
    "vaalpensen": "vaalpens",
    "vacua": "vacuum",
    "vacuums": "vacuum",
    "vagi": "vagus",
    "vaginae": "vagina",
    "vaginas": "vagina",
    "vaginulae": "vaginula",
    "valleculae": "vallecula",
    "valute": "valuta",
    "valuten": "valuta",
    "valvulae": "valvula",
    "vaporetti": "vaporetto",
    "vaporettos": "vaporetto",
    "vaqueros": "vaquero",
    "varellaes": "varella",
    "varely": "varella",
    "variae lectiones": "varia lectio",
    "varices": "varix",
    "vasa": "vas",
    "vascula": "vasculum",
    "vasculitides": "vasculitis",
    "vasculitises": "vasculitis",
    "vasculums": "vasculum",
    "vedutas": "veduta",
    "vedute": "veduta",
    "veg": "veg",
    "veges": "veg",
    "vehicula": "vehiculum",
    "vela": "velum",
    "velamina": "velamen",
    "velaria": "velarium",
    "velums": "velum",
    "venae": "vena",
    "veneres": "venus",
    "venuses": "venus",
    "vermicellos": "vermicello",
    "vermiculi": "vermiculus",
    "vermontese": "vermontese",
    "veronese": "veronese",
    "verrucae": "verruca",
    "vers libres": "vers libre",
    "versiculi": "versiculus",
    "versos": "verso",
    "vertebrae": "vertebra",
    "vertexes": "vertex",
    "vertices": "vertex",
    "verticilli": "verticillus",
    "vesiculae": "vesicula",
    "vestigia": "vestigium",
    "vetturini": "vetturino",
    "viali": "viale",
    "viatica": "viaticum",
    "vibices": "vibex",
    "vibracula": "vibraculum",
    "vibratos": "vibrato",
    "vibriones": "vibrio",
    "vibrios": "vibrio",
    "viennese": "viennese",
    "vietnamese": "vietnamese",
    "vilas": "vila",
    "vile": "vila",
    "villanelle": "villanella",
    "villi": "villus",
    "villini": "villino",
    "vincula": "vinculum",
    "violette": "violetta",
    "violoncellos": "violoncello",
    "vires": "vis",
    "virgae": "virga",
    "virgos": "virgo",
    "viri": "virus",
    "virii": "virus",
    "virtuosi": "virtuoso",
    "virtuosos": "virtuoso",
    "viruses": "virus",
    "viscera": "viscus",
    "vitellaria": "vitellarium",
    "vittae": "vitta",
    "vivaria": "vivarium",
    "vivariums": "vivarium",
    "vlakten": "vlakte",
    "vlaktes": "vlakte",
    "voeux": "voeu",
    "volcanelli": "volcanello",
    "volcanoes": "volcano",
    "volcanos": "volcano",
    "vomicae": "vomica",
    "vomicas": "vomica",
    "vomitoria": "vomitorium",
    "vomitoriums": "vomitorium",
    "voorbokke": "voorbok",
    "vorspiele": "vorspiel",
    "vorstellungen": "vorstellung",
    "vortexes": "vortex",
    "vortices": "vortex",
    "wadies": "wadi",
    "wadis": "wadi",
    "wadys": "wady",
    "waldoes": "waldo",
    "waldos": "waldo",
    "waltzes": "waltz",
    "wanderjahre": "wanderjahr",
    "wandervogel": "wandervogel",
    "washers up": "washer up",
    "washers-up": "washer-up",
    "we": "I",
    "weirdos": "weirdo",
    "wenchowese": "wenchowese",
    "wharfs": "wharf",
    "wharves": "wharf",
    "what": "what",
    "whatever": "whatever",
    "whatsoever": "whatsoever",
    "white staves": "white staff",
    "whiting": "whiting",
    "whitmanese": "whitmanese",
    "who": "who",
    "whoever": "whoever",
    "whom": "whom",
    "whomever": "whomever",
    "whomsoever": "whomsoever",
    "whose": "whose",
    "whosesoever": "whosesoever",
    "whosever": "whosever",
    "whosoever": "whosoever",
    "wildebeest": "wildebeest",
    "wildebeests": "wildebeest",
    "wives": "wife",
    "wolves": "wolf",
    "women children": "woman child",
    "women servants": "woman servant",
    "women-children": "woman-child",
    "women-servants": "woman-servant",
    "womenchildren": "womanchild",
    "wreaths": "wreath",
    "xanthomas": "xanthoma",
    "xanthomata": "xanthoma",
    "xiphiplastra": "xiphiplastron",
    "xiphopagi": "xiphopagus",
    "xiphoplastra": "xiphoplastron",
    "xoana": "xoanon",
    "xysti": "xystus",
    "yakimans": "yakiman",
    "yamoussoukros": "yamoussoukro",
    "yarramans": "yarraman",
    "yarramen": "yarraman",
    "yengeese": "yengeese",
    "yeses": "yes",
    "yo yos": "yo yo",
    "yo-yos": "yo-yo",
    "yoes": "yo",
    "yokohamans": "yokohaman",
    "you": "you",
    "yours": "yours",
    "yourselves": "yourself",
    "yumans": "yuman",
    "yunnanese": "yunnanese",
    "yuon": "yuo",
    "zadrugas": "zadruga",
    "zadruge": "zadruga",
    "zeitgebers": "zeitgeber",
    "zemirot": "zemirah",
    "zemiroth": "zemirah",
    "zeppoli": "zeppole",
    "zeros": "zero",
    "zibos": "zibo",
    "zingani": "zingano",
    "zite": "zita",
    "zitelle": "zitella",
    "ziti": "zita",
    "zizzes": "zizz",
    "zloty": "zloty",
    "zlotys": "zloty",
    "zoaeae": "zoaea",
    "zoaria": "zoarium",
    "zoeae": "zoea",
    "zoeas": "zoea",
    "zooeae": "zooea",
    "zooecia": "zooecium",
    "zoophori": "zoophorus",
    "zoosporangia": "zoosporangium",
    "zophori": "zophorus",
    "zucche": "zucca",
    "zucchini": "zucchini",
    "zucchinis": "zucchini",
    "zuppe": "zuppa",
    "zuzim": "zuz",
    "zuzzes": "zuz",
    "zyga": "zygon",
    "zygantra": "zygantrum",
    "zygomas": "zygoma",
    "zygomata": "zygoma"
}

modern_plural_convert_rule_regex = rei(r"^(?:(.*?)-general|(.*?) general|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) it|son-of-a-(.*?)|son of a (.*?)|(.*?)-(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)|(.*?) (about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)|(.*?)-errant|(.*?) errant|(.*?)-(above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)|(.*?) (above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)|(.*)staff|(.*)stave|(.*)genus|(.*)person|(.*)hertz|(.*)siemens|(.*)brother|(.*)child|(.*)cow|(.*)mensch|(.*)elf|(.*)hoof|(.*)knife|(.*)leaf|(.*)loaf|(.*)shelf|(.*)thief|(.*)wife|(.*)wolf|(.*)human|(.*)foot|(.*)goose|(.*)man|(.*)mouse|(.*)tooth|(.*)buzz|(.*)fizz|(.*)frizz|(.*)fuzz|(.*)jazz|(.*)proboscis|(.*)bema|(.*)drama|(.*)edema|(.*)lemma|(.*)sarcoma|(.*)stoma|(.*)nova|(.*)umbra|(.*)lumen|(.*)datum|(.*)stratum|(.*)bacillus|(.*)nucleus|(.+)thesis|(.*)hedron|(.*)helion|(.*)legomenon|(.*)phenomenon|(.*)helix|(.*)measles|(.*)pox|(.*)bass|(.*)deer|(.*)fish|(.*)fowl|(.*)sheep|(.*)wildebeest|(.*)craft|(.+)star general|(.+)oux|(.+[aeiou])o|(.+[aeo])lf|(.+)nge|(.+[aiy])nx|(.+)arf|(.+)ceps|(.+[^aeoui][aeouiy])che|(.+[cs])h|(.+)eaf|(.+)eau|(.+)ieu|(.+)nife|(.+)oe|(.+)o|(.+)quy|(.+[aeiou])y|(.+[rnlpwaeio])se|(.+)ss|(.+[aeo])use|(.+)um|(.+)us|(.+[^ns])sis|(.+)trix|(.+)x|(.+)y|(.+[^aeiouy])z|(.+)z|(.+)zoon|(.+)s|(.+))$")

modern_plural_convert_outputs = [
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-general",
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} general",
    lambda subterms: f"{subterms[0]} them",
    lambda subterms: f"sons-of-{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}",
    lambda subterms: f"sons of {convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}",
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-{subterms[1]}-{subterms[2]}",
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} {subterms[1]} {subterms[2]}",
    lambda subterms: f"{subterms[0]}-{convert_to_modern_plural(subterms[1]) if is_singular(subterms[1]) else subterms[1]}",
    lambda subterms: f"{subterms[0]} {convert_to_modern_plural(subterms[1]) if is_singular(subterms[1]) else subterms[1]}",
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-errant",
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} errant",
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-{subterms[1]}",
    lambda subterms: f"{convert_to_modern_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} {subterms[1]}",
    lambda subterms: f"{subterms[0]}staffs",
    lambda subterms: f"{subterms[0]}staves",
    lambda subterms: f"{subterms[0]}genera",
    lambda subterms: f"{subterms[0]}people",
    lambda subterms: f"{subterms[0]}hertz",
    lambda subterms: f"{subterms[0]}siemens",
    lambda subterms: f"{subterms[0]}brothers",
    lambda subterms: f"{subterms[0]}children",
    lambda subterms: f"{subterms[0]}cows",
    lambda subterms: f"{subterms[0]}menschen",
    lambda subterms: f"{subterms[0]}elves",
    lambda subterms: f"{subterms[0]}hoofs",
    lambda subterms: f"{subterms[0]}knives",
    lambda subterms: f"{subterms[0]}leaves",
    lambda subterms: f"{subterms[0]}loaves",
    lambda subterms: f"{subterms[0]}shelves",
    lambda subterms: f"{subterms[0]}thieves",
    lambda subterms: f"{subterms[0]}wives",
    lambda subterms: f"{subterms[0]}wolves",
    lambda subterms: f"{subterms[0]}humans",
    lambda subterms: f"{subterms[0]}feet",
    lambda subterms: f"{subterms[0]}geese",
    lambda subterms: f"{subterms[0]}men",
    lambda subterms: f"{subterms[0]}mice",
    lambda subterms: f"{subterms[0]}teeth",
    lambda subterms: f"{subterms[0]}buzzes",
    lambda subterms: f"{subterms[0]}fizzes",
    lambda subterms: f"{subterms[0]}frizzes",
    lambda subterms: f"{subterms[0]}fuzzes",
    lambda subterms: f"{subterms[0]}jazzes",
    lambda subterms: f"{subterms[0]}proboscises",
    lambda subterms: f"{subterms[0]}bemas",
    lambda subterms: f"{subterms[0]}dramas",
    lambda subterms: f"{subterms[0]}edemas",
    lambda subterms: f"{subterms[0]}lemmas",
    lambda subterms: f"{subterms[0]}sarcomas",
    lambda subterms: f"{subterms[0]}stomas",
    lambda subterms: f"{subterms[0]}novas",
    lambda subterms: f"{subterms[0]}umbras",
    lambda subterms: f"{subterms[0]}lumens",
    lambda subterms: f"{subterms[0]}data",
    lambda subterms: f"{subterms[0]}stratums",
    lambda subterms: f"{subterms[0]}bacilli",
    lambda subterms: f"{subterms[0]}nuclei",
    lambda subterms: f"{subterms[0]}theses",
    lambda subterms: f"{subterms[0]}hedrons",
    lambda subterms: f"{subterms[0]}helions",
    lambda subterms: f"{subterms[0]}legomena",
    lambda subterms: f"{subterms[0]}phenomena",
    lambda subterms: f"{subterms[0]}helices",
    lambda subterms: f"{subterms[0]}measles",
    lambda subterms: f"{subterms[0]}pox",
    lambda subterms: f"{subterms[0]}basses",
    lambda subterms: f"{subterms[0]}deer",
    lambda subterms: f"{subterms[0]}fish",
    lambda subterms: f"{subterms[0]}fowls",
    lambda subterms: f"{subterms[0]}sheep",
    lambda subterms: f"{subterms[0]}wildebeests",
    lambda subterms: f"{subterms[0]}craft",
    lambda subterms: f"{subterms[0]}star generals",
    lambda subterms: f"{subterms[0]}oux",
    lambda subterms: f"{subterms[0]}os",
    lambda subterms: f"{subterms[0]}lves",
    lambda subterms: f"{subterms[0]}nges",
    lambda subterms: f"{subterms[0]}nxes",
    lambda subterms: f"{subterms[0]}arves",
    lambda subterms: f"{subterms[0]}ceps",
    lambda subterms: f"{subterms[0]}ches",
    lambda subterms: f"{subterms[0]}hes",
    lambda subterms: f"{subterms[0]}eaves",
    lambda subterms: f"{subterms[0]}eaus",
    lambda subterms: f"{subterms[0]}ieus",
    lambda subterms: f"{subterms[0]}nives",
    lambda subterms: f"{subterms[0]}oes",
    lambda subterms: f"{subterms[0]}os",
    lambda subterms: f"{subterms[0]}quies",
    lambda subterms: f"{subterms[0]}ys",
    lambda subterms: f"{subterms[0]}ses",
    lambda subterms: f"{subterms[0]}sses",
    lambda subterms: f"{subterms[0]}uses",
    lambda subterms: f"{subterms[0]}ums",
    lambda subterms: f"{subterms[0]}uses",
    lambda subterms: f"{subterms[0]}ses",
    lambda subterms: f"{subterms[0]}trices",
    lambda subterms: f"{subterms[0]}xes",
    lambda subterms: f"{subterms[0]}ies",
    lambda subterms: f"{subterms[0]}zes",
    lambda subterms: f"{subterms[0]}zzes",
    lambda subterms: f"{subterms[0]}zoa",
    lambda subterms: f"{subterms[0]}ses",
    lambda subterms: f"{subterms[0]}s",
]
modern_plural_convert_slices = [
    (0, [1]),
    (1, [2]),
    (2, [3]),
    (3, [4]),
    (4, [5]),
    (5, [6, 7, 8]),
    (5, [6, 7, 8]),
    (5, [6, 7, 8]),
    (6, [9, 10, 11]),
    (6, [9, 10, 11]),
    (6, [9, 10, 11]),
    (7, [12, 13]),
    (7, [12, 13]),
    (8, [14, 15]),
    (8, [14, 15]),
    (9, [16]),
    (10, [17]),
    (11, [18, 19]),
    (11, [18, 19]),
    (12, [20, 21]),
    (12, [20, 21]),
    (13, [22]),
    (14, [23]),
    (15, [24]),
    (16, [25]),
    (17, [26]),
    (18, [27]),
    (19, [28]),
    (20, [29]),
    (21, [30]),
    (22, [31]),
    (23, [32]),
    (24, [33]),
    (25, [34]),
    (26, [35]),
    (27, [36]),
    (28, [37]),
    (29, [38]),
    (30, [39]),
    (31, [40]),
    (32, [41]),
    (33, [42]),
    (34, [43]),
    (35, [44]),
    (36, [45]),
    (37, [46]),
    (38, [47]),
    (39, [48]),
    (40, [49]),
    (41, [50]),
    (42, [51]),
    (43, [52]),
    (44, [53]),
    (45, [54]),
    (46, [55]),
    (47, [56]),
    (48, [57]),
    (49, [58]),
    (50, [59]),
    (51, [60]),
    (52, [61]),
    (53, [62]),
    (54, [63]),
    (55, [64]),
    (56, [65]),
    (57, [66]),
    (58, [67]),
    (59, [68]),
    (60, [69]),
    (61, [70]),
    (62, [71]),
    (63, [72]),
    (64, [73]),
    (65, [74]),
    (66, [75]),
    (67, [76]),
    (68, [77]),
    (69, [78]),
    (70, [79]),
    (71, [80]),
    (72, [81]),
    (73, [82]),
    (74, [83]),
    (75, [84]),
    (76, [85]),
    (77, [86]),
    (78, [87]),
    (79, [88]),
    (80, [89]),
    (81, [90]),
    (82, [91]),
    (83, [92]),
    (84, [93]),
    (85, [94]),
    (86, [95]),
    (87, [96]),
    (88, [97]),
    (89, [98]),
    (90, [99]),
    (91, [100]),
    (92, [101]),
    (93, [102]),
    (94, [103]),
    (95, [104]),
    (96, [105]),
    (97, [106]),
    (98, [107]),
    (99, [108]),
    (100, [109]),
    (101, [110]),
    (102, [111]),
    (103, [112]),
]

classical_plural_convert_rule_regex = rei(r"^(?:(.*?)-general|(.*?) general|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) it|son-of-a-(.*?)|son of a (.*?)|(.*?)-(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)|(.*?) (about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)|(.*?)-errant|(.*?) errant|(.*?)-(above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)|(.*?) (above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)|(.*)staff|(.*)stave|(.*)genus|(.*)person|(.*)hertz|(.*)siemens|(.*)brother|(.*)child|(.*)cow|(.*)mensch|(.*)elf|(.*)hoof|(.*)knife|(.*)leaf|(.*)loaf|(.*)shelf|(.*)thief|(.*)wife|(.*)wolf|(.*)human|(.*)foot|(.*)goose|(.*)man|(.*)mouse|(.*)tooth|(.*)buzz|(.*)fizz|(.*)frizz|(.*)fuzz|(.*)jazz|(.*)proboscis|(.*)bema|(.*)drama|(.*)edema|(.*)lemma|(.*)sarcoma|(.*)stoma|(.*)nova|(.*)umbra|(.*)lumen|(.*)datum|(.*)stratum|(.*)bacillus|(.*)nucleus|(.+)thesis|(.*)hedron|(.*)helion|(.*)legomenon|(.*)phenomenon|(.*)helix|(.*)measles|(.*)pox|(.*)bass|(.*)deer|(.*)fish|(.*)fowl|(.*)sheep|(.*)wildebeest|(.*)craft|(.+)star general|(.+)oux|(.+[aeiou])o|(.+[aeo])lf|(.+)nge|(.+[aiy])nx|(.+)arf|(.+)ceps|(.+[^aeoui][aeouiy])che|(.+[cs])h|(.+)eaf|(.+)eau|(.+)ieu|(.+)nife|(.+)oe|(.+)o|(.+)quy|(.+[aeiou])y|(.+[rnlpwaeio])se|(.+)ss|(.+[aeo])use|(.+)um|(.+)us|(.+[^ns])sis|(.+)trix|(.+)x|(.+)y|(.+[^aeiouy])z|(.+)z|(.+)zoon|(.+)s|(.+))$")

classical_plural_convert_outputs = [
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-general",
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} general",
    lambda subterms: f"{subterms[0]} them",
    lambda subterms: f"sons-of-{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}",
    lambda subterms: f"sons of {convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}",
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-{subterms[1]}-{subterms[2]}",
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} {subterms[1]} {subterms[2]}",
    lambda subterms: f"{subterms[0]}-{convert_to_classical_plural(subterms[1]) if is_singular(subterms[1]) else subterms[1]}",
    lambda subterms: f"{subterms[0]} {convert_to_classical_plural(subterms[1]) if is_singular(subterms[1]) else subterms[1]}",
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-errant",
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} errant",
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]}-{subterms[1]}",
    lambda subterms: f"{convert_to_classical_plural(subterms[0]) if is_singular(subterms[0]) else subterms[0]} {subterms[1]}",
    lambda subterms: f"{subterms[0]}staves",
    lambda subterms: f"{subterms[0]}staves",
    lambda subterms: f"{subterms[0]}genera",
    lambda subterms: f"{subterms[0]}persons",
    lambda subterms: f"{subterms[0]}hertz",
    lambda subterms: f"{subterms[0]}siemens",
    lambda subterms: f"{subterms[0]}brethren",
    lambda subterms: f"{subterms[0]}children",
    lambda subterms: f"{subterms[0]}kine",
    lambda subterms: f"{subterms[0]}menschen",
    lambda subterms: f"{subterms[0]}elves",
    lambda subterms: f"{subterms[0]}hooves",
    lambda subterms: f"{subterms[0]}knives",
    lambda subterms: f"{subterms[0]}leaves",
    lambda subterms: f"{subterms[0]}loaves",
    lambda subterms: f"{subterms[0]}shelves",
    lambda subterms: f"{subterms[0]}thieves",
    lambda subterms: f"{subterms[0]}wives",
    lambda subterms: f"{subterms[0]}wolves",
    lambda subterms: f"{subterms[0]}humans",
    lambda subterms: f"{subterms[0]}feet",
    lambda subterms: f"{subterms[0]}geese",
    lambda subterms: f"{subterms[0]}men",
    lambda subterms: f"{subterms[0]}mice",
    lambda subterms: f"{subterms[0]}teeth",
    lambda subterms: f"{subterms[0]}buzzes",
    lambda subterms: f"{subterms[0]}fizzes",
    lambda subterms: f"{subterms[0]}frizzes",
    lambda subterms: f"{subterms[0]}fuzzes",
    lambda subterms: f"{subterms[0]}jazzes",
    lambda subterms: f"{subterms[0]}proboscides",
    lambda subterms: f"{subterms[0]}bemata",
    lambda subterms: f"{subterms[0]}dramas",
    lambda subterms: f"{subterms[0]}edemata",
    lambda subterms: f"{subterms[0]}lemmata",
    lambda subterms: f"{subterms[0]}sarcomata",
    lambda subterms: f"{subterms[0]}stomata",
    lambda subterms: f"{subterms[0]}novae",
    lambda subterms: f"{subterms[0]}umbrae",
    lambda subterms: f"{subterms[0]}lumina",
    lambda subterms: f"{subterms[0]}data",
    lambda subterms: f"{subterms[0]}strata",
    lambda subterms: f"{subterms[0]}bacilli",
    lambda subterms: f"{subterms[0]}nuclei",
    lambda subterms: f"{subterms[0]}theses",
    lambda subterms: f"{subterms[0]}hedra",
    lambda subterms: f"{subterms[0]}helia",
    lambda subterms: f"{subterms[0]}legomena",
    lambda subterms: f"{subterms[0]}phenomena",
    lambda subterms: f"{subterms[0]}helices",
    lambda subterms: f"{subterms[0]}measles",
    lambda subterms: f"{subterms[0]}pox",
    lambda subterms: f"{subterms[0]}bass",
    lambda subterms: f"{subterms[0]}deer",
    lambda subterms: f"{subterms[0]}fish",
    lambda subterms: f"{subterms[0]}fowl",
    lambda subterms: f"{subterms[0]}sheep",
    lambda subterms: f"{subterms[0]}wildebeest",
    lambda subterms: f"{subterms[0]}craft",
    lambda subterms: f"{subterms[0]}star generals",
    lambda subterms: f"{subterms[0]}oux",
    lambda subterms: f"{subterms[0]}os",
    lambda subterms: f"{subterms[0]}lves",
    lambda subterms: f"{subterms[0]}nges",
    lambda subterms: f"{subterms[0]}nges",
    lambda subterms: f"{subterms[0]}arves",
    lambda subterms: f"{subterms[0]}ceps",
    lambda subterms: f"{subterms[0]}ches",
    lambda subterms: f"{subterms[0]}hes",
    lambda subterms: f"{subterms[0]}eaves",
    lambda subterms: f"{subterms[0]}eaux",
    lambda subterms: f"{subterms[0]}ieux",
    lambda subterms: f"{subterms[0]}nives",
    lambda subterms: f"{subterms[0]}oes",
    lambda subterms: f"{subterms[0]}os",
    lambda subterms: f"{subterms[0]}quies",
    lambda subterms: f"{subterms[0]}ys",
    lambda subterms: f"{subterms[0]}ses",
    lambda subterms: f"{subterms[0]}sses",
    lambda subterms: f"{subterms[0]}uses",
    lambda subterms: f"{subterms[0]}a",
    lambda subterms: f"{subterms[0]}uses",
    lambda subterms: f"{subterms[0]}ses",
    lambda subterms: f"{subterms[0]}trices",
    lambda subterms: f"{subterms[0]}xes",
    lambda subterms: f"{subterms[0]}ies",
    lambda subterms: f"{subterms[0]}zes",
    lambda subterms: f"{subterms[0]}zzes",
    lambda subterms: f"{subterms[0]}zoa",
    lambda subterms: f"{subterms[0]}ses",
    lambda subterms: f"{subterms[0]}s",
]
classical_plural_convert_slices = [
    (0, [1]),
    (1, [2]),
    (2, [3]),
    (3, [4]),
    (4, [5]),
    (5, [6, 7, 8]),
    (5, [6, 7, 8]),
    (5, [6, 7, 8]),
    (6, [9, 10, 11]),
    (6, [9, 10, 11]),
    (6, [9, 10, 11]),
    (7, [12, 13]),
    (7, [12, 13]),
    (8, [14, 15]),
    (8, [14, 15]),
    (9, [16]),
    (10, [17]),
    (11, [18, 19]),
    (11, [18, 19]),
    (12, [20, 21]),
    (12, [20, 21]),
    (13, [22]),
    (14, [23]),
    (15, [24]),
    (16, [25]),
    (17, [26]),
    (18, [27]),
    (19, [28]),
    (20, [29]),
    (21, [30]),
    (22, [31]),
    (23, [32]),
    (24, [33]),
    (25, [34]),
    (26, [35]),
    (27, [36]),
    (28, [37]),
    (29, [38]),
    (30, [39]),
    (31, [40]),
    (32, [41]),
    (33, [42]),
    (34, [43]),
    (35, [44]),
    (36, [45]),
    (37, [46]),
    (38, [47]),
    (39, [48]),
    (40, [49]),
    (41, [50]),
    (42, [51]),
    (43, [52]),
    (44, [53]),
    (45, [54]),
    (46, [55]),
    (47, [56]),
    (48, [57]),
    (49, [58]),
    (50, [59]),
    (51, [60]),
    (52, [61]),
    (53, [62]),
    (54, [63]),
    (55, [64]),
    (56, [65]),
    (57, [66]),
    (58, [67]),
    (59, [68]),
    (60, [69]),
    (61, [70]),
    (62, [71]),
    (63, [72]),
    (64, [73]),
    (65, [74]),
    (66, [75]),
    (67, [76]),
    (68, [77]),
    (69, [78]),
    (70, [79]),
    (71, [80]),
    (72, [81]),
    (73, [82]),
    (74, [83]),
    (75, [84]),
    (76, [85]),
    (77, [86]),
    (78, [87]),
    (79, [88]),
    (80, [89]),
    (81, [90]),
    (82, [91]),
    (83, [92]),
    (84, [93]),
    (85, [94]),
    (86, [95]),
    (87, [96]),
    (88, [97]),
    (89, [98]),
    (90, [99]),
    (91, [100]),
    (92, [101]),
    (93, [102]),
    (94, [103]),
    (95, [104]),
    (96, [105]),
    (97, [106]),
    (98, [107]),
    (99, [108]),
    (100, [109]),
    (101, [110]),
    (102, [111]),
    (103, [112]),
]

singular_convert_rule_regex = rei(r"^(?:(.*?)-general|(.*?) general|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) them|sons-of-(.*?)|sons of (.*?)|(.*?)-(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)|(.*?) (about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)|(.*?)-errant|(.*?) errant|(.*?)-(above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)|(.*?) (above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)|(.*)staffs|(.*)staves|(.*)genera|(.*)people|(.*)persons|(.*)hertz|(.*)siemens|(.*)brothers|(.*)brethren|(.*)children|(.*)cows|(.*)kine|(.*)menschen|(.*)elves|(.*)hoofs|(.*)hooves|(.*)knives|(.*)leaves|(.*)loaves|(.*)shelves|(.*)thieves|(.*)wives|(.*)wolves|(.*)humans|(.*)feet|(.*)geese|(.*)men|(.*)mice|(.*)teeth|(.*)buzzes|(.*)fizzes|(.*)frizzes|(.*)fuzzes|(.*)jazzes|(.*)proboscises|(.*)proboscides|(.*)bemas|(.*)bemata|(.*)dramas|(.*)edemas|(.*)edemata|(.*)lemmas|(.*)lemmata|(.*)sarcomas|(.*)sarcomata|(.*)stomas|(.*)stomata|(.*)novas|(.*)novae|(.*)umbras|(.*)umbrae|(.*)lumens|(.*)lumina|(.*)data|(.*)stratums|(.*)strata|(.*)bacilli|(.*)nuclei|(.+)theses|(.*)hedrons|(.*)hedra|(.*)helions|(.*)helia|(.*)legomena|(.*)phenomena|(.*)helices|(.*)measles|(.*)pox|(.*)basses|(.*)bass|(.*)deer|(.*)fish|(.*)fowls|(.*)fowl|(.*)sheep|(.*)wildebeests|(.*)wildebeest|(.*)craft|(.+)star generals|(.+)oux|(.+[aeiou])os|(.+[aeo])lves|(.+)nges|(.+[aiy])nxes|(.+[aiy])nges|(.+)arves|(.+)ceps|(.+[^aeoui][aeouiy])ches|(.+[cs])hes|(.+)eaves|(.+)eaus|(.+)eaux|(.+)ieus|(.+)ieux|(.+)nives|(.+)oes|(.+)os|(.+)quies|(.+[aeiou])ys|(.+[rnlpwaeio])ses|(.+)sses|(.+[aeo])uses|(.+)uses|(.+[^ns])ses|(.+)trices|(.+)xes|(.+)ies|(.+)zzes|(.+)zoa|(.+)ae|(.+)ses|(.+)s)$")

singular_convert_outputs = [
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]}-general",
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]} general",
    lambda subterms: f"{subterms[0]} it",
    lambda subterms: f"son-of-a-{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]}",
    lambda subterms: f"son of a {convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]}",
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]}-{subterms[1]}-{subterms[2]}",
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]} {subterms[1]} {subterms[2]}",
    lambda subterms: f"{subterms[0]}-{convert_to_singular(subterms[1]) if is_plural(subterms[1]) else subterms[1]}",
    lambda subterms: f"{subterms[0]} {convert_to_singular(subterms[1]) if is_plural(subterms[1]) else subterms[1]}",
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]}-errant",
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]} errant",
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]}-{subterms[1]}",
    lambda subterms: f"{convert_to_singular(subterms[0]) if is_plural(subterms[0]) else subterms[0]} {subterms[1]}",
    lambda subterms: f"{subterms[0]}staff",
    lambda subterms: f"{subterms[0]}staff",
    lambda subterms: f"{subterms[0]}genus",
    lambda subterms: f"{subterms[0]}person",
    lambda subterms: f"{subterms[0]}person",
    lambda subterms: f"{subterms[0]}hertz",
    lambda subterms: f"{subterms[0]}siemens",
    lambda subterms: f"{subterms[0]}brother",
    lambda subterms: f"{subterms[0]}brother",
    lambda subterms: f"{subterms[0]}child",
    lambda subterms: f"{subterms[0]}cow",
    lambda subterms: f"{subterms[0]}cow",
    lambda subterms: f"{subterms[0]}mensch",
    lambda subterms: f"{subterms[0]}elf",
    lambda subterms: f"{subterms[0]}hoof",
    lambda subterms: f"{subterms[0]}hoof",
    lambda subterms: f"{subterms[0]}knife",
    lambda subterms: f"{subterms[0]}leaf",
    lambda subterms: f"{subterms[0]}loaf",
    lambda subterms: f"{subterms[0]}shelf",
    lambda subterms: f"{subterms[0]}thief",
    lambda subterms: f"{subterms[0]}wife",
    lambda subterms: f"{subterms[0]}wolf",
    lambda subterms: f"{subterms[0]}human",
    lambda subterms: f"{subterms[0]}foot",
    lambda subterms: f"{subterms[0]}goose",
    lambda subterms: f"{subterms[0]}man",
    lambda subterms: f"{subterms[0]}mouse",
    lambda subterms: f"{subterms[0]}tooth",
    lambda subterms: f"{subterms[0]}buzz",
    lambda subterms: f"{subterms[0]}fizz",
    lambda subterms: f"{subterms[0]}frizz",
    lambda subterms: f"{subterms[0]}fuzz",
    lambda subterms: f"{subterms[0]}jazz",
    lambda subterms: f"{subterms[0]}proboscis",
    lambda subterms: f"{subterms[0]}proboscis",
    lambda subterms: f"{subterms[0]}bema",
    lambda subterms: f"{subterms[0]}bema",
    lambda subterms: f"{subterms[0]}drama",
    lambda subterms: f"{subterms[0]}edema",
    lambda subterms: f"{subterms[0]}edema",
    lambda subterms: f"{subterms[0]}lemma",
    lambda subterms: f"{subterms[0]}lemma",
    lambda subterms: f"{subterms[0]}sarcoma",
    lambda subterms: f"{subterms[0]}sarcoma",
    lambda subterms: f"{subterms[0]}stoma",
    lambda subterms: f"{subterms[0]}stoma",
    lambda subterms: f"{subterms[0]}nova",
    lambda subterms: f"{subterms[0]}nova",
    lambda subterms: f"{subterms[0]}umbra",
    lambda subterms: f"{subterms[0]}umbra",
    lambda subterms: f"{subterms[0]}lumen",
    lambda subterms: f"{subterms[0]}lumen",
    lambda subterms: f"{subterms[0]}datum",
    lambda subterms: f"{subterms[0]}stratum",
    lambda subterms: f"{subterms[0]}stratum",
    lambda subterms: f"{subterms[0]}bacillus",
    lambda subterms: f"{subterms[0]}nucleus",
    lambda subterms: f"{subterms[0]}thesis",
    lambda subterms: f"{subterms[0]}hedron",
    lambda subterms: f"{subterms[0]}hedron",
    lambda subterms: f"{subterms[0]}helion",
    lambda subterms: f"{subterms[0]}helion",
    lambda subterms: f"{subterms[0]}legomenon",
    lambda subterms: f"{subterms[0]}phenomenon",
    lambda subterms: f"{subterms[0]}helix",
    lambda subterms: f"{subterms[0]}measles",
    lambda subterms: f"{subterms[0]}pox",
    lambda subterms: f"{subterms[0]}bass",
    lambda subterms: f"{subterms[0]}bass",
    lambda subterms: f"{subterms[0]}deer",
    lambda subterms: f"{subterms[0]}fish",
    lambda subterms: f"{subterms[0]}fowl",
    lambda subterms: f"{subterms[0]}fowl",
    lambda subterms: f"{subterms[0]}sheep",
    lambda subterms: f"{subterms[0]}wildebeest",
    lambda subterms: f"{subterms[0]}wildebeest",
    lambda subterms: f"{subterms[0]}craft",
    lambda subterms: f"{subterms[0]}star general",
    lambda subterms: f"{subterms[0]}oux",
    lambda subterms: f"{subterms[0]}o",
    lambda subterms: f"{subterms[0]}lf",
    lambda subterms: f"{subterms[0]}nge",
    lambda subterms: f"{subterms[0]}nx",
    lambda subterms: f"{subterms[0]}nx",
    lambda subterms: f"{subterms[0]}arf",
    lambda subterms: f"{subterms[0]}ceps",
    lambda subterms: f"{subterms[0]}che",
    lambda subterms: f"{subterms[0]}h",
    lambda subterms: f"{subterms[0]}eaf",
    lambda subterms: f"{subterms[0]}eau",
    lambda subterms: f"{subterms[0]}eau",
    lambda subterms: f"{subterms[0]}ieu",
    lambda subterms: f"{subterms[0]}ieu",
    lambda subterms: f"{subterms[0]}nife",
    lambda subterms: f"{subterms[0]}oe",
    lambda subterms: f"{subterms[0]}o",
    lambda subterms: f"{subterms[0]}quy",
    lambda subterms: f"{subterms[0]}y",
    lambda subterms: f"{subterms[0]}se",
    lambda subterms: f"{subterms[0]}ss",
    lambda subterms: f"{subterms[0]}use",
    lambda subterms: f"{subterms[0]}us",
    lambda subterms: f"{subterms[0]}sis",
    lambda subterms: f"{subterms[0]}trix",
    lambda subterms: f"{subterms[0]}x",
    lambda subterms: f"{subterms[0]}y",
    lambda subterms: f"{subterms[0]}z",
    lambda subterms: f"{subterms[0]}zoon",
    lambda subterms: f"{subterms[0]}a",
    lambda subterms: f"{subterms[0]}s",
    lambda subterms: f"{subterms[0]}",
]
singular_convert_slices = [
    (0, [1]),
    (1, [2]),
    (2, [3]),
    (3, [4]),
    (4, [5]),
    (5, [6, 7, 8]),
    (5, [6, 7, 8]),
    (5, [6, 7, 8]),
    (6, [9, 10, 11]),
    (6, [9, 10, 11]),
    (6, [9, 10, 11]),
    (7, [12, 13]),
    (7, [12, 13]),
    (8, [14, 15]),
    (8, [14, 15]),
    (9, [16]),
    (10, [17]),
    (11, [18, 19]),
    (11, [18, 19]),
    (12, [20, 21]),
    (12, [20, 21]),
    (13, [22]),
    (14, [23]),
    (15, [24]),
    (16, [25]),
    (17, [26]),
    (18, [27]),
    (19, [28]),
    (20, [29]),
    (21, [30]),
    (22, [31]),
    (23, [32]),
    (24, [33]),
    (25, [34]),
    (26, [35]),
    (27, [36]),
    (28, [37]),
    (29, [38]),
    (30, [39]),
    (31, [40]),
    (32, [41]),
    (33, [42]),
    (34, [43]),
    (35, [44]),
    (36, [45]),
    (37, [46]),
    (38, [47]),
    (39, [48]),
    (40, [49]),
    (41, [50]),
    (42, [51]),
    (43, [52]),
    (44, [53]),
    (45, [54]),
    (46, [55]),
    (47, [56]),
    (48, [57]),
    (49, [58]),
    (50, [59]),
    (51, [60]),
    (52, [61]),
    (53, [62]),
    (54, [63]),
    (55, [64]),
    (56, [65]),
    (57, [66]),
    (58, [67]),
    (59, [68]),
    (60, [69]),
    (61, [70]),
    (62, [71]),
    (63, [72]),
    (64, [73]),
    (65, [74]),
    (66, [75]),
    (67, [76]),
    (68, [77]),
    (69, [78]),
    (70, [79]),
    (71, [80]),
    (72, [81]),
    (73, [82]),
    (74, [83]),
    (75, [84]),
    (76, [85]),
    (77, [86]),
    (78, [87]),
    (79, [88]),
    (80, [89]),
    (81, [90]),
    (82, [91]),
    (83, [92]),
    (84, [93]),
    (85, [94]),
    (86, [95]),
    (87, [96]),
    (88, [97]),
    (89, [98]),
    (90, [99]),
    (91, [100]),
    (92, [101]),
    (93, [102]),
    (94, [103]),
    (95, [104]),
    (96, [105]),
    (97, [106]),
    (98, [107]),
    (99, [108]),
    (100, [109]),
    (101, [110]),
    (102, [111]),
    (103, [112]),
    (104, [113]),
    (105, [114]),
    (106, [115]),
    (107, [116]),
    (108, [117]),
    (109, [118]),
    (110, [119]),
    (111, [120]),
    (112, [121]),
    (113, [122]),
    (114, [123]),
    (115, [124]),
    (116, [125]),
    (117, [126]),
    (118, [127]),
    (119, [128]),
    (120, [129]),
    (121, [130]),
    (122, [131]),
    (123, [132]),
    (124, [133]),
]

plural_recognize_rules = {
    rei(r"^(?:(.+)ae|(.+)os|(.+[aeiou])os|(.+[aeiou])ys|(.*)men|(.*)pox|(.+)ies|(.+)oes|(.+)oux|(.+)xes|(.+)zoa|(.+[^ns])ses|(.+[cs])hes|(.+[rnlpwaeio])ses|(.*)bass|(.*)cows|(.*)data|(.*)deer|(.*)feet|(.*)fish|(.*)fowl|(.*)kine|(.*)mice|(.+)ceps|(.+)eaus|(.+)eaux|(.+)ieus|(.+)ieux|(.+)nges|(.+)sses|(.+)uses|(.+)zzes|(.+[^aeoui][aeouiy])ches|(.+[aeo])lves|(.+[aeo])uses|(.+[aiy])nges|(.+[aiy])nxes|(.*)bemas|(.*)craft|(.*)elves|(.*)fowls|(.*)geese|(.*)hedra|(.*)helia|(.*)hertz|(.*)hoofs|(.*)novae|(.*)novas|(.*)sheep|(.*)teeth|(.*)wives|(.+)arves|(.+)eaves|(.+)nives|(.+)quies|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) them|(.*)basses|(.*)bemata|(.*)buzzes|(.*)dramas|(.*)edemas|(.*)fizzes|(.*)fuzzes|(.*)genera|(.*)hooves|(.*)humans|(.*)jazzes|(.*)knives|(.*)leaves|(.*)lemmas|(.*)loaves|(.*)lumens|(.*)lumina|(.*)nuclei|(.*)people|(.*)staffs|(.*)staves|(.*)stomas|(.*)strata|(.*)umbrae|(.*)umbras|(.*)wolves|(.+)theses|(.+)trices|(.*)bacilli|(.*)edemata|(.*)frizzes|(.*)hedrons|(.*)helices|(.*)helions|(.*)lemmata|(.*)measles|(.*)persons|(.*)shelves|(.*)siemens|(.*)stomata|(.*)thieves|(.*)brethren|(.*)brothers|(.*)children|(.*)legomena|(.*)menschen|(.*)sarcomas|(.*)stratums|(.*)phenomena|(.*)sarcomata|(.*)wildebeest|(.*)proboscides|(.*)proboscises|(.*)wildebeests|(.+)star generals)$"): {},
    rei(r"^sons-of-(.*?)$"): {"conditional": lambda match: is_plural(match.group(1))},
    rei(r"^sons of (.*?)$"): {"conditional": lambda match: is_plural(match.group(1))},
    rei(r"^(.*?)-(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)$"): {"conditional": lambda match: is_plural(match.group(1))},
    rei(r"^(.*?) (about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)$"): {"conditional": lambda match: is_plural(match.group(1))},
    rei(r"^(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)$"): {"conditional": lambda match: is_plural(match.group(2))},
    rei(r"^(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)$"): {"conditional": lambda match: is_plural(match.group(2))},
    rei(r"^(.*?)-errant$"): {"conditional": lambda match: is_plural(match.group(1))},
    rei(r"^(.*?) errant$"): {"conditional": lambda match: is_plural(match.group(1))},
    rei(r"^(.*?)-(above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)$"): {"conditional": lambda match: is_plural(match.group(1))},
    rei(r"^(.*?) (above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)$"): {"conditional": lambda match: is_plural(match.group(1))},
}

singular_recognize_rules = {
    rei(r"^(?:(.+)o|(.+)x|(.+)y|(.+)z|(.+[^aeiouy])z|(.+[aeiou])o|(.+[aeiou])y|(.+[cs])h|(.+)oe|(.+)ss|(.+)um|(.+)us|(.+[aeo])lf|(.+[aiy])nx|(.+[rnlpwaeio])se|(.*)cow|(.*)elf|(.*)man|(.*)pox|(.+)arf|(.+)eaf|(.+)eau|(.+)ieu|(.+)nge|(.+)oux|(.+)quy|(.+[^aeoui][aeouiy])che|(.+[^ns])sis|(.+[aeo])use|(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) it|(.*)bass|(.*)bema|(.*)buzz|(.*)deer|(.*)fish|(.*)fizz|(.*)foot|(.*)fowl|(.*)fuzz|(.*)hoof|(.*)jazz|(.*)leaf|(.*)loaf|(.*)nova|(.*)wife|(.*)wolf|(.+)ceps|(.+)nife|(.+)trix|(.+)zoon|(.*)child|(.*)craft|(.*)datum|(.*)drama|(.*)edema|(.*)frizz|(.*)genus|(.*)goose|(.*)helix|(.*)hertz|(.*)human|(.*)knife|(.*)lemma|(.*)lumen|(.*)mouse|(.*)sheep|(.*)shelf|(.*)staff|(.*)stave|(.*)stoma|(.*)thief|(.*)tooth|(.*)umbra|(.*)hedron|(.*)helion|(.*)mensch|(.*)person|(.+)thesis|(.*)brother|(.*)measles|(.*)nucleus|(.*)sarcoma|(.*)siemens|(.*)stratum|(.*)bacillus|(.*)legomenon|(.*)proboscis|(.*)phenomenon|(.*)wildebeest|(.+)star general)$"): {},
    rei(r"^son-of-a-(.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^son of a (.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?)-(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?) (about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)$"): {"conditional": lambda match: is_singular(match.group(2))},
    rei(r"^(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)$"): {"conditional": lambda match: is_singular(match.group(2))},
    rei(r"^(.*?)-errant$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?) errant$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?)-(above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?) (above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^son-of-a-(.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^son of a (.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?)-(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?) (about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with)-(.*?)$"): {"conditional": lambda match: is_singular(match.group(2))},
    rei(r"^(about|above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|in|near|off|of|onto|on|out|over|since|till|to|under|until|unto|upon|with) (.*?)$"): {"conditional": lambda match: is_singular(match.group(2))},
    rei(r"^(.*?)-errant$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?) errant$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?)-(above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)$"): {"conditional": lambda match: is_singular(match.group(1))},
    rei(r"^(.*?) (above|across|after|among|around|athwart|at|before|behind|below|beneath|besides?|between|betwixt|beyond|but|by|during|except|for|from|into|near|of|onto|since|till|to|under|until|unto|upon|with)$"): {"conditional": lambda match: is_singular(match.group(1))},
}

def known_plural(word: str) -> bool:
    """True if `word` is known to be plural.

    Args:
        word (str): Input word or collocation.

    Returns:
        bool: True if `word` is known to be plural.
    """
    return word in singular_of

def known_singular(word):
    """True if `word` is known to be singular.

    Args:
        word (str): Input word or collocation.

    Returns:
        bool: True if `word` is known to be singular.
    """
    return word in modern_plural_of

def convert_to_modern_plural(word: str) -> str:
    """Convert `word` to modern plural form.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: The modern plural form of `word`.
    """
    if word.lower().endswith(("'s", "'")):
        subword = word[:word.rfind("'")]
        if is_singular(subword):
            subword = convert_to_modern_plural(subword)
            return subword + "'" if subword.endswith(('s', 'S')) else subword + "'s"
        return word

    if word in modern_plural_of:
        return modern_plural_of[word]

    if not word.islower() and word.lower() in modern_plural_of:
        return modern_plural_of[word.lower()]

    if is_plural(word) and not is_singular(word, is_word_plural=True):
        return word

    match = modern_plural_convert_rule_regex.match(word)
    if match:
        for i, group in enumerate(match.groups()):
            if group is not None:
                output_id, slices = modern_plural_convert_slices[i]
                return modern_plural_convert_outputs[output_id]([match.group(index) for index in slices])
    return word

def convert_to_classical_plural(word: str) -> str:
    """Convert `word` to classical plural form.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: The classical plural form of `word`.
    """
    if word.lower().endswith(("'s", "'")):
        subword = word[:word.rfind("'")]
        if is_singular(subword):
            subword = convert_to_classical_plural(subword)
            return subword + "'" if subword.endswith(('s', 'S')) else subword + "'s"
        return word

    if word in classical_plural_of:
        return classical_plural_of[word]

    if not word.islower() and word.lower() in classical_plural_of:
        return classical_plural_of[word.lower()]

    if is_plural(word) and not is_singular(word, is_word_plural=True):
        return word

    match = classical_plural_convert_rule_regex.match(word)
    if match:
        for i, group in enumerate(match.groups()):
            if group is not None:
                output_id, slices = classical_plural_convert_slices[i]
                return classical_plural_convert_outputs[output_id]([match.group(index) for index in slices])
    return word

def convert_to_singular(word: str) -> str:
    """Convert `word` to singular form.

    Args:
        word (str): Input word or collocation.

    Returns:
        str: The singular form of `word`.
    """
    if word.lower().endswith(("'s", "'")):
        subword = word[:word.rfind("'")]
        if is_plural(subword):
            subword = convert_to_singular(subword)
            return subword + "'" if subword.endswith(('s', 'S')) else subword + "'s"
        return word

    if word in singular_of:
        return singular_of[word]

    if not word.islower() and word.lower() in singular_of:
        return singular_of[word.lower()]

    if is_singular(word):
        return word

    match = singular_convert_rule_regex.match(word)
    if match:
        for i, group in enumerate(match.groups()):
            if group is not None:
                output_id, slices = singular_convert_slices[i]
                return singular_convert_outputs[output_id]([match.group(index) for index in slices])
    return word

def is_plural(word):
    """Detect whether `word` is in plural form.

    Args:
        word (str): Input word or collocation.

    Returns:
        bool: True if `word` is deemed plural.
    """
    if known_plural(word) or (not word.islower() and known_plural(word.lower())):
        return True

    if known_singular(word) or (not word.islower() and known_singular(word.lower())):
        return False

    for rule in plural_recognize_rules:
        match = rule.match(word)
        if match:
            if not "conditional" in plural_recognize_rules[rule] or plural_recognize_rules[rule]["conditional"](match):
                return True

    return word.endswith(('s', 'S'))

def is_singular(word, is_word_plural=None):
    """Detect whether `word` is in singular form.

    Args:
        word (str): Input word or collocation.

    Returns:
        bool: True if `word` is deemed singular.
    """
    if known_singular(word) or (not word.islower() and known_singular(word.lower())):
        return True

    if known_plural(word) or (not word.islower() and known_plural(word.lower())):
        return False

    for rule in singular_recognize_rules:
        match = rule.match(word)
        if match:
            if not "conditional" in singular_recognize_rules[rule] or singular_recognize_rules[rule]["conditional"](match):
                return True

    if is_word_plural is not None:
        return not is_word_plural
    return not is_plural(word)

